// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"bufio"
	"fmt"
	"os"
	"strings"
	"time"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"golang.org/x/crypto/ssh/terminal"
	yaml "gopkg.in/yaml.v2"

	xbus "xbus.io/go-xbus"
	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/api/control"
	"xbus.io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/utils"
)

// accountCmd represents the account command.
var accountCmd = &cobra.Command{
	Use:   "account",
	Short: "Manage accounts that are allowed to connect to Xbus.",
	Long: `This command contains several sub-commands to manage accounts and the
permissions they have within Xbus.

In particular, it is possible from this interface to accept / reject pending
account requests.`,
}

// accountListCmd represents the account list sub-command.
var accountListCmd = &cobra.Command{
	Use:   "list",
	Short: "List accounts.",
	Long:  "Display accounts allowed to connect to Xbus, as a list.",
	Run:   listAccounts,
}

// accountAcceptCmd represents the account accept sub-command.
var accountAcceptCmd = &cobra.Command{
	Use:   "accept <identifier>",
	Short: "Accept an account.",
	Long:  "Accept a pending account request so it can be allowed to connect to Xbus.",
	Run:   acceptAccount,
}

// accountRejectCmd represents the account reject sub-command.
var accountRejectCmd = &cobra.Command{
	Use:   "reject <identifier>",
	Short: "Reject an account.",
	Long:  "Reject a pending account request; it will not be allowed to connect to Xbus.",
	Run:   rejectAccount,
}

// accountDeleteCmd represents the account delete sub-command.
var accountDeleteCmd = &cobra.Command{
	Use:   "delete <identifier>",
	Short: "Delete an account.",
	Long:  "Delete an account.",
	Run:   deleteAccount,
}

var accountRenewAPIKeyCmd = &cobra.Command{
	Use:   "renew-apikey",
	Short: "Renew the APIKey of an account",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 1 {
			cmd.Usage()
			os.Exit(1)
		}
		os.Exit(accountRenewAPIKey(args[0]))
	},
}

var accountGetAPIKeyCmd = &cobra.Command{
	Use:   "get-apikey",
	Short: "Get the current APIKey of an account",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 1 {
			cmd.Usage()
			os.Exit(1)
		}
		os.Exit(accountGetAPIKey(args[0]))
	},
}

var accountSetAPIKeyCmd = &cobra.Command{
	Use:   "set-apikey",
	Short: "Set the APIKey of an account",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 2 {
			cmd.Usage()
			os.Exit(1)
		}
		os.Exit(accountSetAPIKey(args[0], args[1]))
	},
}

// accountResetCertCmd is 'account reset-cert' sub-command.
var accountResetCertCmd = &cobra.Command{
	Use:   "reset-cert",
	Short: "Reset an account certificate (/!\\ temporarily disable the account /!\\)",
	Long: `Disable an account and set its certificate as renewable,
so the client may submit a new CSR, which then has to be accepted.`,
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 1 {
			cmd.Usage()
			os.Exit(1)
		}
		os.Exit(accountResetCert(args[0]))
	},
}

var accountPasswordCmd = &cobra.Command{
	Use:   "password",
	Short: "Set/change the password of an account",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 1 {
			cmd.Usage()
			os.Exit(1)
		}
		var reset = cmd.Flag("reset").Value.String() == "true"
		os.Exit(accountPassword(args[0], reset))
	},
}

// accountExportCmd is the 'account export' sub-command
var accountExportCmd = &cobra.Command{
	Use:   "export",
	Short: `Export an account configuration file.`,
	Long:  `To be really usable, the certificate will need to be renewed.`,
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 1 {
			cmd.Usage()
			os.Exit(1)
		}
		os.Exit(accountExport(args[0]))
	},
}

// Arguments of the "accept" command.
const (
	acceptCmdArgFingerprint = iota
	acceptCmdArgCount
)

// Arguments of the "reject" command.
const (
	rejectCmdArgFingerprint = iota
	rejectCmdArgCount
)

// Arguments of the "delete" command.
const (
	deleteCmdArgIdentifier = iota
	deleteCmdArgCount
)

// listAccounts parses arguments to list accounts.
func listAccounts(cmd *cobra.Command, args []string) {
	csrHash := viper.GetBool("csr-hash")
	certDetails := viper.GetBool("cert-details")
	os.Exit(doListAccounts(csrHash, certDetails))
}

// doListAccounts lists accounts.
func doListAccounts(csrHash bool, certDetails bool) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	accounts, err := conn.AccountList()
	if err != nil {
		Logger.Errorf("Cannot retrieve accounts: %s", err)
		return 1
	}

	accountCount := len(accounts)
	if accountCount == 0 {
		fmt.Println("No account found.")
	} else {
		fmt.Printf("%d accounts found:\n", accountCount)
		for _, account := range accounts {
			fmt.Println(fmtAccount(account, csrHash, certDetails))
		}
	}
	return 0
}

// acceptAccount parses arguments to accept a pending account request.
func acceptAccount(cmd *cobra.Command, args []string) {
	if len(args) < acceptCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", acceptCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	expire, err := getCertExpire(cmd)
	if err != nil {
		Logger.Errorf("%s", err)
		cmd.Usage()
		os.Exit(1)
	}

	os.Exit(doAcceptAccount(args[acceptCmdArgFingerprint], expire))
}

// doAcceptAccount accepts a pending account request.
func doAcceptAccount(identifier string, expire time.Time) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	id, err := findAccountID(conn, identifier)
	if err == xbus.ErrNotFound {
		Logger.Errorf("No such account: %s", identifier)
		return 1
	} else if err != nil {
		Logger.Errorf("%s", err)
		return 1
	}

	// Got a match! Send an "accept" request.
	repAccount, err := conn.AccountAccept(id, expire)
	if err != nil {
		Logger.Errorf("Could not accept the account %q: %s", id, err)
		return 1
	}
	if repAccount.Status == api.Account_PENDING {
		Logger.Errorf("Could not accept the account %q: Still pending.", id)
		return 1
	}
	Logger.Noticef("Successfully accepted the account %q.", id)
	return 0
}

// rejectAccount parses arguments to reject a pending account request.
func rejectAccount(cmd *cobra.Command, args []string) {
	if len(args) < rejectCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", rejectCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	os.Exit(doRejectAccount(args[rejectCmdArgFingerprint]))
}

// deleteAccount parses arguments to delete an account
func deleteAccount(cmd *cobra.Command, args []string) {
	if len(args) < deleteCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", deleteCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	os.Exit(doDeleteAccount(
		args[deleteCmdArgIdentifier],
		cmd.Flag("delete-actors").Value.String() == "true"))
}

// doRejectAccount rejects a pending account request.
func doRejectAccount(identifier string) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	id, err := findAccountID(conn, identifier)
	if err == xbus.ErrNotFound {
		Logger.Errorf("No such account: %s", identifier)
		return 1
	} else if err != nil {
		Logger.Errorf("%s", err)
		return 1
	}

	// Got a match! Send a "reject" request.
	repAccount, err := conn.AccountReject(id)
	if err != nil {
		Logger.Errorf("Could not reject the account %q: %s", id, err)
		return 1
	}
	if repAccount.Status == api.Account_PENDING {
		Logger.Errorf("Could not reject the account %q: Still pending.", id)
		return 1
	}
	Logger.Noticef("Successfully rejected the account %q.", id)
	return 0
}

// doDeleteAccount deletes an account
func doDeleteAccount(identifier string, deleteActors bool) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Errorf("Error getting account %q: %s", identifier, err)
		return 1
	}

	actors, err := conn.ActorList([]api.Actor{{AccountID: account.ID}})
	if err != nil {
		Logger.Errorf("Error listing account %q actors: %s", identifier, err)
		return 1

	}

	if len(actors) != 0 && !deleteActors {
		// Get confirmation from the user

		fmt.Printf("The following actors will be deleted along with the account:\n")
		for _, actor := range actors {
			fmt.Printf(" - %s\n", actor.Name)
		}
		fmt.Printf("Do you confirm the deletion (y/N)? ")
		reader := bufio.NewReader(os.Stdin)
		text, _ := reader.ReadString('\n')
		if !(text == "Y\n" || text == "y\n" || text == "yes\n") {
			return 1
		}
	}

	// Got a match! Send a "delete" request.
	err = conn.AccountDelete(account.ID)
	if err != nil {
		Logger.Errorf("Could not delete the account %q: %s", account.Name, err)
		return 1
	}
	return 0
}

func accountRenewAPIKey(identifier string) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Errorf("Could not find account %s: %s", identifier, err)
		return 1
	}

	account, err = conn.AccountRenewAPIKey(account.ID)
	if err != nil {
		Logger.Errorf("Error renewing the account: %s", err)
		return 1
	}

	Logger.Noticef("New API Key:\n%s", account.APIKey)

	return 0
}

func accountGetAPIKey(identifier string) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Errorf("Could not find account %s: %s", identifier, err)
		return 1
	}

	account, err = conn.Control.Account.GetAPIKey(
		control.AccountIDRequest{AccountID: account.ID})
	if err != nil {
		Logger.Errorf("Error getting the account apikey: %s", err)
		return 1
	}

	Logger.Noticef("API Key:\n%s", account.APIKey)

	return 0
}

func accountSetAPIKey(identifier, key string) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Errorf("Could not find account %s: %s", identifier, err)
		return 1
	}

	account.APIKey = key

	account, err = conn.Control.Account.SetAPIKey(
		control.AccountUpdateRequest{Account: account})
	if err != nil {
		Logger.Errorf("Error setting the account key: %s", err)
		return 1
	}

	Logger.Noticef("New API Key:\n%s", account.APIKey)

	return 0
}

func accountResetCert(identifier string) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Errorf("Could not find account %s: %s", identifier, err)
		return 1
	}

	// TODO add a confirmation message
	if err := conn.Control.Account.ResetCert(
		control.AccountIDRequest{AccountID: account.ID},
	); err != nil {
		Logger.Errorf("Failed: %s", err)
	}
	Logger.Noticef("The account %s is now ready for a certificate renewal.", account.Name)
	return 0
}

func accountPassword(identifier string, reset bool) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Errorf("Could not find account %s: %s", identifier, err)
		return 1
	}

	var newpassword string

	if !reset {
		fd := int(os.Stdin.Fd())
		if !terminal.IsTerminal(fd) {
			Logger.Errorf("Cannot read password from non-terminal")
			return 1
		}
		if outFd := int(os.Stdout.Fd()); !terminal.IsTerminal(outFd) {
			Logger.Errorf("Cannot read password if stdin is a terminal and stdout is not")
			return 1
		}

		os.Stdout.WriteString("New password: ")
		b, err := terminal.ReadPassword(fd)
		os.Stdout.WriteString("\n")
		if err != nil {
			Logger.Errorf("Failed to read password: %s", err)
			return 1
		}
		newpassword = string(b)

		os.Stdout.WriteString("Confirm password: ")
		b, err = terminal.ReadPassword(fd)
		os.Stdout.WriteString("\n")
		if err != nil {
			Logger.Errorf("Failed to read password: %s", err)
			return 1
		}
		if newpassword != string(b) {
			Logger.Errorf("Passwords mismatched. Please try again")
			return 1
		}

	}

	// TODO add a confirmation message
	if err := conn.Control.Account.PasswordSet(
		control.AccountPasswordSetRequest{
			AccountID: account.ID,
			Password:  newpassword,
		},
	); err != nil {
		Logger.Errorf("Failed: %s", err)
	}
	if reset {
		Logger.Noticef("Reset %s password.", account.Name)
	} else {
		Logger.Noticef("Updated %s password.", account.Name)
	}
	return 0
}

func accountExport(identifier string) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Errorf("Could not find account %s: %s", identifier, err)
		return 1
	}

	actors, err := conn.ActorList([]api.Actor{{AccountID: account.ID}})
	if err != nil {
		Logger.Errorf("Could not load actors of account %s: %s", identifier, err)
		return 1
	}

	var actorsOutput yaml.MapSlice
	for i, actor := range actors {
		key := actor.Name
		conf := yaml.MapSlice{
			{Key: "kind", Value: strings.ToLower(actor.Kind.String())},
			{Key: "roles", Value: actor.Roles},
		}
		if strings.Contains(key, ".") {
			key = fmt.Sprintf("%d", i)
			conf = append(conf, yaml.MapItem{Key: "name", Value: actor.Name})
		}
		actorsOutput = append(actorsOutput, yaml.MapItem{
			Key:   key,
			Value: conf,
		})
	}

	output := yaml.MapSlice{
		{Key: "account-name", Value: account.Name},
		{Key: "nats-url", Value: conn.GetConn().ConnectedUrl()},
		{Key: "actors", Value: actorsOutput},
	}
	b, err := yaml.Marshal(output)
	if err != nil {
		Logger.Errorf("Error marshaling the account data: %s", err)
		return 1
	}
	os.Stdout.Write(b)
	return 0
}

func findAccountID(conn *xbus.Client, identifier string) (api.UUID, error) {
	if id, err := api.UUIDFromString(identifier); err == nil {
		return id, nil
	}
	account, err := conn.AccountFind(identifier)
	return account.ID, err
}

func fmtAccount(account api.Account, csrHash bool, certDetails bool) string {
	formattedAccount := fmt.Sprintf(`
ID: %s
Name: %s
Type: %s
Status: %s`,
		account.ID,
		account.Name,
		account.Type,
		account.Status,
	)
	if csrHash {
		formattedAccount += fmt.Sprintf(`
CSR fingerprint: %s`,
			tools.GetCSRHash(account.CSR, true),
		)
	}
	if certDetails {
		formattedAccount += fmt.Sprintf(`
CSR: %s
Cert: %s`,
			account.CSR,
			account.Cert,
		)
	}
	return formattedAccount
}

func getCertExpire(cmd *cobra.Command) (time.Time, error) {
	expireString, err := cmd.Flags().GetString("cert-expire")
	if err != nil {
		return time.Time{}, err
	}
	return utils.ParseDateOrDuration(expireString)
}

func init() {
	RootCmd.AddCommand(accountCmd)

	accountCmd.AddCommand(accountListCmd)
	accountListFlags := accountListCmd.PersistentFlags()
	accountListFlags.Bool("cert-details", false, "Display cert details")
	viper.BindPFlag("cert-details", accountListFlags.Lookup("cert-details"))
	accountListFlags.Bool("csr-hash", true, "Display CSR fingerprints")
	viper.BindPFlag("csr-hash", accountListFlags.Lookup("csr-hash"))

	flags := accountAcceptCmd.Flags()
	flags.String("cert-expire", "1y", "Expire date or duration for the account certificate")

	viper.BindPFlag("accountaccept.cert-expire", flags.Lookup("cert-expire"))

	flags = accountPasswordCmd.Flags()
	flags.Bool("reset", false, "Reset the account password")

	accountCmd.AddCommand(accountAcceptCmd)
	accountCmd.AddCommand(accountRejectCmd)

	flags = accountDeleteCmd.Flags()
	flags.Bool("delete-actors", false, "Force delete the related actors")

	accountCmd.AddCommand(accountDeleteCmd)

	accountCmd.AddCommand(accountRenewAPIKeyCmd)
	accountCmd.AddCommand(accountGetAPIKeyCmd)
	accountCmd.AddCommand(accountSetAPIKeyCmd)
	accountCmd.AddCommand(accountResetCertCmd)
	accountCmd.AddCommand(accountPasswordCmd)
	accountCmd.AddCommand(accountExportCmd)
}
