// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"context"
	"errors"
	"fmt"
	"os"
	"time"

	"github.com/spf13/cobra"

	"xbus.io/go-xbus/api"
)

var envelopeCmd = &cobra.Command{
	Use:   "envelope",
	Short: "Envelope management",
}

var envelopePurgeCmd = &cobra.Command{
	Use:   "purge",
	Short: "Remove envelopes that are not in an existing process",
	Args:  cobra.NoArgs,
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient(false)
		if err != nil {
			return
		}
		ctx, _ := contextCancelOnSignal(
			context.Background(), os.Interrupt,
		)
		var count uint32
		fmt.Print("\nPurge: ")
		err = c.EnvelopePurge(
			ctx, func(c uint32) {
				fmt.Printf("\rPurge: %d", c)
				count = c
			},
		)
		if err != nil {
			Logger.Errorf("%s", err)
			return
		}
		Logger.Noticef("Purged %d envelope(s)", count)
	},
}

var envelopeGetCmd = &cobra.Command{
	Use:   "get <ID>",
	Short: "Retrieve a single envelope",
	Args: func(cmd *cobra.Command, args []string) error {
		if err := cobra.ExactArgs(1)(cmd, args); err != nil {
			return err
		}
		_, err := api.UUIDFromString(args[0])
		if err != nil {
			return errors.New("Invalid uuid")
		}
		return nil
	},
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient(false)
		if err != nil {
			return
		}
		id := api.MustUUIDFromString(args[0])
		ctx, _ := contextCancelOnSignal(
			context.Background(), os.Interrupt,
		)
		if cmd.Flag("stream").Value.String() == "true" {
			if err := c.API.EnvelopeStorage.Retrieve(
				ctx, "account", c.ID().String(),
				api.EnvelopePosition{EnvelopeID: id, Start: true},
				func(ctx context.Context, fragment api.EnvelopeFragmentAndPosition) {
					b, err := api.JSONMarshal(
						[]api.ItemEncoding{api.ItemJSON, api.ItemText, api.ItemBinary},
						fragment.GetFragment())
					if err != nil {
						Logger.Errorf("%s", err)
					}
					os.Stdout.Write(b)
					os.Stdout.Write([]byte("\n"))
				},
			); err != nil {
				Logger.Errorf("receiving: %s", err)
			}
		} else {
			receiver := c.RetrieveEnvelope(ctx, "account", c.ID().String(), id)
			env, err := receiver.Complete(5 * time.Second)
			if err != nil {
				Logger.Errorf("receiving: %s", err)
				os.Exit(1)
			}
			b, err := api.JSONMarshal(
				[]api.ItemEncoding{api.ItemJSON, api.ItemText, api.ItemBinary},
				env)
			if err != nil {
				Logger.Errorf("%s", err)
				os.Exit(1)
			}
			os.Stdout.Write(b)
		}
	},
}

func init() {
	RootCmd.AddCommand(envelopeCmd)
	// envelope purge
	envelopeCmd.AddCommand(envelopePurgeCmd)

	// envelope get
	flags := envelopeGetCmd.PersistentFlags()
	flags.Bool("stream", false, "Store a streamed version of the envelope")
	envelopeCmd.AddCommand(envelopeGetCmd)
}
