// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"io"
	"io/ioutil"
	"os"

	"xbus.io/go-xbus/api"

	"github.com/spf13/cobra"
)

func pipelineInfo(identifier, version string) api.PipelineInfo {
	var info api.PipelineInfo
	if id, err := api.UUIDFromString(identifier); err == nil {
		info.ID = id
	} else {
		info.Name = identifier
		info.Version = version
	}
	return info
}

func pipelineList(cmd *cobra.Command, args []string) {
	c, err := GetClient(false)
	if err != nil {
		return
	}
	if len(args) > 1 {
		cmd.Usage()
		os.Exit(-1)
	}
	var name string
	if len(args) == 1 {
		name = args[0]
	}

	pipelines, err := c.PipelineQuery(name)
	if err != nil {
		Logger.Errorf("%s", err)
		return
	}
	for _, info := range pipelines {
		fmt.Printf("%v\n", info)
	}
}

func pipelineGet(identifier, version, filename string) error {
	c, err := GetClient(false)
	if err != nil {
		return err
	}

	info := pipelineInfo(identifier, version)

	str, err := c.PipelineGet(info)
	if err != nil {
		return err
	}

	var out io.Writer
	if filename == "-" {
		out = os.Stdout
	} else {
		f, err := os.Create(filename)
		if err != nil {
			return err
		}
		defer f.Close()
		out = f
	}
	out.Write([]byte(str))
	return nil
}

func pipelineSave(identifier, version, filename string) error {
	c, err := GetClient(false)
	if err != nil {
		return err
	}

	content, err := ioutil.ReadFile(filename)
	if err != nil {
		return fmt.Errorf("Cannot read pipeline definition file: %s", err)
	}

	info := pipelineInfo(identifier, version)

	info, _, validationMessages, err := c.PipelineSave(info, string(content))
	if len(validationMessages) != 0 {
		fmt.Println("Validation messages:")
		for _, msg := range validationMessages {
			fmt.Println("-", msg)
		}
	}

	if err != nil {
		return err
	}

	Logger.Noticef("Successfully saved pipeline %s-%s graph", info.Name, info.Version)
	return nil
}

func pipelineStatus(identifier, version string, status api.PipelineInfo_Status) error {
	c, err := GetClient(false)
	if err != nil {
		return err
	}

	info := pipelineInfo(identifier, version)
	info.Status = status

	info, err = c.PipelineSetStatus(info)
	if err != nil {
		return err
	}
	Logger.Noticef("%s-%s status changed to %s", info.Name, info.Version, info.Status)
	return nil
}

var pipelineCmd = &cobra.Command{
	Use:   "pipeline",
	Short: "Pipeline configuration subcommands",
	Long:  "Pipeline subcommands allow to list and modify pipelines",
}

var pipelineListCmd = &cobra.Command{
	Use:   "list [name]",
	Short: "List pipelines, optionnaly filtered by name",
	Run:   pipelineList,
}

var pipelineGetCmd = &cobra.Command{
	Use:   "get <name or id> [version]",
	Short: "Get a pipeline definition",
	Long:  "Get a pipeline definition (its graph) and write it to stdout",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) == 0 || len(args) > 2 {
			cmd.Usage()
			os.Exit(-1)
		}
		var (
			identifier = args[0]
			version    string
			filename   = cmd.Flag("output").Value.String()
		)
		if len(args) == 2 {
			version = args[1]
		}
		if err := pipelineGet(identifier, version, filename); err != nil {
			Logger.Errorf("Error getting pipeline: %s", err)
			os.Exit(-1)
		}
	},
}

var pipelineSaveCmd = &cobra.Command{
	Use:   "save <name or id> [version] <filepath>",
	Short: "Save a pipeline",
	Long:  "Save a pipeline definition (its graph)",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) < 2 || len(args) > 3 {
			cmd.Usage()
			os.Exit(-1)
		}
		var (
			identifier = args[0]
			version    string
			filename   string
		)
		if len(args) == 2 {
			filename = args[1]
		} else {
			version = args[1]
			filename = args[2]
		}
		if err := pipelineSave(identifier, version, filename); err != nil {
			Logger.Errorf("Error saving pipeline: %s", err)
			os.Exit(-1)
		}
	},
}

var pipelineStatusCmd = &cobra.Command{
	Use:   "status <name or id> [version] [--active|--inactive]",
	Short: "Change the status of a pipeline version",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) < 1 || len(args) > 2 {
			cmd.Usage()
			os.Exit(-1)
		}
		var (
			identifier = args[0]
			version    string
		)
		if len(args) == 2 {
			version = args[1]
		}
		var status api.PipelineInfo_Status
		if cmd.Flag("active").Value.String() == "true" {
			status = api.PipelineInfo_ACTIVE
		} else if cmd.Flag("inactive").Value.String() == "true" {
			status = api.PipelineInfo_INACTIVE
		}
		if err := pipelineStatus(identifier, version, status); err != nil {
			Logger.Errorf("Error setting pipeline status: %s", err)
			os.Exit(-1)
		}
	},
}

func init() {
	RootCmd.AddCommand(pipelineCmd)
	pipelineCmd.AddCommand(pipelineListCmd)
	pipelineCmd.AddCommand(pipelineGetCmd)
	pipelineCmd.AddCommand(pipelineSaveCmd)
	pipelineCmd.AddCommand(pipelineStatusCmd)

	pipelineStatusCmd.Flags().Bool("active", false, "Activate the pipeline")
	pipelineStatusCmd.Flags().Bool("inactive", false, "Deactivate the pipeline")

	pipelineGetCmd.Flags().StringP("output", "o", "-", "Target filename (default to stdout)")
}
