// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"bufio"
	"context"
	"fmt"
	"os"
	"os/signal"
	"regexp"
	"strconv"
	"strings"
	"sync"
	"time"

	xbus "xbus.io/go-xbus"
	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/api/control"

	"github.com/olekukonko/tablewriter"
	"github.com/spf13/cobra"
)

var durationRe = regexp.MustCompile("^([0-9]+)([smhdw])$")
var unixEpochRe = regexp.MustCompile("^[0-9]+$")

func readDatetimeArg(arg string) (time.Time, error) {
	if matches := durationRe.FindStringSubmatch(arg); len(matches) != 0 {
		// its a duration, lets interpret it
		qty, err := strconv.Atoi(matches[1])
		if err != nil {
			return time.Time{}, err
		}
		var duration time.Duration
		switch matches[2] {
		case "s":
			duration = time.Second * time.Duration(qty)
		case "m":
			duration = time.Minute * time.Duration(qty)
		case "h":
			duration = time.Hour * time.Duration(qty)
		case "d":
			duration = time.Hour * 24 * time.Duration(qty)
		case "w":
			duration = time.Hour * 24 * 7 * time.Duration(qty)
		default:
			return time.Time{}, fmt.Errorf("Unknown duration unit: %s", matches[2])
		}
		return time.Now().Add(-duration), nil
	}
	if unixEpochRe.MatchString(arg) {
		// its a unix timestamp
		qty, err := strconv.Atoi(arg)
		if err != nil {
			return time.Time{}, err
		}
		return time.Unix(int64(qty), 0), nil
	}
	// It's not a duration nor a unix timestamp, attempt to parse it as a
	// datetime, allowing various iso-like formats
	for _, format := range []string{
		time.RFC3339,
		"2006-05-04T15:02:01Z0700",
		"2006-05-04T15:02:01",
		"2006-05-04 15:02:01Z0700",
		"2006-05-04 15:02:01",
		"2006-05-04",
	} {
		t, err := time.ParseInLocation(format, arg, time.Local)
		if err == nil {
			return t, nil
		}
	}
	return time.Time{}, fmt.Errorf("Could not parse as a valid duration or date: %s", arg)
}

func contextCancelOnSignal(ctx context.Context, s ...os.Signal) (context.Context, context.CancelFunc) {
	ctx, cancel := context.WithCancel(ctx)

	c := make(chan os.Signal, 1)
	for _, sig := range s {
		signal.Notify(c, sig)
	}

	go func() {
		select {
		case <-ctx.Done():
		case <-c:
			signal.Stop(c)
			cancel()
		}
	}()

	return ctx, cancel
}

var psCmd = &cobra.Command{
	Use:   "ps",
	Short: "Process list",
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient(false)
		if err != nil {
			return
		}
		filter, err := getProcessFilter(
			c, args, filterFlags{IncludeLiveStatuses: true})
		if err != nil {
			Logger.Errorf("Invalid filter: %s", err)
			return
		}
		processList, err := getProcessList(c, filter, "", false)
		if err != nil {
			Logger.Errorf("Error: %s", err)
			return
		}
		pipelineList, err := c.PipelineQuery("")
		if err != nil {
			Logger.Errorf("%s", err)
			return
		}
		actorList, err := c.ActorList(nil)
		if err != nil {
			Logger.Errorf("%s", err)
			return
		}
		printProcessList(processList, pipelineList, actorList)
	},
}

var psExportCmd = &cobra.Command{
	Use:   "export",
	Short: "Export process data",
	Run: func(cmd *cobra.Command, args []string) {
		var (
			quiet             = cmd.Flag("quiet").Value.String() == "true"
			full              = cmd.Flag("full").Value.String() == "true"
			exportEnvelope    = cmd.Flag("export-envelope").Value.String() == "true"
			exportJobs        = cmd.Flag("export-jobs").Value.String() == "true"
			exportLogs        = cmd.Flag("export-logs").Value.String() == "true"
			exportProcessLogs = cmd.Flag("export-process-logs").Value.String() == "true"

			exportEnvelopeSizeLimit int64
			exportLogLevel          api.LogLevel
		)
		if err := exportLogLevel.UnmarshalText(
			[]byte(cmd.Flag("export-logs-level").Value.String()),
		); err != nil {
			Logger.Errorf(
				"Invalid log level: '%s'. Got %s",
				cmd.Flag("export-logs-level").Value.String(), err)
			os.Exit(1)
		}

		if flag := cmd.Flag("export-envelope-size-limit"); flag.Changed {
			value, err := strconv.Atoi(flag.Value.String())
			if err != nil {
				Logger.Errorf(
					"Invalid value '%s' for --export-envelope-size-limit: %s",
					flag.Value.String(), err)
				os.Exit(1)
			}
			exportEnvelopeSizeLimit = int64(value)
		}

		c, err := GetClient(false)
		if err != nil {
			return
		}
		filter, err := getProcessFilter(
			c, args, filterFlags{IncludeAllStatuses: true})
		if err != nil {
			Logger.Errorf("Invalid filter: %s", err)
			return
		}
		list, err := getProcessList(c, filter, "", quiet)
		var (
			pids     []api.UUID
			pidbatch [][]api.UUID
		)
		for _, p := range list {
			pids = append(pids, p.ID)
			if len(pids) >= 10000 {
				pidbatch = append(pidbatch, pids)
				pids = nil
			}
		}
		if len(pids) > 0 {
			pidbatch = append(pidbatch, pids)
		}
		ctx, _ := contextCancelOnSignal(
			context.Background(), os.Interrupt,
		)

		wg := sync.WaitGroup{}
		wg.Add(1)
		ch := make(chan control.OctetStreamList, 1000)

		go func() {
			for msg := range ch {
				for _, chunk := range msg.Chunks {
					fmt.Println(string(chunk.Data))
				}
			}
			wg.Done()
		}()

		for _, pids := range pidbatch {
			if err := c.Control.Process.ExportStream(
				ctx,
				control.ProcessExportRequest{
					ProcessIDs:        pids,
					ExportEnvelope:    exportEnvelope,
					EnvelopeSizeLimit: exportEnvelopeSizeLimit,
					ExportJobs:        exportJobs,
					ExportLogs:        exportLogs,
					ExportProcessLogs: exportProcessLogs,
					ExportFull:        full,
				},
				func(ctx context.Context, msg control.OctetStreamList) {
					ch <- msg
				},
			); err != nil {
				Logger.Errorf("Error exporting stream: %s", err)
				os.Exit(1)
			}
		}
		close(ch)
		wg.Wait()
	},
}

var psPurgeCmd = &cobra.Command{
	Use:   "purge",
	Short: "Purge process data",
	Long: `Remove all the data related to some process(es).
Defaults to all the processes terminated more than 1 month ago`,
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient(false)
		if err != nil {
			return
		}
		filter, err := getProcessFilter(
			c, args, filterFlags{
				IncludeTerminatedStatuses: true,
				StatusUnchangedSince:      "31d", // Approx 1 month
			})
		if err != nil {
			Logger.Errorf("Invalid filter: %s", err)
			return
		}
		confirm := "Do you confirm the definitive deletion of %d process(es) (there is NO going back!) (yes/NO)?"
		if cmd.Flag("yes").Value.String() == "true" {
			confirm = ""
		}
		list, err := getProcessList(c, filter, confirm, false)
		if err != nil {
			fmt.Println(err)
			os.Exit(1)
		}
		if len(list) == 0 {
			fmt.Println("No process matched the filter")
			os.Exit(0)
		}
		var (
			pids     []api.UUID
			pidbatch [][]api.UUID
		)
		for _, p := range list {
			pids = append(pids, p.ID)
			if len(pids) >= 10000 {
				pidbatch = append(pidbatch, pids)
				pids = nil
			}
		}
		if len(pids) > 0 {
			pidbatch = append(pidbatch, pids)
		}
		var (
			total         = len(list)
			offset uint32 = 0
		)
		ctx, _ := contextCancelOnSignal(
			context.Background(), os.Interrupt,
		)
		fmt.Print("\nPurge: ")
		for _, batch := range pidbatch {
			if err := c.ProcessPurge(ctx, batch, func(pr uint32, maxPR uint32) {
				fmt.Printf("\rPurge: %d/%d", offset+pr, total)
			}); err != nil {
				fmt.Print("\n")
				Logger.Errorf("Error: %s", err)
				return
			}
			offset += uint32(len(batch))
		}
		fmt.Print("\n")
		Logger.Noticef("Purged %d process(es)", offset)
	},
}

var psPauseCmd = &cobra.Command{
	Use:   "pause",
	Short: "Pause a process",
	Run: func(cmd *cobra.Command, args []string) {
		os.Exit(doCommand(
			args, filterFlags{}, control.ProcessControlRequest_PAUSE,
			cmd.Flag("yes").Value.String() == "true",
		))
	},
}

var psResumeCmd = &cobra.Command{
	Use:   "resume",
	Short: "Resume a process",
	Run: func(cmd *cobra.Command, args []string) {
		os.Exit(doCommand(
			args, filterFlags{}, control.ProcessControlRequest_RESUME,
			cmd.Flag("yes").Value.String() == "true",
		))
	},
}

var psCancelCmd = &cobra.Command{
	Use:   "cancel",
	Short: "Cancel a process",
	Run: func(cmd *cobra.Command, args []string) {
		os.Exit(doCommand(
			args, filterFlags{}, control.ProcessControlRequest_CANCEL,
			cmd.Flag("yes").Value.String() == "true",
		))
	},
}

func doCommand(
	args []string, defaultFilterFlags filterFlags,
	command control.ProcessControlRequest_Command, confirm bool,
) int {
	c, err := GetClient(false)
	if err != nil {
		return 1
	}
	filter, err := getProcessFilter(
		c, args, defaultFilterFlags)
	if err != nil {
		Logger.Errorf("Invalid filter: %s", err)
		return 1
	}
	msg := ""
	if confirm {
		msg = fmt.Sprintf(
			"Will run the command %s on %%d process(es). Do you confirm (yes/NO)?", command,
		)
	}
	list, err := getProcessList(c, filter, msg, false)
	if err != nil {
		fmt.Println(err)
		os.Exit(1)
	}
	if len(list) == 0 {
		fmt.Println("No process matched the filter")
		os.Exit(0)
	}
	errorCount := 0
	for i, p := range list {
		err = c.ProcessControl(p.ID, command)
		if err != nil {
			errorCount++
			Logger.Errorf("Process %s: %s", p.ID, err)
		}
		fmt.Printf("\r%d success, %d errors / %d     ", (i + 1 - errorCount), errorCount, len(list))
	}
	fmt.Printf("\n%d success, %d errors", len(list)-errorCount, errorCount)
	if errorCount != 0 {
		Logger.Errorf("Encountered %d errors, the command could not be applied to all processes.", errorCount)
		return 1
	}
	return 0
}

func getProcessList(client *xbus.Client, filter api.ProcessFilter, confirmText string, quiet bool) ([]api.Process, error) {
	var processList []api.Process
	if !quiet {
		fmt.Printf("Loading process list...")
	}
	ctx, _ := contextCancelOnSignal(
		context.Background(), os.Interrupt,
	)
	if err := client.ProcessQuery(ctx, filter, func(list []api.Process) {
		processList = append(processList, list...)
		if !quiet {
			fmt.Printf("\rLoading process list... %d", len(processList))
		}
	}); err != nil {
		return nil, err
	}
	if !quiet {
		fmt.Printf("\rLoaded %d processes       \n", len(processList))
	}
	if !quiet && len(processList) != 0 && confirmText != "" {
		printProcessList(processList, nil, nil)
		reader := bufio.NewReader(os.Stdin)
		fmt.Printf(confirmText, len(processList))
		text, _ := reader.ReadString('\n')
		if !(text == "Y\n" || text == "y\n" || text == "yes\n") {
			return nil, nil
		}
	}
	return processList, nil
}

func printProcessList(
	processList []api.Process, pipelineList []api.PipelineInfo, actorList []api.Actor,
) {
	var pList = processList
	if len(processList) > 1000 {
		pList = pList[:1000]
	}

	pipelines := make(map[api.UUID]api.PipelineInfo)
	for _, g := range pipelineList {
		pipelines[g.ID] = g
	}
	actors := make(map[api.UUID]api.Actor)
	for _, a := range actorList {
		actors[a.ID] = a
	}
	table := tablewriter.NewWriter(os.Stdout)
	table.SetHeader([]string{"id", "created at", "pipeline", "emitter", "status"})

	for _, p := range pList {
		createdAt := ""
		if p.CreatedAt != nil {
			createdAt = p.CreatedAt.Format(time.RFC3339)
		}
		pipeline := p.PipelineID.String()
		if g, ok := pipelines[p.PipelineID]; ok {
			pipeline = g.GetName() + " " + g.GetVersion()
		}
		actor := p.TriggerEmitterID.String()
		if a, ok := actors[p.TriggerEmitterID]; ok {
			actor = a.GetName()
		}
		table.Append([]string{
			p.ID.String(),
			createdAt,
			pipeline,
			actor,
			p.Status.String(),
		})
	}
	table.Render()
	if len(processList) > 1000 {
		fmt.Printf("+ %d more processes...\n", len(processList)-1000)
	}
}

type filterFlags struct {
	CreatedFrom               string
	CreatedTo                 string
	StatusChangedSince        string
	StatusUnchangedSince      string
	IncludeStatusInitial      bool
	IncludeStatusRunning      bool
	IncludeStatusPaused       bool
	IncludeStatusDone         bool
	IncludeStatusError        bool
	IncludeLiveStatuses       bool
	IncludeTerminatedStatuses bool
	IncludeAllStatuses        bool
	AckedResponseOnly         bool
	NotAckedResponseOnly      bool
	EnvelopeIDs               string
	Pipelines                 string
}

var globalFlags filterFlags

func getProcessFilter(client *xbus.Client, args []string, defaults filterFlags) (api.ProcessFilter, error) {
	var (
		flags  = globalFlags
		filter api.ProcessFilter
	)

	if (globalFlags == filterFlags{}) && len(args) == 0 {
		flags = defaults
	}

	if flags.CreatedFrom != "" {
		t, err := readDatetimeArg(flags.CreatedFrom)
		if err != nil {
			return filter, fmt.Errorf("Error reading 'created-from': %s", err)
		}
		filter.CreatedFrom = &t
	}

	if flags.CreatedTo != "" {
		t, err := readDatetimeArg(flags.CreatedTo)
		if err != nil {
			return filter, fmt.Errorf("Error reading 'created-to': %s", err)
		}
		filter.CreatedTo = &t
	}

	if flags.StatusChangedSince != "" {
		t, err := readDatetimeArg(flags.StatusChangedSince)
		if err != nil {
			return filter, fmt.Errorf("Error reading 'status-changed-since': %s", err)
		}
		filter.StatusChangedFrom = &t
	}

	if flags.StatusUnchangedSince != "" {
		t, err := readDatetimeArg(flags.StatusUnchangedSince)
		if err != nil {
			return filter, fmt.Errorf("Error reading 'status-unchanged-since': %s", err)
		}
		filter.StatusChangedTo = &t
	}

	if flags.IncludeAllStatuses || flags.IncludeLiveStatuses {
		flags.IncludeStatusInitial = true
		flags.IncludeStatusRunning = true
		flags.IncludeStatusPaused = true
	}

	if flags.IncludeAllStatuses || flags.IncludeTerminatedStatuses {
		flags.IncludeStatusDone = true
		flags.IncludeStatusError = true
	}

	if flags.IncludeStatusInitial {
		filter.Status = append(filter.Status, api.Process_INITIAL)
	}

	if flags.IncludeStatusRunning {
		filter.Status = append(filter.Status, api.Process_RUNNING)
	}

	if flags.IncludeStatusPaused {
		filter.Status = append(filter.Status, api.Process_PAUSED)
	}

	if flags.IncludeStatusDone {
		filter.Status = append(filter.Status, api.Process_DONE)
	}

	if flags.IncludeStatusError {
		filter.Status = append(filter.Status, api.Process_ERROR)
	}

	if flags.AckedResponseOnly {
		filter.ResultAcked = append(filter.ResultAcked, true)
	}

	if flags.NotAckedResponseOnly {
		filter.ResultAcked = append(filter.ResultAcked, false)
	}

	if flags.EnvelopeIDs != "" {
		var sep string
		if strings.Contains(flags.EnvelopeIDs, " ") {
			sep = " "
		} else {
			sep = ","
		}
		for _, sid := range strings.Split(flags.EnvelopeIDs, sep) {
			id, err := api.UUIDFromString(sid)
			if err != nil {
				return api.ProcessFilter{}, err
			}
			filter.EnvelopeID = append(filter.EnvelopeID, id)
		}
	}

	if flags.Pipelines != "" {
		allPipelines, err := client.PipelineQuery("")
		if err != nil {
			return api.ProcessFilter{}, err
		}
		list := strings.Split(flags.Pipelines, ",")
		for _, item := range list {
			var (
				name  = item
				match = func(v string) bool { return true }
				found bool
			)
			for _, op := range []struct {
				op string
				fn func(v1, v2 string) bool
			}{
				{"==", func(v1, v2 string) bool {
					return v1 == v2
				}},
				{"=", func(v1, v2 string) bool {
					return v1 == v2
				}},
			} {
				if i := strings.Index(item, op.op); i != -1 {
					name = item[:i]
					match = func(v string) bool {
						return op.fn(v, item[i+len(op.op):])
					}
					break
				}
			}

			for _, pipeline := range allPipelines {
				if pipeline.GetName() == name && match(pipeline.GetVersion()) {
					found = true
					filter.PipelineID = append(filter.PipelineID, pipeline.ID)
				}
			}
			if !found {
				return filter, fmt.Errorf("No graph matching '%s'", item)
			}
		}
	}

	for _, arg := range args {
		id, err := api.UUIDFromString(arg)
		if err != nil {
			return api.ProcessFilter{}, fmt.Errorf("Invalid processid: %s (%s)", arg, err)
		}
		filter.ID = append(filter.ID, id)
	}

	return filter, nil
}

func init() {
	RootCmd.AddCommand(psCmd)

	flags := psCmd.PersistentFlags()

	// A few flags to create filters
	flags.StringVar(
		&globalFlags.CreatedFrom,
		"created-from", "",
		"Process was created at or after this date (or duration before now)",
	)
	flags.StringVar(
		&globalFlags.CreatedTo,
		"created-to", "",
		"Process was created before or at this date (or duration before now)",
	)
	flags.StringVar(
		&globalFlags.StatusChangedSince,
		"status-changed-since", "",
		"Status of the process changed at least once after this date (or duration before now)")

	flags.StringVar(
		&globalFlags.StatusUnchangedSince,
		"status-unchanged-since", "",
		"Status of the process did not change after this date (or duration before now)")

	flags.BoolVar(
		&globalFlags.IncludeStatusInitial,
		"status-initial", false,
		"Include processes in the 'initial' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusRunning,
		"status-running", false,
		"Include processes in the 'running' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusPaused,
		"status-paused", false,
		"Include processes in the 'paused' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusDone,
		"status-done", false,
		"Include processes in the 'done' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusError,
		"status-error", false,
		"Include processes in the 'error' status")

	flags.BoolVar(
		&globalFlags.IncludeLiveStatuses,
		"live", false,
		"Include processes that are not terminated")

	flags.BoolVar(
		&globalFlags.IncludeTerminatedStatuses,
		"terminated", false,
		"Include processes that are terminated")

	flags.BoolVar(
		&globalFlags.IncludeAllStatuses,
		"all", false,
		"Include processes with any status",
	)

	flags.BoolVar(
		&globalFlags.AckedResponseOnly,
		"result-acked", false,
		"Include processes for which the result is acked")

	flags.BoolVar(
		&globalFlags.NotAckedResponseOnly,
		"result-not-acked", false,
		"Include processes for which the result is not acked")

	flags.StringVar(
		&globalFlags.EnvelopeIDs,
		"envelope-id", "",
		"Include processes that are related to one on the given envelopes",
	)

	flags.StringVar(
		&globalFlags.Pipelines,
		"pipeline", "",
		"Include pipelines matching the selection (ex: name1,name2==version,name3)",
	)

	psCmd.AddCommand(psPauseCmd)
	psCmd.AddCommand(psResumeCmd)
	psCmd.AddCommand(psCancelCmd)
	psCmd.AddCommand(psExportCmd)
	psCmd.AddCommand(psPurgeCmd)

	flags = psExportCmd.PersistentFlags()
	flags.Bool("export-envelope", false, "Export the process envelopes")
	flags.Int64("export-envelope-size-limit", 0,
		"When exporting the envelopes, limit the envelope exported size")
	flags.Bool("export-jobs", false, "Export the process jobs")
	flags.Bool("export-logs", false, "Export the actor logs associated to the process")
	flags.String("export-logs-level", "notice", "Log level when exporting actor logs ('notice', 'warning', 'error')")
	flags.Bool("export-process-logs", false, "Export the process logs")

	flags.Bool("full", false, "Export all the optional data (See also the --export-xxx options)")
	flags.Bool("quiet", false, "Disable extra logging so stdout only contains the export")
}
