// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"crypto/tls"
	"crypto/x509"
	"fmt"
	"os"
	"time"

	"github.com/slicebit/qb"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/services/control"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

// accountCmd represents the account command.
var accountCmd = &cobra.Command{
	Use:   "account",
	Short: "Manage accounts that are allowed to connect to Xbus.",
	Long: `This command contains several sub-commands to manage accounts and the
permissions they have within Xbus.

In particular, it is possible from this interface to accept / reject pending
account requests.

Note that this is usually done via the "xbusctl" program, although the commands
are available here as well.`,
	PersistentPreRun: preRunCheckCertCertDir,
}

// accountListCmd represents the account list sub-command.
var accountListCmd = &cobra.Command{
	Use:   "list",
	Short: "List accounts.",
	Long:  "Display accounts allowed to connect to Xbus, as a list.",
	Run:   listAccounts,
}

// accountAcceptCmd represents the account accept sub-command.
var accountAcceptCmd = &cobra.Command{
	Use:   "accept <id|name|fingerprint>",
	Short: "Accept an account.",
	Long:  "Accept a pending account request so it can be allowed to connect to Xbus.",
	Run:   acceptAccount,
}

// accountRejectCmd represents the account reject sub-command.
var accountRejectCmd = &cobra.Command{
	Use:   "reject <id|name|fingerprint>",
	Short: "Reject an account.",
	Long:  "Reject a pending account request; it will not be allowed to connect to Xbus.",
	Run:   rejectAccount,
}

var accountResetCertCmd = &cobra.Command{
	Use:   "reset-cert <id|name>",
	Short: "Un-accept an account so it can reset its cert",
	Run:   resetCert,
}

// Arguments of the "accept" command.
const (
	acceptCmdArgKey = iota
	acceptCmdArgCount
)

// Arguments of the "reject" command.
const (
	rejectCmdArgKey = iota
	rejectCmdArgCount
)

const (
	resetCertArgKey = iota
	resetCertArgCount
)

// listAccounts parses arguments to list accounts.
func listAccounts(cmd *cobra.Command, args []string) {
	csrHash := viper.GetBool("csr-hash")
	certDetails := viper.GetBool("cert-details")
	os.Exit(doListAccounts(getAccountStorage(), csrHash, certDetails))
}

// doListAccounts lists accounts.
func doListAccounts(accountStorage storage.AccountStorage, csrHash bool, certDetails bool) int {
	accounts := accountStorage.List()

	accountCount := len(accounts)
	if accountCount == 0 {
		fmt.Println("No account found.")
	} else {
		fmt.Printf("%d accounts found:\n", accountCount)
		for _, account := range accounts {
			fmt.Println(fmtAccount(account, csrHash, certDetails))
		}
	}
	return 0
}

// acceptAccount parses arguments to accept a pending account request.
func acceptAccount(cmd *cobra.Command, args []string) {
	if len(args) < acceptCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", acceptCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	certDir := viper.GetString("cert-dir")

	expire, err := getExpire(cmd, "cert-expire")
	if err != nil {
		Logger.Errorf("%s", err)
		cmd.Usage()
		os.Exit(1)
	}
	os.Exit(doAcceptAccount(getAccountStorage(), args[acceptCmdArgKey], certDir, expire))
}

// doAcceptAccount accepts a pending account request.
func doAcceptAccount(
	accountStorage storage.AccountStorage, key string, certDir string, expire time.Time,
) int {

	// Load the servers's client CA in order to sign CSRs.
	clientCACert, clientCAKey, err := loadClientCA(certDir)
	if err != nil {
		Logger.Errorf("Cannot load the client CA: %s", err)
		return 1
	}

	account, found := findAccount(accountStorage, key)
	if !found {
		return 1
	}

	// Got a match! Ensure the account is pending.
	if account.Status != api.Account_PENDING {
		Logger.Errorf("The specified account (%s) is not pending a validation.",
			account.Name)
		return 1
	}

	// Got a match! Send an "accept" request.
	repAccount, err := control.AccountAccept(accountStorage, account.ID, clientCACert, clientCAKey, expire)
	if err != nil {
		Logger.Errorf("Could not accept the account %q: %s", account.Name, err)
		return 1
	}
	if repAccount.Status == api.Account_PENDING {
		Logger.Errorf("Could not accept the account %q: Still pending.", account.Name)
		return 1
	}
	Logger.Noticef("Successfully accepted the account %q.", repAccount.Name)
	return 0
}

// rejectAccount parses arguments to reject a pending account request.
func rejectAccount(cmd *cobra.Command, args []string) {
	if len(args) < rejectCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", rejectCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	os.Exit(doRejectAccount(getAccountStorage(), args[rejectCmdArgKey]))
}

// doRejectAccount rejects a pending account request.
func doRejectAccount(accountStorage storage.AccountStorage, key string) int {
	account, found := findAccount(accountStorage, key)
	if !found {
		return 1
	}

	// Got a match! Ensure the account is pending.
	if account.Status != api.Account_PENDING {
		Logger.Errorf("The specified account (%s) is not pending a validation.",
			account.Name)
		return 1
	}

	// Got a match! Send a "reject" request.
	repAccount, err := control.AccountReject(accountStorage, account.ID)
	if err != nil {
		Logger.Errorf("Could not reject the account %q: %s", account.Name, err)
		return 1
	}
	if repAccount.Status == api.Account_PENDING {
		Logger.Errorf("Could not reject the account %q: Still pending.", account.Name)
		return 1
	}
	Logger.Noticef("Successfully rejected the account %q.", repAccount.Name)
	return 0
}

func resetCert(cmd *cobra.Command, args []string) {
	if len(args) < resetCertArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", resetCertArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	os.Exit(doResetCert(getAccountStorage(), args[resetCertArgKey]))
}

func doResetCert(accountStorage storage.AccountStorage, identifier string) int {
	account, found := findAccount(accountStorage, identifier)
	if !found {
		return 1
	}

	if account.Status == api.Account_PENDING {
		Logger.Errorf(
			"The account is already in pending state. " +
				"It must be accepted or rejected first")
		return 1
	}

	if account.Status != api.Account_ENABLED && account.Status != api.Account_DISABLED {
		Logger.Errorf(
			"The account current state does now allow a cert reset. "+
				"Expected ENABLED or DISABLED, got %s",
			account.Status)
		return 1
	}

	account.Status = api.Account_PENDING
	account.CSR = ""

	var err error
	account, err = accountStorage.Update(account)

	if err != nil {
		Logger.Errorf("Error saving the account: %s", err)
		return 1
	}
	return 0
}

// findAccount finds a pending account with the specified fingerprint or name or id
func findAccount(
	accountStorage storage.AccountStorage, key string,
) (
	account api.Account, found bool,
) {

	var ok bool
	id, err := api.UUIDFromString(key)
	if err == nil {
		account, ok = accountStorage.Get(id)
		if !ok {
			Logger.Errorf("No account by this ID: %s", id)
			return
		}
	} else if len(key) == 59 {
		for _, a := range accountStorage.List() {
			// TODO To be done elsewhere? Should fingerprints be cached?
			if tools.GetCSRHash(a.CSR, true) == key {
				account = a
				ok = true
				break
			}
		}
		if !ok {
			Logger.Errorf("No account by this fingerprint: %s", key)
			return
		}
	} else {
		account, ok = accountStorage.GetByName(key)
		if !ok {
			Logger.Errorf("No account by this name: %s", key)
			return
		}
	}

	return account, true
}

func fmtAccount(account api.Account, csrHash bool, certDetails bool) string {
	formattedAccount := fmt.Sprintf(`
ID: %s
Name: %s
Type: %s
Status: %s`,
		account.ID,
		account.Name,
		account.Type,
		account.Status,
	)
	if csrHash {
		formattedAccount += fmt.Sprintf(`
CSR fingerprint: %s`,
			tools.GetCSRHash(account.CSR, true),
		)
	}
	if certDetails {
		formattedAccount += fmt.Sprintf(`
CSR: %s
Cert: %s`,
			account.CSR,
			account.Cert,
		)
	}
	return formattedAccount
}

// getAccountStorage provides access to the DB account storage.
func getAccountStorage() (accountStorage storage.AccountStorage) {
	m, db, err := model.NewDB("postgres",
		viper.GetString("database.dsn"),
		viper.GetInt("database.max-connections"),
		Logger, qb.LDefault)
	if err != nil {
		Logger.Errorf("Cannot connect to the database: %s", err)
		return nil
	}
	return sql.NewAccountStorage(m, db, Logger)
}

// loadClientCA loads the server CA made for client connections and provides the cert & the private
// key, so account signing functions, in particular, can use it.
// TODO This should be elsewhere.
func loadClientCA(certDir string) (
	clientCACert *x509.Certificate, clientCAKey interface{}, err error,
) {
	clientCAInfo, err := tls.LoadX509KeyPair(certDir+"/ClientCA.crt", certDir+"/ClientCA.key")
	if err != nil {
		return nil, nil, err
	}
	clientCACert, err = x509.ParseCertificate(clientCAInfo.Certificate[0])
	if err != nil {
		return nil, nil, err
	}
	clientCAKey = clientCAInfo.PrivateKey
	return clientCACert, clientCAKey, nil
}

func init() {
	RootCmd.AddCommand(accountCmd)

	accountCmd.AddCommand(accountListCmd)
	accountListFlags := accountListCmd.PersistentFlags()
	accountListFlags.Bool("cert-details", false, "Display cert details")
	viper.BindPFlag("cert-details", accountListFlags.Lookup("cert-details"))
	accountListFlags.Bool("csr-hash", true, "Display CSR fingerprints")
	viper.BindPFlag("csr-hash", accountListFlags.Lookup("csr-hash"))

	flags := accountAcceptCmd.Flags()
	flags.String("cert-expire", "1y", "Date of duration for the account certificate to expire")
	viper.BindPFlag("account.cert-expire", flags.Lookup("cert-expire"))

	accountCmd.AddCommand(accountAcceptCmd)

	accountCmd.AddCommand(accountRejectCmd)
	accountCmd.AddCommand(accountResetCertCmd)
}
