import json

import docutils
import protobufdomain
from docutils import nodes
from docutils.parsers.rst import Directive


def parse(directive, content):
    view = docutils.statemachine.StringList(content.split("\n"), "<autoprotobuf>")
    node = nodes.container()
    directive.state.nested_parse(view, directive.content_offset, node)
    return [node]


class JsonDocDirective(Directive):
    required_arguments = 1
    has_content = False

    def run(self):
        path = self.arguments[0].strip()
        data = json.load(open(path, "r", encoding="UTF-8"))
        env = self.state.document.settings.env
        env.temp_data.setdefault("proto:files", []).extend(data["files"])
        return []


def indent_lines(text, prefix):
    return "\n".join(prefix + line for line in text.split("\n"))


def fmt_rpc_desc(text, prefix):
    return "\n".join(prefix + line.lstrip("/ ") for line in text.split("\n"))


def fmt_desc(text, prefix, prefix_firstline=True):
    lines = text.split("\n")
    if len(lines) == 0:
        return ""

    if prefix_firstline:
        return "\n".join(prefix + line.lstrip("/") for line in text.split("\n"))

    lines = lines[0:1] + [prefix + line.lstrip("/") for line in text.split("\n")]

    return "\n".join(lines)


class AutoPackageDirective(Directive):
    required_arguments = 1

    def run(self):
        package = self.arguments[0].strip()

        env = self.state.document.settings.env
        files = env.temp_data.get("proto:files", [])

        filedoc = None

        if len(files) == 0:
            raise ValueError("No doc files" % package)

        for f in files:
            if f["package"] == package:
                filedoc = f

        if filedoc is None:
            raise ValueError("No doc file for package '%s'" % package)

        rst_content = fmt_desc(filedoc["description"], "") + "\n\n"

        if len(filedoc["services"]) > 0:
            rst_content += "**Services**\n\n"

        for s in filedoc["services"]:
            rst_content += "*service* **%s**\n" % s["name"]

            rst_content += indent_lines(s["description"], "    ")
            rst_content += "\n\n"

            for m in s["methods"]:
                rst_content += "    .. protobuf:rpc:: %s(%s)\n\n" % (
                    m["name"],
                    m["requestType"],
                )
                if m["requestFullType"] != "nrpc.Void":
                    rst_content += "        :arg arg: :protobuf:message:`%s`\n" % (
                        m["requestType"]
                    )
                rst_content += "        :returns: :protobuf:message:`%s`\n" % (
                    m["responseType"],
                )
                rst_content += fmt_rpc_desc(m["description"], "        ")
                rst_content += "\n"

        if len(filedoc["messages"]) > 0:
            rst_content += "**Message types**\n\n"

        for m in filedoc["messages"]:
            rst_content += ".. protobuf:message:: {package}.{message}\n\n".format(
                message=m["name"], package=package
            )

            if m["description"]:
                rst_content += fmt_desc(m["description"], "    ") + "\n\n"

            for f in m["fields"]:
                rst_content += "    :field %s: %s\n" % (
                    f["name"],
                    fmt_desc(f["description"], " " * 8, False),
                )
                rst_content += "    :type %s: %s\n" % (f["name"], f["type"])

        env.temp_data["proto:currentpackage"] = package

        return parse(self, rst_content)


protobufdomain.ProtobufDomain.directives.update(
    {"jsondoc": JsonDocDirective, "autopackage": AutoPackageDirective}
)


def setup(env):
    pass
