package fullenv

import (
	"fmt"
	"os"
	"os/exec"
	"path"
	"strings"
	"syscall"
	"time"

	"github.com/spf13/pflag"
	"github.com/spf13/viper"

	// Embed QB dialects we need
	_ "github.com/slicebit/qb/dialects/postgres"

	xbus "xbus.io/go-xbus"
	"xbus.io/go-xbus/api"
)

type clientSettings struct {
	name             string
	customXbusClient string
	register         bool
	autostart        bool
	inProcess        bool
	apikey           string
	password         string
	actorConfig      map[string]map[string]string
}

type clientRunner struct {
	dirname       string
	configFile    string
	xbusClientBin string
	logger        xbus.Logger
	settings      clientSettings
	cmd           *exec.Cmd
	inClient      *xbus.Client
	running       bool
}

func (c *clientRunner) configPath() string {
	return path.Join(c.dirname, c.configFile)
}

func (c *clientRunner) buildArgs(args ...string) (string, []string) {
	xbusclientCfgFilename := path.Join(
		c.dirname, c.configFile)

	args = append([]string{
		"--config", xbusclientCfgFilename,
	}, args...)

	return c.xbusClientBin, args
}

func (c *clientRunner) Settings() ProgramSettings {
	return ProgramSettings{
		c.settings.autostart,
		0,
	}
}

func (c *clientRunner) Running() bool {
	return c.running
}

func (c *clientRunner) Run(args ...string) error {
	program, finalArgs := c.buildArgs(args...)

	cmd := exec.Command(program, finalArgs...)
	cmd.Dir = c.dirname

	b, err := cmd.CombinedOutput()
	if err != nil {
		c.logger.Errorf("xbus-client: %s", b)
		return fmt.Errorf("Could not run xbus-client : %s", err)
	}

	return nil
}

func (c *clientRunner) Setup(xbusctl CmdRunner) error {
	if c.settings.inProcess {
		v := viper.New()
		xbus.DefineAndBindPFlags(v, &pflag.FlagSet{})
		v.SetConfigFile(c.dirname + "/" + c.configFile)
		err := v.ReadInConfig()
		if err != nil {
			return err
		}
		coptions := xbus.COptionsFromViper(
			v, api.Account_ACTOR, c.dirname, c.logger)
		inProcessClient := xbus.NewClient(coptions, c.logger)
		if err := inProcessClient.CreateKey(false); err != nil {
			return err
		}
		if err := inProcessClient.CreateCSR("localhost", "", c.settings.name, false); err != nil {
			return err
		}
		c.inClient = inProcessClient
	}
	if c.settings.register {
		if c.settings.inProcess {
			if err := c.inClient.Connect(); err != nil {
				return err
			}
			status, err := c.inClient.Register()
			if err != nil {
				return err
			}
			if status != api.Registration_PENDING {
				return fmt.Errorf(
					"Invalid return from Register: %s (should be 'pending')", status)
			}
			if err := xbusctl.Run(
				"account", "accept", c.inClient.ID().String(),
			); err != nil {
				return err
			}

			for _, actor := range c.inClient.GetOptions().ActorList {
				if err := xbusctl.Run("actor", "accept", actor.Name); err != nil {
					return err
				}
			}

			status, err = c.inClient.Register()
			if err != nil {
				return err
			}
			if status != api.Registration_ACCEPTED {
				return fmt.Errorf(
					"Invalid return from Register: %s (should be 'accepted')",
					status)
			}
			// Reconnect so the new permissions are applied
			c.inClient.Close()
			if err := c.inClient.Connect(); err != nil {
				return err
			}
		} else {
			if err := c.Run("register"); err != nil {
				return err
			}

			// Load the conf to read the ids
			v := viper.New()
			xbus.DefineAndBindPFlags(v, &pflag.FlagSet{})
			v.SetConfigFile(c.configPath())
			if err := v.ReadInConfig(); err != nil {
				return err
			}
			coptions := xbus.COptionsFromViper(
				v, api.Account_ACTOR, c.dirname, c.logger)

			// Accept the account and actor registration
			c.logger.Noticef("Accepting account %s", coptions.Name)
			if err := xbusctl.Run(
				"account", "accept", coptions.Name,
			); err != nil {
				return err
			}


			for _, actor := range coptions.ActorList {
				c.logger.Noticef("Accepting %s", actor.Name)
				if err := xbusctl.Run(
					"actor", "accept", actor.Name,
				); err != nil {
					return err
				}
			}

			if err := c.Run("register"); err != nil {
				return err
			}
		}
		if c.settings.apikey != "" {
			xbusctl.Run("account", "set-apikey", c.settings.name, c.settings.apikey)
		}
		for name, entries := range c.settings.actorConfig {
			for key, value := range entries {
				if err := xbusctl.Run(
					"actor", "config", "set", name, key+"="+value,
				); err != nil {
					return err
				}

			}
		}
	}
	return nil
}

func (c *clientRunner) Start() error {
	if c.running {
		return fmt.Errorf("Client '%s' already running", c.settings.name)
	}
	if c.settings.inProcess {
		if err := c.inClient.Startup(true); err != nil {
			return err
		}
	} else {
		program, args := c.buildArgs("serve")
		if c.settings.customXbusClient != "" {
			splitted := strings.Split(c.settings.customXbusClient, " ")
			program = splitted[0]
			args = append(splitted[1:], args[:len(args)-1]...)
		}
		cmd := exec.Command(program, args...)
		cmd.Dir = c.dirname

		if err := captureCmdOutputs("client-"+c.settings.name, cmd, c.logger); err != nil {
			return err
		}

		err := cmd.Start()
		if err != nil {
			return err
		}
		c.cmd = cmd
	}
	c.running = true
	return nil
}

func (c *clientRunner) Stop() error {
	if !c.running {
		return fmt.Errorf("Client '%s' is not running", c.settings.name)
	}
	if c.settings.inProcess {
		c.inClient.Shutdown()
		c.running = false
	} else {
		c.cmd.Process.Signal(os.Interrupt)
		var timer *time.Timer
		timer = time.AfterFunc(time.Second, func() {
			c.logger.Errorf(
				"Client '%s' still running. Attempting a SIGQUIT",
				c.settings.name)
			if err := c.cmd.Process.Signal(syscall.SIGQUIT); err != nil {
				c.logger.Errorf(
					"SIGQUIT failed with '%s', killing the process.", err)
				c.cmd.Process.Kill()
			} else {
				timer = time.AfterFunc(time.Second, func() {
					c.logger.Errorf(
						"Client '%s' still running. Killing it",
						c.settings.name)
					c.cmd.Process.Kill()
				})
			}
		})
		defer timer.Stop()
		return c.cmd.Wait()
	}

	return nil
}
