package fullenv

import (
	"fmt"
	"os"
	"os/exec"
	"path"
	"syscall"
	"time"

	// Embed QB dialects we need
	_ "github.com/slicebit/qb/dialects/postgres"

	"xbus.io/go-xbus"
)

type httpSettings struct {
	name      string
	autostart bool
}

type httpRunner struct {
	dirname     string
	configFile  string
	xbusHTTPBin string
	logger      xbus.Logger
	settings    httpSettings
	cmd         *exec.Cmd
	running     bool
}

func (c *httpRunner) configPath() string {
	return path.Join(c.dirname, c.configFile)
}

func (c *httpRunner) buildArgs(args ...string) (string, []string) {
	xbushttpCfgFilename := path.Join(
		c.dirname, c.configFile)

	args = append([]string{
		"--config", xbushttpCfgFilename,
	}, args...)

	return c.xbusHTTPBin, args
}

func (c *httpRunner) Settings() ProgramSettings {
	return ProgramSettings{
		c.settings.autostart,
		0,
	}
}

func (c *httpRunner) Running() bool {
	return c.running
}

func (c *httpRunner) Run(args ...string) error {
	program, finalArgs := c.buildArgs(args...)

	cmd := exec.Command(program, finalArgs...)
	cmd.Dir = c.dirname

	b, err := cmd.CombinedOutput()
	if err != nil {
		c.logger.Errorf("xbus-http: %s", b)
		return fmt.Errorf("Could not run xbus-http : %s", err)
	}

	return nil
}

func (c *httpRunner) Setup(xbusctl CmdRunner) error {
	if err := c.Run("register"); err != nil {
		return err
	}

	// Accept the account and actor registration
	c.logger.Noticef("Accepting account %s", c.settings.name)
	if err := xbusctl.Run(
		"account", "accept", c.settings.name,
	); err != nil {
		return err
	}

	if err := c.Run("register"); err != nil {
		return err
	}
	return nil
}

func (c *httpRunner) Start() error {
	if c.running {
		return fmt.Errorf("XbusHTTP '%s' already running", c.settings.name)
	}
	program, args := c.buildArgs("serve")
	cmd := exec.Command(program, args...)
	cmd.Dir = c.dirname

	if err := captureCmdOutputs("http-"+c.settings.name, cmd, c.logger); err != nil {
		return err
	}

	err := cmd.Start()
	if err != nil {
		return err
	}
	c.cmd = cmd
	c.running = true
	return nil
}

func (c *httpRunner) Stop() error {
	if !c.running {
		return fmt.Errorf("HTTP '%s' is not running", c.settings.name)
	}
	c.cmd.Process.Signal(os.Interrupt)
	var timer *time.Timer
	timer = time.AfterFunc(time.Second, func() {
		c.logger.Errorf(
			"HTTP '%s' still running. Attempting a SIGQUIT",
			c.settings.name)
		if err := c.cmd.Process.Signal(syscall.SIGQUIT); err != nil {
			c.logger.Errorf(
				"SIGQUIT failed with '%s', killing the process.", err)
			c.cmd.Process.Kill()
		} else {
			timer = time.AfterFunc(time.Second, func() {
				c.logger.Errorf(
					"HTTP '%s' still running. Killing it",
					c.settings.name)
				c.cmd.Process.Kill()
			})
		}
	})
	defer timer.Stop()
	return c.cmd.Wait()
}
