package fullstack

import (
	"context"
	"fmt"
	"testing"
	"time"

	xbus "xbus.io/go-xbus"
	"xbus.io/go-xbus/api"
	"github.com/stretchr/testify/assert"
)

func TestDirectorInvalidLogCrash(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	xsub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			return fmt.Errorf("test\000")
		})
	assert.Nil(t, err)
	defer xsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})

	envStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	status, err := emitter.SendEnvelope(nil, api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	timeout := time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("Process state not updated")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_RUNNING, envState.Status)
	}

	timeout = time.After(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("Process state not updated to 'error'")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_ERROR.String(), envState.Status.String())
	}
}

func TestGetRunningActorRequestList(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")

	var consumerAPC *xbus.ActorProcessingContext
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	xsub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			consumerAPC = apc
			t.Logf("Here i am")
			e, err := apc.CompleteEnvelope("default", time.Second)
			if err != nil {
				ft.T.Error(err)
				return err
			}
			hit <- e
			apc.Detach()
			return nil
		})
	assert.Nil(t, err)
	defer xsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))

	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})

	envStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	status, err := emitter.SendEnvelope(nil, api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

	select {
	case <-time.After(5 * time.Second):
		t.Errorf("Process state not updated")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_RUNNING, envState.Status)
	}

	select {
	case <-time.After(5 * time.Second):
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}

	// Let time to the director to update the job status.
	time.Sleep(250 * time.Millisecond)

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	var runningTasks []api.ActorProcessRequest

	if err := consumer.Client.API.Director.GetRunningActorRequestList(
		ctx, consumer.ID.String(),
		func(ctx context.Context, req api.ActorProcessRequest) {
			runningTasks = append(runningTasks, req)
		},
	); err != nil {
		t.Fatal(err)
	}

	assert.Equal(t, 1, len(runningTasks))

	consumerAPC.Done()

	select {
	case <-time.After(5 * time.Second):
		t.Errorf("Process state not updated to 'done'")
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_DONE, envState.Status)
	}
}
