package http

import (
	"encoding/base64"
	"encoding/json"
	"errors"
	"fmt"
	"io/ioutil"
	"net/http"
	"strings"
	"time"

	"github.com/gogo/protobuf/jsonpb"
	"github.com/gogo/protobuf/proto"
	"github.com/gorilla/mux"
	nats "github.com/nats-io/go-nats"

	xbus "xbus.io/go-xbus"
	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/envelope"
)

// ErrNoAPIKey is when no API key is in the request
var ErrNoAPIKey = errors.New("No API key found in the request")

// ErrNotNow ...
var ErrNotNow = errors.New("The actor is not ready, try again later")

// ErrTemporaryError ...
var ErrTemporaryError = errors.New(
	"An error occured that should not put the process in fail state")

// NewGateway initializes a Gateway
func NewGateway(options xbus.COptions, logger xbus.Logger) *Gateway {
	router := mux.NewRouter()
	g := &Gateway{
		XbusOptions: options,
		Log:         logger,
		Router:      router,
	}
	router.HandleFunc("/{actor}/request", g.RequestHandler)
	router.HandleFunc("/{actor}/output", g.OutputHandler)
	router.HandleFunc("/{actor}/processrequest", g.ProcessRequestHandler)
	router.HandleFunc("/{actor}/processingend", g.ProcessingEndHandler)

	router.HandleFunc("/{actor}/output/{output}", g.Beta11OutputHandler)
	router.HandleFunc("/{actor}/input", g.Beta11InputHandler)
	return g
}

// Gateway provides a http router that handle a http-rest gateway to xbus
type Gateway struct {
	XbusOptions xbus.COptions
	Log         xbus.Logger
	Router      *mux.Router

	webhooks []*WebHookHandler
	clients  []*xbus.Client
}

// Startup starts handling the webhooks
func (g *Gateway) Startup() error {
	options := g.XbusOptions
	options.ConnOptions.Token = "gateway"

	client := xbus.NewClient(
		options, xbus.NewPrefixedLogger("gateway ", g.Log))

	if err := client.Connect(); err != nil {
		return err
	}
	defer client.Close()

	actors, err := client.ActorList(nil)
	if err != nil {
		return nil
	}

	actorsByAccount := make(map[api.UUID][]xbus.ActorOptions)

	for _, actorDef := range actors {
		if url := actorDef.GetConfigEntry("http.webhook.url"); url != "" {
			options := xbus.ActorOptions{Actor: actorDef}
			options.Service.Type = "http.webhook"

			actorsByAccount[actorDef.AccountID] = append(
				actorsByAccount[actorDef.AccountID], options)

		}
	}

	// Now connects to each client and create the webhooks
	for accountid, actors := range actorsByAccount {
		clientOptions := g.XbusOptions
		clientOptions.ConnOptions.Token = "webhook:" + accountid.String()
		clientOptions.ActorList = actors

		client := xbus.NewClient(
			clientOptions, xbus.NewPrefixedLogger(accountid.String()+" ", g.Log))

		if err := client.Startup(true); err != nil {
			g.Shutdown()
			return err
		}
		g.clients = append(g.clients, client)
	}
	return nil
}

// Shutdown the webhooks
func (g *Gateway) Shutdown() {
	for _, client := range g.clients {
		client.Shutdown()
	}
}

func (g *Gateway) returnError(w http.ResponseWriter, status int, err string) {
	b, encErr := json.Marshal(map[string]string{
		"error": err,
	})
	if encErr != nil {
		w.WriteHeader(http.StatusInternalServerError)
		msg := fmt.Sprintf("Error encoding error reply: %s", err)
		g.Log.Errorf(msg)
		w.Write([]byte(msg))
	} else {
		w.WriteHeader(status)
		w.Write(b)
	}
}

// Credentials holds user credentials extracted from the request
type Credentials struct {
	APIKey   string
	User     string
	Password string
}

// ExtractCredentials extracts authentication information from a http.Request
func ExtractCredentials(r *http.Request) (*Credentials, error) {
	var creds Credentials
	creds.APIKey = r.Header.Get("Xbus-API-Key")

	if v := r.URL.Query().Get("key"); v != "" {
		creds.APIKey = v
	}

	// protect against unwanted injections
	if strings.Contains(creds.APIKey, ":") {
		return nil, fmt.Errorf("Invalid APIKey: %s", creds.APIKey)
	}

	var (
		auth      = r.Header.Get("Authorization")
		authType  string
		authValue string
	)

	if creds.APIKey != "" && auth != "" {
		return nil, fmt.Errorf("Invalid headers: too many credentials")
	}

	if auth != "" {
		splitted := strings.SplitN(auth, " ", 2)
		if len(splitted) != 2 {
			return nil, fmt.Errorf("Invalid Authorization header")
		}
		authType = splitted[0]
		authValue = splitted[1]
	}

	if authType == "Basic" {
		data, err := base64.StdEncoding.DecodeString(authValue)
		if err != nil {
			fmt.Println("error:", err)
			return nil, fmt.Errorf("decoding Authorization: %s", err)
		}
		splitted := strings.SplitN(string(data), ":", 2)
		if len(splitted) != 2 {
			return nil, fmt.Errorf("decoding Authorization: invalid user:password")
		}
		creds.User = splitted[0]
		creds.Password = splitted[1]
	} else if authType != "" {
		return nil, fmt.Errorf("Invalid Authorization type: %s", authType)
	}

	return &creds, nil
}

// SetupCredentials extract credentials from a request and inject them in a
// nats.Options
func SetupCredentials(options nats.Options, r *http.Request) (nats.Options, error) {
	creds, err := ExtractCredentials(r)
	if err != nil {
		return options, err
	}
	options.Token = creds.APIKey
	options.User = creds.User
	options.Password = creds.Password
	return options, nil
}

func (g *Gateway) newClient(r *http.Request) (*xbus.Client, error) {
	apikey := r.Header.Get("Xbus-API-Key")

	if v := r.URL.Query().Get("key"); v != "" {
		apikey = v
	}

	if apikey == "" {
		return nil, ErrNoAPIKey
	}

	coptions := g.XbusOptions
	options, err := SetupCredentials(g.XbusOptions.ConnOptions, r)
	if err != nil {
		return nil, err
	}
	coptions.ConnOptions = options

	client := xbus.NewClient(coptions, xbus.NewPrefixedLogger(apikey+" ", g.Log))

	if err := client.Connect(); err != nil {
		return nil, err
	}

	return client, nil
}

func (g *Gateway) getClient(r *http.Request) (*xbus.Client, error) {
	return g.newClient(r)
}

func (g *Gateway) getActor(r *http.Request) (*xbus.Actor, error) {
	vars := mux.Vars(r)
	actorIdentifier := vars["actor"]

	if actorIdentifier == "" {
		panic("No actor in request vars")
	}

	client, err := g.getClient(r)
	if err != nil {
		return nil, err
	}

	var actor *xbus.Actor
	if id, err := api.UUIDFromString(actorIdentifier); err == nil {
		actor = client.GetActor(id)
	} else {
		actor = client.GetActorByName(actorIdentifier)
	}
	if actor == nil {
		return nil, fmt.Errorf("No such actor: %s", actorIdentifier)
	}
	return actor, nil
}

func wrapEnvelope(e *api.Envelope) envelope.Envelope {
	wrapped := envelope.NewOpenedEnvelopeWithID(e.ID)

	go func() {
		for _, event := range e.Events {
			msg, writer := envelope.NewMessageChunkWriterWithID(
				event.ID, event.Type, event.Checksum)
			wrapped.Add(msg)
			for _, item := range event.Items {
				if err := writer.Write(item); err != nil {
					writer.Close()
					wrapped.CloseWithError(err)
					return
				}
			}
			if err := writer.Close(); err != nil {
				wrapped.CloseWithError(err)
				return
			}
		}
		if err := wrapped.Close(); err != nil {
			wrapped.CloseWithError(err)
			return
		}
	}()

	return wrapped
}

// OutputHandler is the handler for the '{actor}/output' route
func (g *Gateway) OutputHandler(w http.ResponseWriter, r *http.Request) {
	actor, err := g.getActor(r)
	if err != nil {
		w.WriteHeader(http.StatusNotFound)
		fmt.Fprintf(w, "Error: %s", err)
		return
	}
	defer actor.GetClient().Close()

	var request api.OutputRequest

	// TODO for big envelopes, a streamed decoding would be better, but would
	// be restricted by the fact that it would have to wait for the last event
	// to be decoded before sending anything (not true anymore with the new
	// go-xbus envelope api.
	// The simple solution to that will be, for big envelopes, to add item-less
	// events with just their types in from of the Events list
	contentType := r.Header.Get("Content-Type")
	if contentType == "application/x-protobuf" {
		buffer, err := ioutil.ReadAll(r.Body)
		if err != nil {
			w.WriteHeader(http.StatusBadRequest)
			fmt.Fprintf(w, "Error reading body: %s", err)
			return
		}
		if err := proto.Unmarshal(buffer, &request); err != nil {
			w.WriteHeader(http.StatusBadRequest)
			fmt.Fprintf(w, "Error decoding body: %s", err)
			return
		}
	} else {
		decoder := json.NewDecoder(r.Body)
		if err := decoder.Decode(&request); err != nil {
			w.WriteHeader(http.StatusBadRequest)
			fmt.Fprintf(w, "Error decoding body: %s", err)
			return
		}
	}

	if request.Output == "" {
		request.Output = "default"
	}

	if request.Envelope.ID.IsZero() {
		id, err := api.UUIDV4()
		if err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			fmt.Fprintf(w, "Error generating the envelope ID: %s", err)
			return
		}
		request.Envelope.ID = id
	}

	out := wrapEnvelope(request.Envelope)

	if err := xbus.SendEnvelope(
		r.Context(), out,
		api.EnvelopeMaxSize/3, xbus.EnvelopeSenderDefaultDelay, xbus.DefaultMaxChunksByFragment,
		func(fragment *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
			return actor.SendEnvelope(request.Context, request.Output, *fragment)
		},
	); err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		fmt.Fprintf(w, "Internal Error '%s'", err)
		return
	}

	w.WriteHeader(http.StatusOK)
	w.Write([]byte("OK"))
}

type incomingFragment struct {
	input    string
	reply    string
	envelope api.Envelope
}

// ProcessRequestHandler is the handler for the '{actor}/processrequest' route
func (g *Gateway) ProcessRequestHandler(w http.ResponseWriter, r *http.Request) {
	actor, err := g.getActor(r)
	if err != nil {
		w.WriteHeader(http.StatusNotFound)
		fmt.Fprintf(w, "{\"error\": \"%s\"}", err)
		return
	}
	defer func() {
		// Without a delay, the last messages sent to the server are lost
		// FIXME this should not be necessary. Dig into nats to understand what
		// is going on. Before that, upgrade go-nats.
		time.AfterFunc(time.Second, actor.GetClient().Close)
	}()

	contentType := "application/json"

	accept := r.Header.Get("Accept")
	if strings.Contains(accept, "application/x-protobuf") {
		contentType = "application/x-protobuf"
	}

	var itemEncodings []api.ItemEncoding

	if contentType == "application/json" {
		itemEncodings, err := api.ParseItemEncodingList(
			r.Header.Get("Accept-Item-Encoding"))
		if err != nil {
			w.WriteHeader(http.StatusBadRequest)
			fmt.Fprintf(w, "Invalid 'Accept-Item-Encoding': %s", err)
			return
		}
		if len(itemEncodings) == 0 {
			itemEncodings = []api.ItemEncoding{api.ItemBinary}
		}
	}

	var (
		sub        *xbus.Subscription
		processing = make(chan struct{})
	)

	sub, err = actor.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			// Signal the http handler that we are processing something
			close(processing)

			apc.ActorLeaving()

			var (
				b   []byte
				err error
			)

			w.Header().Set("Content-Type", contentType)

			if contentType == "application/x-protobuf" {
				b, err = proto.Marshal(apc.Request)
			} else {
				b, err = api.JSONMarshal(itemEncodings, apc.Request)
			}

			if err != nil {
				w.WriteHeader(http.StatusInternalServerError)
				return err
			}

			w.WriteHeader(http.StatusOK)

			w.Write(b)

			// TODO If needed, send all the envelope remainings

			apc.Detach()
			return nil
		},
	)

	select {
	case <-r.Context().Done():
		if err := sub.Unsubscribe(); err != nil {
			g.Log.Errorf("Error unsubscribing the handler: %s", err)
		}
		return
	case <-processing:
		// The processing is ongoing, the sub.Unsubscribe will wait for its
		// completion
		if err := sub.Unsubscribe(); err != nil {
			g.Log.Errorf("Error unsubscribing the handler: %s", err)
		}

	case <-time.After(30 * time.Second):
		// Unsubscribe, then check if a processing was done meanwhile
		if err := sub.Unsubscribe(); err != nil {
			g.Log.Errorf("Error unsubscribing the handler: %s", err)
		}

		select {
		case <-processing:
			// Yes, so we have nothing to write back
		default:
			w.WriteHeader(http.StatusNoContent)
		}
	}
}

type beta11CompatProcessingState struct {
	api.ActorProcessingState
	Result string `json:"result"`
}

//ProcessingEndHandler is the handler for the '{actor}/processingend' route
func (g *Gateway) ProcessingEndHandler(w http.ResponseWriter, r *http.Request) {
	actor, err := g.getActor(r)
	if err != nil {
		w.WriteHeader(http.StatusNotFound)
		fmt.Fprintf(w, "Error: %s", err)
		return
	}
	var message beta11CompatProcessingState
	dec := json.NewDecoder(r.Body)
	if err := dec.Decode(&message); err != nil {
		g.returnError(w, http.StatusBadRequest, fmt.Sprintf("Could not decode input payload: %s", err))
		return
	}
	if message.Result == "success" {
		message.Status = api.ActorProcessingState_SUCCESS
	} else if message.Result == "error" {
		message.Status = api.ActorProcessingState_ERROR
	}
	if err := actor.GetClient().RawProcessingEnd(
		actor.ID,
		message.Context,
		message.Status,
		message.Messages,
	); err != nil {
		g.returnError(w, http.StatusInternalServerError, err.Error())
		return
	}
}

// RequestHandler is the handler for the '{actor}/request' route
func (g *Gateway) RequestHandler(w http.ResponseWriter, r *http.Request) {
	actor, err := g.getActor(r)
	if err != nil {
		w.WriteHeader(http.StatusNotFound)
		fmt.Fprintf(w, "Error: %s", err)
		return
	}
	defer actor.GetClient().Close()

	var request api.OutputRequest

	// TODO for big envelopes, a streamed decoding would be better, but would
	// be restricted by the fact that it would have to wait for the last event
	// to be decoded before sending anything.
	// The simple solution to that will be, for big envelopes, to add item-less
	// events with just their types in from of the Events list
	contentType := r.Header.Get("Content-Type")
	if contentType == "application/x-protobuf" {
		buffer, err := ioutil.ReadAll(r.Body)
		if err != nil {
			w.WriteHeader(http.StatusBadRequest)
			fmt.Fprintf(w, "Error reading body: %s", err)
			return
		}
		if err := proto.Unmarshal(buffer, &request); err != nil {
			w.WriteHeader(http.StatusBadRequest)
			fmt.Fprintf(w, "Error decoding body: %s", err)
			return
		}
	} else {
		if err := jsonpb.Unmarshal(r.Body, &request); err != nil {
			w.WriteHeader(http.StatusBadRequest)
			fmt.Fprintf(w, "Error decoding body: %s", err)
			return
		}
	}

	var itemEncodings []api.ItemEncoding

	if contentType == "application/json" {
		itemEncodings, err := api.ParseItemEncodingList(
			r.Header.Get("Accept-Item-Encoding"))
		if err != nil {
			w.WriteHeader(http.StatusBadRequest)
			fmt.Fprintf(w, "Invalid 'Accept-Item-Encoding': %s", err)
			return
		}
		if len(itemEncodings) == 0 {
			itemEncodings = []api.ItemEncoding{api.ItemBinary}
		}
	}

	if request.Output != "" {
		w.WriteHeader(http.StatusBadRequest)
		fmt.Fprintf(w, "'output' field should not be defined")
		return
	}
	request.Output = "request"

	if request.Envelope.ID.IsZero() {
		id, err := api.UUIDV4()
		if err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			fmt.Fprintf(w, "Error generating the envelope ID: %s", err)
			return
		}
		request.Envelope.ID = id
	}

	out := wrapEnvelope(request.Envelope)

	resp, err := actor.Request(r.Context(), out)

	if err != nil {
		// XXX Reporting errors should not systematicaly put the blame on the client.
		w.WriteHeader(http.StatusBadRequest)
		fmt.Fprintf(w, err.Error())
		return
	}

	fullResp, err := envelope.CompleteWithContext(r.Context(), resp)
	if err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		fmt.Fprintf(w, err.Error())
		return
	}

	w.Header().Set("Content-Type", contentType)

	var b []byte
	if contentType == "application/x-protobuf" {
		b, err = proto.Marshal(fullResp)
	} else {
		b, err = api.JSONMarshal(itemEncodings, fullResp)
	}

	if err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		fmt.Fprintf(w, "Error marshaling the response: %s", err)
	}

	w.WriteHeader(http.StatusOK)
	w.Write(b)
}
