package http

import (
	"encoding/json"
	"fmt"
	"net/http"
	"time"

	"github.com/gorilla/mux"

	xbus "xbus.io/go-xbus"
	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/envelope"
)

// Beta11Envelope is a envelope structure from the beta11 version
type Beta11Envelope struct {
	api.Envelope
	Context *api.ProcessingContext `json:"context"`
}

// Beta11OutputHandler is the handler for the deprecated '{actor}/output/{output}' route
func (g *Gateway) Beta11OutputHandler(w http.ResponseWriter, r *http.Request) {
	vars := mux.Vars(r)
	outputName := vars["output"]

	actor, err := g.getActor(r)
	if err != nil {
		w.WriteHeader(http.StatusNotFound)
		fmt.Fprintf(w, "Error: %s", err)
		return
	}
	defer actor.GetClient().Close()

	var inputEnvelope Beta11Envelope

	contentType := r.Header.Get("Content-Type")

	if contentType == "application/json" {
		decoder := json.NewDecoder(r.Body)
		if err := decoder.Decode(&inputEnvelope); err != nil {
			w.WriteHeader(http.StatusBadRequest)
			fmt.Fprintf(w, "Error decoding body: %s", err)
			return
		}
	} else {
		w.WriteHeader(http.StatusBadRequest)
		w.Write([]byte("Invalid content-type"))
	}

	out := envelope.NewOpenedEnvelopeWithID(inputEnvelope.ID)

	var errC = make(chan error)

	go func(out envelope.Envelope, errC chan<- error) {
		errC <- xbus.SendEnvelope(
			r.Context(), out,
			api.EnvelopeMaxSize/3, xbus.EnvelopeSenderDefaultDelay, xbus.DefaultMaxChunksByFragment,
			func(fragment *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
				return actor.SendEnvelope(inputEnvelope.Context, outputName, *fragment)
			},
		)
	}(out, errC)

	for _, event := range inputEnvelope.Events {
		msg, writer := envelope.NewMessageChunkWriterWithID(event.ID, event.Type, event.Checksum)
		out.Add(msg)
		for _, item := range event.Items {
			if err := writer.Write(item); err != nil {
				writer.Close()
				w.WriteHeader(http.StatusInternalServerError)
				fmt.Fprintf(w, "Internal Error '%s'", err)
				go func() { <-errC }()
				return
			}
		}
		if err := writer.Close(); err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			fmt.Fprintf(w, "Internal Error '%s'", err)
			go func() { <-errC }()
			return
		}
	}
	if err := out.Close(); err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		fmt.Fprintf(w, "Internal Error '%s'", err)
		go func() { <-errC }()
		return
	}

	if err := <-errC; err != nil {
		w.WriteHeader(http.StatusInternalServerError)
		fmt.Fprintf(w, "Internal Error '%s'", err)
		return
	}
	w.WriteHeader(http.StatusOK)
	w.Write([]byte("ACK"))
}

// Beta11InputHandler is the handler for the deprecated '{actor}/input' route
func (g *Gateway) Beta11InputHandler(w http.ResponseWriter, r *http.Request) {
	actor, err := g.getActor(r)
	if err != nil {
		w.WriteHeader(http.StatusNotFound)
		fmt.Fprintf(w, "{\"error\": \"%s\"}", err)
		return
	}
	defer func() {
		// Without a delay, the last messages sent to the server are lost
		// FIXME this should not be necessary. Dig into nats to understand what
		// is going on. Before that, upgrade go-nats.
		time.AfterFunc(time.Second, actor.GetClient().Close)
	}()

	itemEncodings, err := api.ParseItemEncodingList(r.Header.Get("Accept-Item-Encoding"))
	if err != nil {
		w.WriteHeader(http.StatusBadRequest)
		fmt.Fprintf(w, "Invalid 'Accept-Item-Encoding': %s", err)
		return
	}
	if len(itemEncodings) == 0 {
		itemEncodings = []api.ItemEncoding{api.ItemBinary}
	}

	var (
		sub        *xbus.Subscription
		processing = make(chan struct{})
	)
	sub, err = actor.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			// Signal the http handler that we are processing something
			close(processing)

			apc.ActorLeaving()

			w.Header().Set("Content-Type", "application/json")

			b, err := api.JSONMarshal(
				itemEncodings, apc.Request.Inputs[0].Envelope,
			)

			if err != nil {
				w.WriteHeader(http.StatusInternalServerError)
				return err
			}

			m := make(map[string]interface{})

			if err := json.Unmarshal(b, &m); err != nil {
				w.WriteHeader(http.StatusInternalServerError)
				return err
			}

			m["context"] = apc.Request.Context

			m = map[string]interface{}{
				"input":    apc.Request.Inputs[0].Name,
				"envelope": m,
			}

			b, err = json.Marshal(m)

			if err != nil {
				w.WriteHeader(http.StatusInternalServerError)
				return err
			}

			w.WriteHeader(http.StatusOK)
			w.Write(b)

			// TODO If needed, send all the envelope remainings

			apc.Detach()
			return nil
		},
	)

	select {
	case <-r.Context().Done():
		if err := sub.Unsubscribe(); err != nil {
			g.Log.Errorf("Error unsubscribing the handler: %s", err)
		}
		return
	case <-processing:
		// The processing is ongoing, the sub.Unsubscribe will wait for its
		// completion
		if err := sub.Unsubscribe(); err != nil {
			g.Log.Errorf("Error unsubscribing the handler: %s", err)
		}

	case <-time.After(30 * time.Second):
		// Unsubscribe, then check if a processing was done meanwhile
		if err := sub.Unsubscribe(); err != nil {
			g.Log.Errorf("Error unsubscribing the handler: %s", err)
		}

		select {
		case <-processing:
			// Yes, so we have nothing to write back
		default:
			w.WriteHeader(http.StatusNoContent)
		}
	}
}
