package http

import (
	"bytes"
	"encoding/json"
	"io/ioutil"
	"net/http"
	"net/http/httptest"
	"strconv"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"

	"xbus.io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/fullenv_tests"
)

type Beta11TestEnvelopeContext struct {
	ProcessID api.UUID
	NodeID    string
}

type Beta11TestEnvelope struct {
	api.Envelope
	Context Beta11TestEnvelopeContext
}

func TestBeta11Emitter(t *testing.T) {
	env, gateway := setupEnv(t, "emitter-test.yaml", true)
	defer env.Close()

	apiKey, err := env.XbusCtl.RunReadLastLine("account", "renew-apikey", "outerclient")
	must(t, err)

	buf := bytes.NewBufferString(`{
		"event_ids": ["53131466-4e2e-11e7-916e-af2b1e6d242d"],
		"events": [{
			"id": "53131466-4e2e-11e7-916e-af2b1e6d242d",
			"type": "test",
			"ItemCount": 1,
			"Items": ["QSBtZXNzYWdl"]
		}]
	}`)

	req := httptest.NewRequest("POST", "/emitter/output/default", buf)
	req.Header.Set("Content-Type", "application/json")
	req.Header.Set("Xbus-API-Key", apiKey)

	w := httptest.NewRecorder()

	gateway.Router.ServeHTTP(w, req)

	if w.Result().StatusCode != 200 {
		var body = make([]byte, 1024*1024)
		t.Errorf("Invalid status: %s", w.Result().Status)
		c, err := w.Result().Body.Read(body)
		if err != nil {
			t.Fatalf("Could not read body: %s", err)
		}
		t.Fatalf("Body is: %s", string(body[:c]))
	}

	if string(w.Body.Bytes()) != "ACK" {
		t.Errorf("Unexpected response: %s", w.Body.Bytes())
		return
	}

	time.Sleep(time.Second)

	assert.Equal(t, 1, len(env.Consumer.Envelopes))

	e := env.Consumer.Envelopes[0]
	assert.Equal(t, "A message", string(e.Events[0].Items[0]))
	assert.Equal(t, "53131466-4e2e-11e7-916e-af2b1e6d242d", e.Events[0].ID.String())
}

type beta11ReplyType struct {
	Input    string
	Envelope Beta11TestEnvelope
}

func TestBeta11Consumer(t *testing.T) {
	env, gateway := setupEnv(t, "consumer-test.yaml", true)
	defer env.Close()

	apiKey, err := env.XbusCtl.RunReadLastLine("account", "renew-apikey", "outerclient")
	must(t, err)

	req := httptest.NewRequest("GET", "/test-consumer/input", nil)
	req.Header.Set("Xbus-API-Key", apiKey)
	req.Header.Set("Accept-Item-Encoding", "json,text,binary")

	w := httptest.NewRecorder()

	sender := env.Emitter.Actor.OpenOutput(nil, "default", "test")
	must(t, sender.AddItems("test", api.Item("a item")))
	must(t, sender.Close())

	envStateChan, sub, err := env.Emitter.Actor.GetClient().
		API.ProcessState.EnvelopeStatesSubscribeChan(
		env.Emitter.Actor.ID.String(),
		sender.GetID().String())
	if err != nil {
		t.Fatal(err)
	}
	defer sub.Unsubscribe()

	done := make(chan bool)

	go func() {
		gateway.Router.ServeHTTP(w, req)
		done <- true
	}()

	select {
	case <-time.After(10 * time.Second):
		t.Fatal("Timeout...")
	case <-done:
		t.Log("Request served, checking result")
	}

	if w.Result().StatusCode != 200 {
		var body = make([]byte, 1024*1024)
		t.Errorf("Invalid status: %s", w.Result().Status)
		c, err := w.Result().Body.Read(body)
		if err != nil {
			t.Fatalf("Could not read body: %s", err)
		}
		t.Fatalf("Body is: %s", string(body[:c]))
	}

	var reply beta11ReplyType
	dec := json.NewDecoder(w.Body)
	if err := dec.Decode(&reply); err != nil {
		t.Fatalf("Error decoding response: %s", err)
	}
	envelope := reply.Envelope
	if len(envelope.Events) != 1 {
		t.Fatalf("Expected 1 event, got %d", len(envelope.Events))
	}
	event := envelope.Events[0]
	if len(event.Items) != 1 {
		t.Fatalf("Expected 1 item, got %d", len(event.Items))
	}
	item := event.Items[0]
	assert.Equal(t, "a item", string(item))

	// TODO decode the envelope and check its items

	{
		buf := bytes.NewBuffer(nil)
		enc := json.NewEncoder(buf)
		if err := enc.Encode(map[string]interface{}{
			"context":     envelope.Context,
			"envelope_id": envelope.ID,
			"result":      "success",
		}); err != nil {
			t.Fatal(err)
		}

		req := httptest.NewRequest("POST", "/test-consumer/processingend", buf)
		req.Header.Set("Xbus-API-Key", apiKey)

		w := httptest.NewRecorder()

		done := make(chan bool)

		go func() {
			gateway.Router.ServeHTTP(w, req)
			done <- true
		}()

		select {
		case <-time.After(10 * time.Second):
			t.Fatal("Timeout...")
		case <-done:
			t.Log("Request served, checking result")
		}

		if w.Result().StatusCode != 200 {
			t.Fatalf("Unexpected status code: %d\nBody: %s", w.Result().StatusCode, w.Body.String())
		}

	}
	timeout := time.After(5 * time.Second)
	for {
		select {
		case <-timeout:
			t.Errorf("Process state not updated to 'done'")
			return
		case envState := <-envStateChan:
			if envState.Status == api.Process_DONE {
				return
			}
		}
	}
}

func TestFloodBeta11(t *testing.T) {
	env, _ := fullenv.NewFullEnvForTest(t, "http-flood", "flood_test.yaml")
	defer env.Close()
	env.Init()
	if err := env.Startup(); err != nil {
		t.Fatal(err)
	}
	defer env.Shutdown()

	apiKey, err := env.XbusCtl.RunReadLastLine("account", "renew-apikey", "test")
	if err != nil {
		t.Fatal(err)
	}
	t.Log("API Key:", apiKey)

	client := &http.Client{}

	req, err := http.NewRequest("GET", "http://localhost:8911/test/input", nil)
	if err != nil {
		t.Fatal(err)
	}

	req.Header.Add("Xbus-API-Key", apiKey)
	req.Header.Add("Accept", "application/json")
	req.Header.Add("Accept-Item-Encoding", "text")

	received := make(map[int]bool)

	for i := 0; i != 40; i++ {
		time.Sleep(250 * time.Millisecond)
		resp, err := client.Do(req)
		if err != nil {
			t.Fatal(err)
		}
		if resp.StatusCode != 200 {
			t.Fatal("Expected a 200 status, got", resp.StatusCode)
		}
		b, err := ioutil.ReadAll(resp.Body)
		if err != nil {
			t.Fatal("Could not read the response body")
		}
		data := make(map[string]interface{})
		if err := json.Unmarshal(b, &data); err != nil {
			t.Fatal(err)
		}
		item := data["envelope"].(map[string]interface{})["events"].([]interface{})[0].(map[string]interface{})["items"].([]interface{})[0].(string)
		value, err := strconv.Atoi(item)
		if err != nil {
			t.Fatal(err)
		}
		if r, _ := received[value]; r {
			t.Fatalf("Received '%d' twice", value)
		}
		received[value] = true
	}
}
