package http

import (
	"bytes"
	"context"
	"encoding/json"
	"io/ioutil"
	"net/http"
	"testing"
	"time"

	"github.com/gogo/protobuf/proto"
	"github.com/stretchr/testify/assert"
	yaml "gopkg.in/yaml.v2"

	"xbus.io/go-xbus/api"
)

func TestWebhookConsumer(t *testing.T) {
	for _, encoding := range []string{"application/json", "application/x-protobuf"} {
		t.Run(encoding, func(t *testing.T) {
			handlerC := make(chan bool)
			// At first the handler will fail with a temporary error
			var httpCode = http.StatusBadGateway
			webhookTestHandler := func(w http.ResponseWriter, r *http.Request) {
				var payload api.ActorProcessRequest
				assert.Equal(t, encoding, r.Header.Get("Content-Type"))
				switch encoding {
				case "application/json":
					dec := json.NewDecoder(r.Body)
					if err := dec.Decode(&payload); err != nil {
						w.WriteHeader(http.StatusBadRequest)
						w.Write([]byte("Cannot decode body: " + err.Error()))
						return
					}
				case "application/x-protobuf":
					bytes, err := ioutil.ReadAll(r.Body)
					if err != nil {
						w.WriteHeader(http.StatusInternalServerError)
						w.Write([]byte("Cannot read body: " + err.Error()))
						return
					}
					if err := proto.Unmarshal(bytes, &payload); err != nil {
						w.WriteHeader(http.StatusBadRequest)
						w.Write([]byte("Cannot decode body: " + err.Error()))
						return
					}
				default:
					panic("Unknown encoding: " + encoding)
				}
				t.Log("Decoded: ", payload)

				replyContentType := r.Header.Get("Accept")

				var reply []byte

				switch replyContentType {
				case "application/json":
					resp := map[string]interface{}{
						"context": map[string]interface{}{
							"processID": payload.Context.ProcessID.String(),
							"nodeID":    payload.Context.NodeID,
						},
						"status": "SUCCESS",
					}
					t.Log("encoding reply", resp)
					bytes, err := json.Marshal(resp)
					if err != nil {
						t.Errorf("Got an error: %s", err)
						w.WriteHeader(http.StatusInternalServerError)
						return
					}
					reply = bytes
				case "application/x-protobuf":
					bytes, err := proto.Marshal(&api.ActorProcessingState{
						Context: *payload.Context,
						Status:  api.ActorProcessingState_SUCCESS,
					})
					if err != nil {
						t.Errorf("Got an error: %s", err)
						w.WriteHeader(http.StatusInternalServerError)
						return
					}
					reply = bytes
				default:
					panic("invalid content type: " + replyContentType)
				}
				t.Log("will reply", string(reply))
				w.Header().Set("Content-Type", replyContentType)
				w.WriteHeader(httpCode)
				w.Write(reply)
				handlerC <- true
			}

			env, gateway := setupEnv(t, "webhook-test.yaml", true)
			defer env.Close()

			env.XbusCtl.MustRun("actor", "config", "set", "test-consumer",
				"http.webhook.content-type="+encoding)

			env.XbusCtl.MustRun("actor", "config", "set", "test-consumer",
				"http.webhook.http-codes.temporary-failure=502,503")

			if err := gateway.Startup(); err != nil {
				t.Error(err)
				return
			}
			defer gateway.Shutdown()

			t.Logf("Gateway is ready")

			sender := env.Emitter.Actor.OpenOutput(nil, "default", "test")
			must(t, sender.AddItems("test", api.Item("a item")))
			must(t, sender.Close())

			// Wait a little before setting our server so the hook fails and reattempt
			// later
			time.Sleep(2 * time.Second)

			mux := http.NewServeMux()
			mux.HandleFunc("/webhook", webhookTestHandler)

			server := http.Server{
				Addr:    "localhost:6758",
				Handler: mux,
			}

			go server.ListenAndServe()

			ctx, cancel := context.WithCancel(context.Background())
			defer cancel()
			defer server.Shutdown(ctx)

			// Wait until the handler failed once so the hook fails and reattempt
			// later
			select {
			case <-handlerC:
				t.Log("Should reattempt soon")
			case <-time.After(10 * time.Second):
				t.Fatal("timeout!")
			}

			httpCode = http.StatusOK

			select {
			case <-handlerC:
				t.Log("yay!")
			case <-time.After(10 * time.Second):
				t.Fatal("timeout!")
			}
			// TODO actively wait for the process completion instead of waiting 1 second
			time.Sleep(time.Second)
			out, err := env.XbusCtl.RunReadOutput("ps", "export", "--full")
			if err != nil {
				t.Fatal(err)
			}
			lines := bytes.Split(out, []byte("\n"))
			out = bytes.Join(lines[1:], []byte("\n"))
			t.Log(string(out))
			var psdetail []map[string]interface{}
			if err := yaml.Unmarshal(out, &psdetail); err != nil {
				t.Fatal(err)
			}
			assert.Equal(t, "DONE", psdetail[0]["state"].(map[interface{}]interface{})["status"])
		})
	}
}

func TestWebhookConsumerFlood(t *testing.T) {
	msgCount := 1000
	handlerC := make(chan bool, msgCount)

	webhookTestHandler := func(w http.ResponseWriter, r *http.Request) {
		var payload api.ActorProcessRequest
		assert.Equal(t, "application/x-protobuf", r.Header.Get("Content-Type"))

		bytes, err := ioutil.ReadAll(r.Body)
		if err != nil {
			w.WriteHeader(http.StatusInternalServerError)
			w.Write([]byte("Cannot read body: " + err.Error()))
			return
		}
		if err := proto.Unmarshal(bytes, &payload); err != nil {
			w.WriteHeader(http.StatusBadRequest)
			w.Write([]byte("Cannot decode body: " + err.Error()))
			return
		}

		t.Log("Decoded: ", payload)

		bytes, err = proto.Marshal(&api.ActorProcessingState{
			Context: *payload.Context,
			Status:  api.ActorProcessingState_SUCCESS,
		})
		if err != nil {
			t.Errorf("Got an error: %s", err)
			w.WriteHeader(http.StatusInternalServerError)
			return
		}

		t.Log("will reply", string(bytes))
		w.Write(bytes)
		handlerC <- true
	}

	env, gateway := setupEnv(t, "webhook-test.yaml", true)
	defer env.Close()

	env.XbusCtl.MustRun("actor", "config", "set", "test-consumer",
		"http.webhook.content-type=proto")

	if err := gateway.Startup(); err != nil {
		t.Error(err)
		return
	}
	defer gateway.Shutdown()

	t.Logf("Gateway is ready")

	mux := http.NewServeMux()
	mux.HandleFunc("/webhook", webhookTestHandler)

	server := http.Server{
		Addr:    "localhost:6758",
		Handler: mux,
	}

	go server.ListenAndServe()

	for i := 0; i != msgCount; i++ {
		status, err := env.Emitter.Actor.SendItems(
			nil, "default", "test", []api.Item{api.Item("item")})
		must(t, err)
		assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)
	}

	ctx, cancel := context.WithCancel(context.Background())
	defer server.Shutdown(ctx)
	defer cancel()

	for i := 0; i != msgCount; i++ {
		select {
		case <-handlerC:
			t.Logf("Got msg %d", i)
		case <-time.After(10 * time.Second):
			t.Fatal("timeout!")
		}
	}
	// TODO actively wait for the process completion instead of waiting 1 second
	time.Sleep(time.Second)

}
