package server

import (
	"fmt"
	"sync"
	"time"

	gnatsd "github.com/nats-io/gnatsd/server"
	"github.com/prometheus/client_golang/prometheus"
)

const refreshMinDelay = 25 * time.Millisecond

var emptyVarz = gnatsd.Varz{}

type natsMetrics struct {
	server *gnatsd.Server

	collectors []prometheus.Collector
	registerer prometheus.Registerer

	varzTimer *time.Timer
	varzLast  *gnatsd.Varz

	m sync.Mutex
}

func (m *natsMetrics) varzGauge(name, help string, f func(*gnatsd.Varz) float64) prometheus.Collector {
	return prometheus.NewGaugeFunc(prometheus.GaugeOpts{
		Name: "nats_server_varz_" + name,
		Help: help,
	}, func() float64 {
		return f(m.getVarz())
	})
}
func (m *natsMetrics) varzCounter(name, help string, f func(*gnatsd.Varz) float64) prometheus.Collector {
	return prometheus.NewCounterFunc(prometheus.CounterOpts{
		Name: "nats_server_varz_" + name,
		Help: help,
	}, func() float64 {
		return f(m.getVarz())
	})
}

func newNatsMetrics(
	natsServer *gnatsd.Server, reg prometheus.Registerer,
) *natsMetrics {
	m := natsMetrics{
		server:     natsServer,
		registerer: reg,
	}
	info := prometheus.NewGauge(prometheus.GaugeOpts{
		Name: "nats_server_info",
		Help: "Informations about the nats server",
		ConstLabels: map[string]string{
			"version": gnatsd.VERSION,
			"proto":   fmt.Sprintf("%d", gnatsd.PROTO),
		},
	})
	info.Set(1)

	m.collectors = []prometheus.Collector{
		info,
		m.varzGauge("connections", "Active client connections",
			func(v *gnatsd.Varz) float64 {
				return float64(v.Connections)
			}),
		m.varzCounter("total_connections", "Total client connections",
			func(v *gnatsd.Varz) float64 {
				return float64(v.TotalConnections)
			}),
		m.varzGauge("routes", "Active routes count",
			func(v *gnatsd.Varz) float64 {
				return float64(v.Routes)
			}),
		m.varzGauge("remotes", "Remotes count",
			func(v *gnatsd.Varz) float64 {
				return float64(v.Remotes)
			}),
		m.varzCounter("msg_in", "Number of message received",
			func(v *gnatsd.Varz) float64 {
				return float64(v.InMsgs)
			}),
		m.varzCounter("msg_out", "Number of message emitted",
			func(v *gnatsd.Varz) float64 {
				return float64(v.OutMsgs)
			}),
		m.varzCounter("bytes_in", "Number of bytes received",
			func(v *gnatsd.Varz) float64 {
				return float64(v.InBytes)
			}),
		m.varzCounter("bytes_out", "Number of bytes emitted",
			func(v *gnatsd.Varz) float64 {
				return float64(v.OutBytes)
			}),
		m.varzGauge("slow_consumer", "Number of slow consumers",
			func(v *gnatsd.Varz) float64 {
				return float64(v.SlowConsumers)
			}),
		m.varzGauge("max_payload", "Max payload",
			func(v *gnatsd.Varz) float64 {
				return float64(v.MaxPayload)
			}),
		m.varzGauge("max_pending", "Max pending",
			func(v *gnatsd.Varz) float64 {
				return float64(v.MaxPending)
			}),
		m.varzGauge("subscriptions", "Number of active subscriptions",
			func(v *gnatsd.Varz) float64 {
				return float64(v.Subscriptions)
			}),
	}
	for _, c := range m.collectors {
		reg.MustRegister(c)
	}
	return &m
}

func (m *natsMetrics) getVarz() (varz *gnatsd.Varz) {
	m.m.Lock()
	if m.varzLast != nil {
		varz = m.varzLast
	} else {
		var err error
		varz, err = m.server.Varz(nil)
		if err != nil {
			varz = &emptyVarz
		} else {
			m.varzLast = varz
			m.varzTimer = time.AfterFunc(refreshMinDelay, func() {
				m.m.Lock()
				m.varzLast = nil
				m.varzTimer = nil
				m.m.Unlock()
			})
		}
	}
	m.m.Unlock()
	return
}

func (m *natsMetrics) Close() {
	m.m.Lock()
	if m.varzTimer != nil {
		m.varzTimer.Stop()
	}
	m.m.Unlock()
	for _, c := range m.collectors {
		m.registerer.Unregister(c)
	}
}
