package server

import (
	"crypto/ecdsa"
	"crypto/tls"
	"crypto/x509"
	gosql "database/sql"
	"errors"
	"fmt"
	"net/http"
	"os"
	"strconv"
	"strings"
	"sync"
	"time"

	raven "github.com/getsentry/raven-go"
	"github.com/golang-migrate/migrate/v4"
	gnatsd "github.com/nats-io/gnatsd/server"
	nats "github.com/nats-io/go-nats"
	"github.com/orus-io/yago"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	"github.com/slicebit/qb"

	xbus "xbus.io/go-xbus"
	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/auth"
	"bitbucket.org/orus-io/xbus/migration"
	"bitbucket.org/orus-io/xbus/pipeline"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/utils/license"
	"bitbucket.org/orus-io/xbus/version"
)

// ErrDBNeedUpgrade is returned if the database is not up-to-date with the
// server version
type ErrDBNeedUpgrade struct {
	CurrentVersion  uint
	RequiredVersion uint
}

// ErrDBNotVersionned is returned by checkDBVersion if the database contains
// some tables but is not versionned, which means it was created before
// version 3.0.0-beta.14
var ErrDBNotVersionned = errors.New(
	"database is not versioned. " +
		"Apply manually the migration scripts up to version 3.0.0-beta.14, " +
		"then run 'xbusd migrate'")

// ErrDBNotInitialized is when the database was never initialized
var ErrDBNotInitialized = errors.New(
	"Database is not initialized. Run 'xbusd init'",
)

// Error returns the formatted error message
func (e ErrDBNeedUpgrade) Error() string {
	return fmt.Sprintf("Database version is too old. Please run 'xbusd migrate'"+
		" Required version: %d, current version: %d",
		e.RequiredVersion, e.CurrentVersion,
	)
}

// InstanceProperties contains the description of a xbus instance
type InstanceProperties struct {
	Name       string
	Production bool
	Logo       string
}

// Options of a server
type Options struct {
	Instance                       InstanceProperties
	DatabaseDSN                    string
	DatabaseLogFlags               qb.LogFlags
	DatabaseMaxConnections         int
	ProcessStorageDBMaxConnections int
	RootCA                         string // a PEM or a path to a PEM file
	ServerCert                     string // a PEM or a path to a PEM file
	ServerKey                      string // a PEM or a path to a PEM file
	ClientCA                       string // a PEM or a path to a PEM file
	ClientCAKey                    string // a PEM or a path to a PEM file
	TLSTimeout                     float64
	NATSHost                       string
	NATSPort                       int
	NATSDebug                      bool
	NATSTrace                      bool
	ServeNATS                      bool
	ServeBroker                    bool
	ServeClientAPI                 bool
	ServeDirector                  bool
	ServeEnvelopeStorage           bool
	ServeControl                   bool
	ServeMetrics                   bool
	ServePostmortem                bool
	ServeProcessState              bool
	ServeRegistration              bool
	ServeSentinel                  bool
	GnatsdOptions                  gnatsd.Options
	SentryDSN                      string
	SentryEnvironment              string
	MetricsHTTPHost                string
	MetricsHTTPPort                int
}

// Server the XBus server
type Server struct {
	options Options

	defaultModel        *model.Model
	defaultDB           *yago.DB
	processStorageModel *model.Model
	processStorageDB    *yago.DB

	accountStorage  storage.AccountStorage
	passwordStorage storage.PasswordStorage
	actorStorage    storage.ActorStorage
	pipelineStorage storage.PipelineStorage
	processStorage  storage.ProcessStorage
	envelopeStorage storage.EnvelopeStorage
	logStorage      storage.LogStorage
	jobStorage      storage.JobStorage
	sessionStorage  storage.SessionStorage

	processManager  *process.Manager
	pipelineManager *pipeline.Manager

	rootCA      *x509.Certificate
	serverCert  *x509.Certificate
	serverKey   *ecdsa.PrivateKey
	clientCA    *x509.Certificate
	clientCAKey *ecdsa.PrivateKey

	natsAuth    auth.XBusAuth
	natsServer  *gnatsd.Server
	natsConn    *nats.Conn
	natsMetrics *natsMetrics

	ServiceManager *services.Manager

	ravenClient *raven.Client

	metricsServer         *http.Server
	collectors            []prometheus.Collector
	defaultDBStats        *gosql.DBStats
	processStorageDBStats *gosql.DBStats
	dbStatsM              sync.Mutex

	log        xbus.Logger
	natsLogger xbus.NATSLogger
}

// NewServer creates a Server
func NewServer(options Options) (*Server, error) {
	if options.DatabaseMaxConnections == 0 {
		options.DatabaseMaxConnections = 10
	}
	if options.ProcessStorageDBMaxConnections == 0 {
		options.ProcessStorageDBMaxConnections = options.DatabaseMaxConnections/3 + 1
	}
	s := Server{
		options:    options,
		log:        xbus.NewNoopLogger(),
		natsLogger: xbus.NewNATSLogger(xbus.NewNoopLogger()),
	}
	if err := s.SetupRaven(); err != nil {
		return nil, err
	}
	return &s, nil
}

// SetupRaven configures the raven client
func (s *Server) SetupRaven() error {
	if s.ravenClient != nil {
		return nil
	}
	if s.options.SentryDSN == "" {
		return nil
	}
	client, err := raven.New(s.options.SentryDSN)
	if err != nil {
		return err
	}
	s.ravenClient = client
	client.SetRelease(version.Version)
	client.SetTagsContext(map[string]string{
		"instance_name":       s.options.Instance.Name,
		"instance_production": fmt.Sprintf("%t", s.options.Instance.Production),
	})
	if s.options.SentryEnvironment != "" {
		client.SetEnvironment(s.options.SentryEnvironment)
	}
	return nil
}

// GetRavenClient returns the raven.Client instance
func (s *Server) GetRavenClient() *raven.Client {
	return s.ravenClient
}

// SetRavenTransport forces a transport on the raven client
// and creates a client even if no SentryDSN is set
func (s *Server) SetRavenTransport(transport raven.Transport) error {
	client, err := raven.New("")
	if err != nil {
		return err
	}
	s.ravenClient = client
	client.SetRelease(version.Version)
	if s.options.SentryEnvironment != "" {
		client.SetEnvironment(s.options.SentryEnvironment)
	}
	s.ravenClient.Transport = transport
	return nil
}

func (s *Server) wrapLogger(logger xbus.Logger) xbus.Logger {
	if s.ravenClient != nil {
		logger = xbus.NewRavenLogger(s.ravenClient, logger)
	}
	return logger
}

// SetLogger changes the logger used by server
func (s *Server) SetLogger(logger xbus.Logger) {
	if logger == nil {
		logger = xbus.NewNoopLogger()
	}

	logger = s.wrapLogger(logger)

	s.log = xbus.NewPrefixedLogger("[server] ", logger)
	s.natsLogger = xbus.NewNATSLogger(xbus.NewPrefixedLogger("[gnatsd] ", logger))
}

// SetSQLDB sets the model & db
func (s *Server) SetSQLDB(m *model.Model, db *yago.DB) {
	s.defaultModel = m
	s.defaultDB = db
}

// SetEnvelopeStorageDB sets the model & db for the envelope storage
func (s *Server) SetEnvelopeStorageDB(m *model.Model, db *yago.DB) {
	s.processStorageModel = m
	s.processStorageDB = db
}

// GetEnvelopeStorage returns the envelope storage
func (s *Server) GetEnvelopeStorage() storage.EnvelopeStorage {
	return s.envelopeStorage
}

func (s *Server) setupSQLDB() error {
	if s.defaultModel == nil {
		m, db, err := model.NewDB(
			"postgres", s.options.DatabaseDSN, s.options.DatabaseMaxConnections,
			xbus.NewPrefixedLogger("[sql] ", s.log),
			s.options.DatabaseLogFlags)
		if err != nil {
			s.log.Errorf("Cannot connect to the database.Got err %s", err)
			return err
		}
		s.SetSQLDB(m, db)
	}
	if s.processStorageModel == nil {
		m, db, err := model.NewDB(
			"postgres", s.options.DatabaseDSN, s.options.ProcessStorageDBMaxConnections,
			xbus.NewPrefixedLogger("[sql(processstorage)] ", s.log),
			s.options.DatabaseLogFlags)
		if err != nil {
			s.log.Errorf("Cannot connect to the database.Got err %s", err)
			return err
		}
		s.SetEnvelopeStorageDB(m, db)
	}
	return nil
}

func (s *Server) checkDBVersion() error {
	if s.defaultModel == nil {
		panic("DB connection not yet initialized")
	}
	_, db, err := model.NewDB(
		"postgres", s.options.DatabaseDSN, 1,
		xbus.NewPrefixedLogger("[sql] ", s.log),
		s.options.DatabaseLogFlags)
	if err != nil {
		return err
	}
	m, source, err := migration.New(db)
	if err != nil {
		return fmt.Errorf("Could not setup the migration: %s", err)
	}
	currentVersion, _, err := m.Version()
	if err == migrate.ErrNilVersion {
		m.Close()
		row := s.defaultDB.Engine.DB().QueryRow(
			"SELECT COUNT(*) FROM pg_tables WHERE schemaname = 'public'")
		var tableCount int
		if err := row.Scan(&tableCount); err != nil {
			return fmt.Errorf("error checking the database: %s", err)
		}
		if tableCount <= 1 {
			return ErrDBNotInitialized
		}
		return ErrDBNotVersionned
	}
	defer m.Close()
	if err != nil {
		return err
	}
	// Lookup the last available db version
	lastVersion, err := source.First()
	if err != nil {
		return err
	}
	for {
		next, err := source.Next(lastVersion)
		if pathError, ok := err.(*os.PathError); err == os.ErrNotExist || ok && pathError.Err == os.ErrNotExist {
			break
		} else if err != nil {
			return err
		}
		lastVersion = next
	}
	if currentVersion != lastVersion {
		return ErrDBNeedUpgrade{
			CurrentVersion:  currentVersion,
			RequiredVersion: lastVersion,
		}
	}
	return nil
}

// AccountStorage returns the underlying AccountStorage
func (s *Server) AccountStorage() storage.AccountStorage {
	return s.accountStorage
}

// PasswordStorage returns the underlying PasswordStorage
func (s *Server) PasswordStorage() storage.PasswordStorage {
	return s.passwordStorage
}

// ActorStorage returns the underlying ActorStorage
func (s *Server) ActorStorage() storage.ActorStorage {
	return s.actorStorage
}

// ProcessStorage returns the underlying ProcessStorage
func (s *Server) ProcessStorage() storage.ProcessStorage {
	return s.processStorage
}

// ProcessManager returns the process manager
func (s *Server) ProcessManager() *process.Manager {
	return s.processManager
}

func (s *Server) setupMetricsServer() {
	mux := http.NewServeMux()
	mux.Handle("/metrics", promhttp.Handler())
	if s.options.MetricsHTTPHost == "" {
		s.options.MetricsHTTPHost = "localhost"
	}
	s.metricsServer = &http.Server{
		Addr:    s.options.MetricsHTTPHost + ":" + strconv.Itoa(s.options.MetricsHTTPPort),
		Handler: mux,
	}
	go func() {
		if err := s.metricsServer.ListenAndServe(); err != nil {
			s.log.Errorf("Error in metricsServer.ListenAndServe(): %s", err)
		}
	}()
	s.log.Noticef("Metrics served on http://%s:%d/metrics", s.options.MetricsHTTPHost, s.options.MetricsHTTPPort)
}

func (s *Server) setupStorage() {
	s.accountStorage = sql.NewAccountStorage(s.defaultModel, s.defaultDB, s.log)
	s.passwordStorage = sql.NewPasswordStorage(s.defaultDB, s.log)
	s.actorStorage = sql.NewActorStorage(s.defaultModel, s.defaultDB, s.log)
	s.envelopeStorage = sql.NewEnvelopeStorage(s.defaultModel, s.defaultDB, s.log)
	s.pipelineStorage = sql.NewPipelineStorage(s.defaultModel, s.defaultDB, s.log)
	s.processStorage = sql.NewProcessStorage(s.processStorageModel, s.processStorageDB, s.log)
	s.logStorage = sql.NewLogStorage(s.defaultModel, s.defaultDB, s.log)
	s.jobStorage = sql.NewJobStorage(s.defaultDB, s.log)
	s.sessionStorage = sql.NewSessionStorage(s.defaultDB, s.log)
}

func (s *Server) setupManagers() {
	s.pipelineManager = pipeline.NewManager(
		s.actorStorage, s.pipelineStorage)
	s.processManager = process.NewManager(
		s.pipelineManager, s.actorStorage, s.processStorage, s.log,
	)
}

func (s *Server) setupCrypto() error {
	rootCA, err := tools.ReadPEMCertificate(s.options.RootCA)
	if err != nil {
		return err
	}
	serverCert, err := tools.ReadPEMCertificate(s.options.ServerCert)
	if err != nil {
		return err
	}
	serverKey, err := tools.ReadPEMPrivateKey(s.options.ServerKey)
	if err != nil {
		return err
	}
	clientCA, err := tools.ReadPEMCertificate(s.options.ClientCA)
	if err != nil {
		return err
	}
	clientCAKey, err := tools.ReadPEMPrivateKey(s.options.ClientCAKey)
	if err != nil {
		return err
	}
	s.rootCA = rootCA
	s.serverCert = serverCert
	s.serverKey = serverKey
	s.clientCA = clientCA
	s.clientCAKey = clientCAKey

	return nil
}

// Startup starts the server
func (s *Server) Startup() (err error) {
	if s.options.Instance.Production && strings.Contains(license.License(), "personal") {
		s.log.Errorf(
			"This binary license is not allowed to run a production instance. You should get a suitable binary, or remove the 'production' flag of your instance.")
		return fmt.Errorf("incompatible binary")
	}
	defer func() {
		if err != nil {
			s.Shutdown()
		}
	}()
	if s.metricsServer == nil && s.options.MetricsHTTPPort != 0 {
		s.setupMetricsServer()
	}
	if err = s.setupSQLDB(); err != nil {
		return err
	}
	if err = s.checkDBVersion(); err != nil {
		return fmt.Errorf("Could not check db version: %s", err)
	}
	s.setupStorage()
	s.setupManagers()
	if err = s.setupCrypto(); err != nil {
		return err
	}
	if s.options.ServeNATS {
		if err = s.setupNATS(); err != nil {
			return err
		}
		if err = s.StartNATS(); err != nil {
			return err
		}
	}
	s.setupMetrics()
	s.setupServices()
	s.log.Noticef("Starting services")
	return s.ServiceManager.Startup()
}

// Shutdown stops the services and nats server (if relevant)
func (s *Server) Shutdown() {
	s.clearMetrics()
	if s.ServiceManager != nil {
		if err := s.ServiceManager.Shutdown(); err != nil {
			s.log.Errorf("Error shutting down services: %s", err)
		}
	}
	if s.options.ServeNATS {
		s.StopNATS()
	}
	if s.ravenClient != nil {
		s.ravenClient.Close()
		s.ravenClient = nil
	}
}

func (s *Server) setupNATS() error {
	certPool := x509.NewCertPool()
	certPool.AddCert(s.clientCA)

	s.natsAuth = auth.NewXBusAuth(
		s.accountStorage, s.passwordStorage, s.actorStorage, s.sessionStorage, s.log)

	gnatsdOptions := s.options.GnatsdOptions
	gnatsdOptions.Host = s.options.NATSHost
	gnatsdOptions.Port = s.options.NATSPort
	gnatsdOptions.NoSigs = true
	gnatsdOptions.TLSConfig = &tls.Config{
		Certificates: []tls.Certificate{
			{
				Certificate: [][]byte{s.serverCert.Raw},
				PrivateKey:  s.serverKey,
				Leaf:        s.serverCert,
			},
		},
		ClientAuth: tls.VerifyClientCertIfGiven,
		ClientCAs:  certPool,
		// make sure no client can downgrade to TLS < 1.2
		MinVersion: tls.VersionTLS12,
	}
	gnatsdOptions.TLSConfig.BuildNameToCertificate()
	gnatsdOptions.CustomClientAuthentication = &s.natsAuth
	gnatsdOptions.TLSTimeout = s.options.TLSTimeout

	s.natsServer = gnatsd.New(&gnatsdOptions)
	s.natsServer.SetLogger(s.natsLogger, s.options.NATSDebug, s.options.NATSTrace)

	return nil
}

// getNATSPort returns the nats server port, when an
// error occurs the returned value should be ignored.
func (s *Server) getNATSPort() (int, error) {
	endpoint := s.natsServer.Addr()
	if endpoint == nil {
		return 0, errors.New("Undefined endpoint")
	}
	subStr := strings.SplitAfter(endpoint.String(), ":")
	strPort := subStr[len(subStr)-1]
	port, err := strconv.Atoi(strPort)
	if err != nil {
		return 0, err
	}
	return port, nil
}

// NATSUrl returns the nats server url
func (s *Server) NATSUrl() string {
	port, err := s.getNATSPort()
	if err != nil {
		panic(err)
	}
	return fmt.Sprintf("nats://%s:%d", s.options.NATSHost, port)
}

// NATSConnectOptions returns nats.Options to connect to the
// bus with full access
func (s *Server) NATSConnectOptions() nats.Options {
	certPool := x509.NewCertPool()
	certPool.AddCert(s.rootCA)

	return nats.Options{
		Url:   s.NATSUrl(),
		Token: s.natsAuth.AuthToken,
		TLSConfig: &tls.Config{
			RootCAs: certPool,
		},
		Secure: true,
	}
}

func (s *Server) waitForNATS() error {

	if !s.natsServer.ReadyForConnections(time.Second) {
		return errors.New("Server still has no listen endpoint")
	}
	url := s.NATSUrl()
	if url == "" {
		return errors.New("Server still has no listen endpoint")
	}

	/*
		opts := s.NATSConnectOptions()

		var (
			ready   bool
			lastErr error
		)
		for time.Now().Before(timeout) {
			// let the goroutine start to listen
			runtime.Gosched()

			natsConn, err := opts.Connect()
			lastErr = err
			if err == nil {
				natsConn.Close()
				ready = true
				break
			}
		}
		if !ready {
			return fmt.Errorf("Cannot connect to the nats server: %s", lastErr)
		}
	*/
	s.log.Noticef("Nats server responding on %s", s.NATSUrl())

	// Store the actual port back into the options so that a restart
	// will reuse the same port, even if the original value is -1 which means
	// a random port.
	port, err := s.getNATSPort()
	if err != nil {
		panic(err)
	}
	s.options.NATSPort = port
	return nil
}

// StartNATS starts the nats server
func (s *Server) StartNATS() error {
	s.log.Noticef("Starting NATS server")

	go s.natsServer.Start()

	if err := s.waitForNATS(); err != nil {
		s.StopNATS()
		return err
	}

	if s.options.MetricsHTTPPort != 0 {
		s.natsMetrics = newNatsMetrics(
			s.natsServer, prometheus.DefaultRegisterer)
	}

	connOptions := nats.Options{
		Url:    s.NATSUrl(),
		Token:  s.natsAuth.AuthToken,
		Name:   fmt.Sprintf("server@%s(xbusd@%s)", api.ProtocolVersion, version.Version),
		Secure: true,
		AsyncErrorCB: func(c *nats.Conn, sub *nats.Subscription, err error) {
			if sub != nil {
				s.log.Errorf("Error on sub %s: %s", sub.Subject, err)
			} else {
				s.log.Errorf("Error %s", err)
			}
		},
	}

	conn, err := connOptions.Connect()

	if err != nil {
		s.log.Errorf("Could not connect to the NATS server: %s", err)
		s.StopNATS()
		return err
	}

	s.natsConn = conn

	return nil
}

// StopNATS stops the nats server
func (s *Server) StopNATS() {
	if s.natsMetrics != nil {
		s.natsMetrics.Close()
		s.natsMetrics = nil
	}
	if s.natsConn != nil {
		s.natsConn.Close()
		s.natsConn = nil
	}
	if s.natsServer != nil {
		s.natsServer.Shutdown()
		s.natsServer = nil
	}
}

// DBStats returns the Stats of the two database connexions, with a cache of 10ms
func (s *Server) DBStats() (defaultDBStats *gosql.DBStats, processStorageDBStats *gosql.DBStats) {
	s.dbStatsM.Lock()
	if s.defaultDBStats != nil {
		defaultDBStats, processStorageDBStats = s.defaultDBStats, s.processStorageDBStats
	} else {
		{
			stats := s.defaultDB.Engine.DB().Stats()
			defaultDBStats = &stats
			s.defaultDBStats = defaultDBStats
		}
		{
			stats := s.processStorageDB.Engine.DB().Stats()
			processStorageDBStats = &stats
			s.processStorageDBStats = processStorageDBStats
		}

		time.AfterFunc(time.Millisecond*10, func() {
			s.dbStatsM.Lock()
			s.defaultDBStats = nil
			s.processStorageDBStats = nil
			s.dbStatsM.Unlock()
		})
	}
	s.dbStatsM.Unlock()
	return
}

// DefaultDBStats returns the Stats of the default database
func (s *Server) DefaultDBStats() *gosql.DBStats {
	defaultDBStats, _ := s.DBStats()
	return defaultDBStats
}

// ProcessStorageDBStats returns the Stats of the process storage database
func (s *Server) ProcessStorageDBStats() *gosql.DBStats {
	_, processStorageDBStats := s.DBStats()
	return processStorageDBStats
}

func (s *Server) dbGauges(name, help string, getter func(*gosql.DBStats) float64) []prometheus.Collector {
	return []prometheus.Collector{
		prometheus.NewGaugeFunc(prometheus.GaugeOpts{
			Name:        "xbus_server_db_" + name,
			Help:        help,
			ConstLabels: map[string]string{"database": "default"},
		}, func() float64 {
			return getter(s.DefaultDBStats())
		}),
		prometheus.NewGaugeFunc(prometheus.GaugeOpts{
			Name:        "xbus_server_db_" + name,
			Help:        help,
			ConstLabels: map[string]string{"database": "process-storage"},
		}, func() float64 {
			return getter(s.ProcessStorageDBStats())
		}),
	}
}
func (s *Server) dbCounters(name, help string, getter func(*gosql.DBStats) float64) []prometheus.Collector {
	return []prometheus.Collector{
		prometheus.NewCounterFunc(prometheus.CounterOpts{
			Name:        "xbus_server_db_" + name,
			Help:        help,
			ConstLabels: map[string]string{"database": "default"},
		}, func() float64 {
			return getter(s.DefaultDBStats())
		}),
		prometheus.NewCounterFunc(prometheus.CounterOpts{
			Name:        "xbus_server_db_" + name,
			Help:        help,
			ConstLabels: map[string]string{"database": "process-storage"},
		}, func() float64 {
			return getter(s.ProcessStorageDBStats())
		}),
	}
}

func (s *Server) setupMetrics() {
	hostname, err := os.Hostname()
	if err != nil {
		s.log.Errorf("Could not get hostname: %s", err)
		hostname = "unknown"
	}
	xbusInfoGauge := prometheus.NewGauge(prometheus.GaugeOpts{
		Name: "xbus_server_info",
		Help: "Informations about the xbus server",
		ConstLabels: map[string]string{
			"version":             version.Version,
			"commit":              version.CommitHash,
			"protocol":            api.ProtocolVersion,
			"instance_name":       s.options.Instance.Name,
			"instance_production": fmt.Sprintf("%t", s.options.Instance.Production),
			"hostname":            hostname,
			"sentry_environment":  s.options.SentryEnvironment,
		},
	})
	xbusInfoGauge.Set(1)
	s.collectors = []prometheus.Collector{
		xbusInfoGauge,
	}
	s.collectors = append(s.collectors, s.dbGauges(
		"max_open_connections", "Maximum number of open connections to the database",
		func(stats *gosql.DBStats) float64 { return float64(stats.MaxOpenConnections) })...,
	)
	s.collectors = append(s.collectors, s.dbGauges(
		"open_connections", "The number of established connections both in use and idle",
		func(stats *gosql.DBStats) float64 { return float64(stats.OpenConnections) })...,
	)
	s.collectors = append(s.collectors, s.dbGauges(
		"in_use", "The number of connections currently in use",
		func(stats *gosql.DBStats) float64 { return float64(stats.InUse) })...,
	)
	s.collectors = append(s.collectors, s.dbGauges(
		"idle", "The number of idle connections",
		func(stats *gosql.DBStats) float64 { return float64(stats.Idle) })...,
	)
	s.collectors = append(s.collectors, s.dbCounters(
		"wait_count", "The total number of connections waited for",
		func(stats *gosql.DBStats) float64 { return float64(stats.WaitCount) })...,
	)
	s.collectors = append(s.collectors, s.dbCounters(
		"wait_duration", "The total time blocked waiting for a new connection, in seconds",
		func(stats *gosql.DBStats) float64 { return float64(stats.WaitDuration) / float64(time.Second) })...,
	)
	s.collectors = append(s.collectors, s.dbCounters(
		"max_idle_closed", "The total number of connections closed due to SetMaxIdleConns",
		func(stats *gosql.DBStats) float64 { return float64(stats.MaxIdleClosed) })...,
	)
	s.collectors = append(s.collectors, s.dbCounters(
		"max_lifetime_closed", "The total number of connections closed due to SetConnMaxLifetime",
		func(stats *gosql.DBStats) float64 { return float64(stats.MaxLifetimeClosed) })...,
	)
	for _, c := range s.collectors {
		prometheus.MustRegister(c)
	}
}

func (s *Server) clearMetrics() {
	for _, c := range s.collectors {
		prometheus.Unregister(c)
	}
}
