// Copyright 2016 Orus.io. All rights reserved.

package broker

import (
	"context"
	"time"

	"github.com/nats-rpc/gogo-nrpc"

	"xbus.io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

const (
	// BrokerQueue is the name we use to queueSub to a queue with the broker
	BrokerQueue = "broker_queue"

	// DefaultPoolSize is the default size of the worker pool
	DefaultPoolSize = 5

	// DefaultPoolMaxPending is the default number of requests that can be
	// pending
	DefaultPoolMaxPending = 5

	// DefaultPoolMaxPendingDuration is the default max pending duration for a
	// a request
	DefaultPoolMaxPendingDuration = time.Second
)

// Options is used to create a Service
type Options struct {
	services.Options

	ActorStorage    storage.ActorStorage
	EnvelopeStorage storage.EnvelopeStorage
	LogStorage      storage.LogStorage
	ProcessManager  *process.Manager

	PoolSize               uint
	PoolMaxPending         uint
	PoolMaxPendingDuration time.Duration
}

// NewService returns a new Service
func (o Options) NewService() *Service {
	if o.PoolSize == 0 {
		o.PoolSize = DefaultPoolSize
	}
	if o.PoolMaxPending == 0 {
		o.PoolMaxPending = DefaultPoolMaxPending
	}
	if o.PoolMaxPendingDuration == 0 {
		o.PoolMaxPendingDuration = DefaultPoolMaxPendingDuration
	}
	return &Service{
		BaseService:    services.NewBaseService("broker", o.Options),
		Opts:           o,
		processManager: o.ProcessManager,
	}
}

// New returns a new Service
func New(options Options) *Service {
	return options.NewService()
}

// Service is the central node of all envelope exchanges
type Service struct {
	services.BaseService
	Opts Options

	processManager *process.Manager

	workers       *nrpc.WorkerPool
	api           *api.Client
	brokerHandler *api.BrokerHandler
}

// Startup initialize all the subscriptions
func (s *Service) Startup() error {
	s.workers = nrpc.NewWorkerPool(
		context.Background(),
		s.Opts.PoolSize, s.Opts.PoolMaxPending, s.Opts.PoolMaxPendingDuration)

	s.brokerHandler = api.NewBrokerConcurrentHandler(
		s.workers, s.Conn, brokerServer{s},
	)

	return s.SubscribeAll(
		s.HandlerQueueSubscribe(s.brokerHandler, BrokerQueue),
	)
}
