// Copyright 2017 Orus.io. All rights reserved.

package clientapi

import (
	"context"
	"fmt"
	"time"

	nrpc "github.com/nats-rpc/gogo-nrpc"

	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

// QueueName is the queue name used by the service
const QueueName = "xbus.services.clientapi"

// Options for the Service
type Options struct {
	services.Options

	ActorStorage   storage.ActorStorage
	AccountStorage storage.AccountStorage
	SessionStorage storage.SessionStorage
}

// NewService instanciate a Service
func NewService(o Options) *Service {
	return &Service{
		o:           o,
		BaseService: services.NewBaseService("clientapi", o.Options),
	}
}

// Service provides an api for administrators
type Service struct {
	services.BaseService

	o Options

	clientAPIHandler *api.ClientAPIHandler
}

// Startup starts the service
func (s *Service) Startup() error {
	whoamiSubject, _ := api.WhoAmISubjects("*")
	s.clientAPIHandler = api.NewClientAPIHandler(
		context.Background(), s.Conn, clientAPIServer{s},
	)
	whoamiHandler := api.NewWhoAmIHandler(
		context.Background(), s.Conn, clientAPIServer{s},
	)
	// The two handlers have the same subject prefix 'xbus.client',
	// so we need to subscribe to their methods directly
	// TODO the api need some rework to avoid this situation
	if err := s.SubscribeAll(
		services.QueueSubscribe(
			api.XbusAPIPrefix+"client.*.getactors", QueueName,
			s.clientAPIHandler.Handler,
		),
		services.QueueSubscribe(
			api.XbusAPIPrefix+"client.*.getactors.*", QueueName,
			s.clientAPIHandler.Handler,
		),
		services.QueueSubscribe(
			api.XbusAPIPrefix+"client.*.getsessiontoken", QueueName,
			s.clientAPIHandler.Handler,
		),
		services.QueueSubscribe(
			api.XbusAPIPrefix+"client.*.getsessiontoken.*", QueueName,
			s.clientAPIHandler.Handler,
		),
		services.QueueSubscribe(
			whoamiSubject, QueueName,
			whoamiHandler.Handler),
		services.QueueSubscribe(
			whoamiSubject+".json", QueueName,
			whoamiHandler.Handler),
	); err != nil {
		s.Shutdown()
		return err
	}
	return nil
}

type clientAPIServer struct {
	*Service
}

func (s clientAPIServer) WhoAmI(
	ctx context.Context, identifier string,
) (api.Account, error) {
	var (
		account api.Account
		ok      bool
	)

	if id, isUUID := api.UUIDFromString(identifier); isUUID == nil {
		account, ok = s.o.AccountStorage.Get(id)
		if !ok {
			panic("No account by this id")
		}
	} else {
		account, ok = s.o.AccountStorage.GetByName(identifier)
		if !ok {
			account, ok = s.o.AccountStorage.GetByAPIKey(identifier)
			if !ok {
				panic("No account by this name or api key")
			}
		}
	}
	return account, nil
}

func (s clientAPIServer) GetActors(ctx context.Context) (
	resp api.ClientGetActorsReply, err error,
) {
	accountSID := nrpc.GetRequest(ctx).ServiceParam("accountID")
	accountID, err := api.UUIDFromString(accountSID)
	if err != nil {
		return
	}

	resp.Actors = s.o.ActorStorage.ListAccount(accountID)
	return
}

func (s clientAPIServer) GetSessionToken(
	ctx context.Context, request api.GetSessionTokenRequest,
) (api.SessionToken, error) {
	accountSID := nrpc.GetRequest(ctx).ServiceParam("accountID")
	accountID, err := api.UUIDFromString(accountSID)
	if err != nil {
		return api.SessionToken{}, err
	}
	if request.InvalidateToken != "" {
		session, err := s.o.SessionStorage.Get(request.InvalidateToken)
		if err != nil {
			s.Log.Errorf("Error invalidating session token: %s", err)
		}
		if session != nil {
			if accountID != session.AccountID {
				return api.SessionToken{}, fmt.Errorf("InvalidateToken value is invalid (trying to invalidate another account session?)")
			}
		}
	}

	// Max session duration is 12h (TODO make it setting)
	var expiresIn = request.ExpiresIn
	if expiresIn > 60*60*12 {
		expiresIn = 60 * 60 * 12
	}

	var newToken = api.SessionToken{
		Token:      tools.RandomString(32),
		ValidUntil: time.Now().Add(time.Duration(expiresIn) * time.Second).UTC(),
	}

	if err := s.o.SessionStorage.Set(storage.Session{
		SessionToken: newToken,
		AccountID:    accountID,
	}); err != nil {
		s.Log.Errorf("Error saving new session: %s", err)
		return api.SessionToken{}, fmt.Errorf("Internal Error")
	}
	// Now that we secured the creation of a new token, we can delete the old one

	if err := s.o.SessionStorage.Delete(request.InvalidateToken); err != nil {
		s.Log.Errorf(
			"Could not delete the session %s, got: %s",
			request.InvalidateToken, err)
	}
	return newToken, nil
}
