// Copyright 2016 Orus.io. All rights reserved.

package control

import (
	"context"
	"gopkg.in/yaml.v2"

	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/api/control"
	"bitbucket.org/orus-io/xbus/pipeline"
)

type pipelineServer struct {
	*Service
}

func (s pipelineServer) Query(
	ctx context.Context, req control.PipelineQueryRequest,
) (resp control.PipelineQueryReply, err error) {
	graphs := s.o.PipelineManager.Query(req.Name, "", false)
	for _, graph := range graphs {
		resp.PipelineList = append(resp.PipelineList,
			api.PipelineInfo{
				ID:      graph.ID,
				Name:    graph.Name,
				Version: graph.Version,
				Status:  graph.Status,
			})
	}
	return
}

func (s pipelineServer) Save(
	ctx context.Context, req control.PipelineSaveRequest,
) (
	resp control.PipelineSaveReply, err error,
) {
	var pl = &pipeline.Pipeline{}
	if err = yaml.Unmarshal([]byte(req.Graph), &pl.Graph); err != nil {
		err = api.ClientProtoError("Error decoding graph: %s", err)
		return
	}

	if req.Info.ID.IsZero() {
		pipes := s.o.PipelineManager.Query(req.Info.Name, req.Info.Version, false)
		if len(pipes) > 1 {
			err = api.ClientProtoError(
				"Multiple pipeline matches: %s-%s", req.Info.Name, req.Info.Version)
			return
		}
		if len(pipes) == 1 {
			pl.ID = pipes[0].ID
		}
		pl.Name = req.Info.Name
		pl.Version = req.Info.Version
	} else {
		curPipeline, exists := s.o.PipelineManager.GetPipeline(req.Info.ID)
		if exists {
			err = api.ClientProtoError(
				"No such pipeline: %s", req.Info.ID)
			return
		}
		pl.ID = req.Info.ID
		pl.Name = curPipeline.Name
		pl.Version = curPipeline.Version
	}

	validator := pipeline.NewValidator(pl.Graph)

	for _, e := range validator.Validate() {
		resp.ValidationMessages = append(resp.ValidationMessages, e.Error())
	}

	for _, e := range validator.ValidateActors(s.o.ActorStorage) {
		resp.ValidationMessages = append(resp.ValidationMessages, e.Error())
	}

	pl, err = s.o.PipelineManager.Save(pl)
	if err != nil {
		err = api.ClientProtoError("Save pipeline error: %s", err)
		return
	}

	var b []byte
	b, err = yaml.Marshal(pl.Graph)
	if err != nil {
		err = api.ServerProtoError(err)
		return
	}
	resp.IsValid = len(resp.ValidationMessages) == 0
	resp.Graph = string(b)
	resp.Info = pl.PipelineInfo
	return
}

func (s pipelineServer) SetStatus(ctx context.Context, req api.PipelineInfo) (resp control.PipelineSetStatusReply, err error) {

	var info = api.PipelineInfo{ID: req.ID}
	var pl pipeline.Pipeline
	if info.ID.IsZero() {
		var pipes = s.o.PipelineManager.Query(req.Name, req.Version, false)
		if len(pipes) == 0 {
			err = api.ClientProtoError(
				"No such graph: %s-%s", req.Name, req.Version)
			return
		}
		if len(pipes) > 1 {
			err = api.ClientProtoError("Multiple graphs found")
			return
		}
		pl = pipes[0]
		info.ID = pl.ID
		info.Name = pl.Name
		info.Version = pl.Version
		info.Status = pl.Status
	} else {
		var ok bool
		pl, ok = s.o.PipelineManager.GetPipeline(info.ID)
		if !ok {
			err = api.ClientProtoError(
				"No such graph: %s", info.ID)
			return
		}
		info.ID = pl.ID
		info.Name = pl.Name
		info.Version = pl.Version
		info.Status = pl.Status
	}
	resp.Info = info
	if setStatusErr := s.o.PipelineManager.SetStatus(pl, req.Status); setStatusErr != nil {
		if list, ok := setStatusErr.(pipeline.ValidationErrorList); ok {
			resp.IsValid = false
			for _, e := range list {
				resp.ValidationMessages = append(resp.ValidationMessages, e.Error())
			}
		} else {
			err = api.ClientProtoError("%s", setStatusErr)
		}
	} else {
		resp.IsValid = true
		resp.Info.Status = req.Status
	}
	return
}

func (s pipelineServer) GetGraph(
	ctx context.Context, req api.PipelineInfo,
) (
	resp api.StringReply, err error,
) {
	var pipeID = req.ID
	var result *pipeline.Pipeline
	if pipeID.IsZero() {
		if req.GetName() == "" {
			err = api.ClientProtoError("Invalid graph query: must have an ID or a Name")
			return
		}
		pipes := s.o.PipelineManager.Query(
			req.GetName(), req.GetVersion(),
			req.GetStatus() == api.PipelineInfo_ACTIVE)
		switch len(pipes) {
		case 0:
			err = api.ClientProtoError("No such graph")
			return
		case 1:
			pl := pipes[0]
			result = &pl
		default:
			var active *pipeline.Pipeline
			// get the draft version if unique
			for _, pl := range pipes {
				switch pl.Status {
				case api.PipelineInfo_DRAFT:
					if result != nil {
						err = api.ClientProtoError(
							"Graph %s has multiple drafts: please specify the version",
							req.GetName())
						return
					}
					result = &pl
				case api.PipelineInfo_ACTIVE:
					active = &pl
				}
			}
			if result == nil && active != nil {
				result = active
			}
			if result == nil {
				err = api.ClientProtoError(
					"Graph %s has no draft of active version: please specify the version you want",
					req.GetName(),
				)
				return
			}
		}
	} else {
		pl, exists := s.o.PipelineManager.GetPipeline(req.ID)
		if !exists {
			err = api.ClientProtoError("No such graph")
			return
		}
		result = &pl
	}
	b, err := yaml.Marshal(result.Graph)
	if err != nil {
		panic(err)
	}
	resp.Value = string(b)
	return
}
