// Copyright 2016 Orus.io. All rights reserved.

// Package core provides the 'core' service, which must be started once and only once
// in every server process
package core

import (
	"context"

	"xbus.io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/pipeline"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

// Options is used to create a 'core' service
type Options struct {
	services.Options

	ActorStorage    storage.ActorStorage
	ProcessStorage  storage.ProcessStorage
	ProcessManager  *process.Manager
	PipelineManager *pipeline.Manager
	LogStorage      storage.LogStorage
	EnvelopeStorage storage.EnvelopeStorage
}

// NewService creates a Emitter
func (o Options) NewService() *Service {
	return &Service{
		BaseService: services.NewBaseService("core", o.Options),
		o:           o,
	}
}

// Service is in charge of emitting envelopes to Actors
type Service struct {
	services.BaseService

	processLogEntries chan *processLogEntries

	o Options

	coreEvents *api.CoreEventsHandler
}

// Startup starts the service
func (s *Service) Startup() error {
	s.Log.Noticef("Startup")
	s.processLogEntries = make(chan *processLogEntries, 10)
	if err := s.initProcessLogEntryHandler(); err != nil {
		close(s.processLogEntries)
		return err
	}
	s.coreEvents = api.NewCoreEventsHandler(context.Background(), s.Conn, nil)
	s.o.ActorStorage.SetEventHandlers(s.onActorUp, s.onActorDown)
	s.o.ProcessStorage.SetEventHandlers(s.onProcessStatusChange)
	s.o.ProcessStorage.SetProcessLogEntryHandler(s.processLogEntryHandler)
	s.o.EnvelopeStorage.SetEventHandler(s.onEnvelope)
	return nil
}

// Shutdown stops the service
func (s *Service) Shutdown() error {
	s.o.ActorStorage.SetEventHandlers(nil, nil)
	s.o.ProcessStorage.SetEventHandlers(nil)
	s.o.ProcessStorage.SetProcessLogEntryHandler(nil)
	s.o.EnvelopeStorage.SetEventHandler(nil)
	close(s.processLogEntries)
	return s.BaseService.Shutdown()
}

func (s *Service) onActorUp(actor api.Actor) {
	s.coreEvents.ActorPublish(api.CoreEvent{
		Name: "up",
		Actor: &api.Actor{
			ID:    actor.ID,
			Name:  actor.Name,
			Kind:  actor.Kind,
			Roles: actor.Roles,
		},
	})
}

func (s *Service) onActorDown(actor api.Actor) {
	s.coreEvents.ActorPublish(api.CoreEvent{
		Name: "down",
		Actor: &api.Actor{
			ID:    actor.ID,
			Name:  actor.Name,
			Kind:  actor.Kind,
			Roles: actor.Roles,
		},
	})
}

func (s *Service) onProcessStatusChange(process storage.Process, oldStatus api.Process_Status, status api.Process_Status) {
	if err := s.coreEvents.ProcessPublish(
		api.CoreEvent{
			Name: "status",
			Process: &api.Process{
				ID:                process.ID,
				TriggerEmitterID:  process.TriggerEmitterID,
				TriggerEnvelopeID: process.TriggerEnvelopeID,
				Status:            status,
			},
			ProcessOldStatus: oldStatus,
		},
	); err != nil {
		s.Log.Errorf("Error publishing ProcessEvent: %s", err)
	}
}

func (s *Service) onNodesReady(event api.ProcessNodeReadyEvent) {
	if err := s.coreEvents.ProcessNodeReadyPublish(event); err != nil {
		s.Log.Errorf("Error publising ProcessNodeReady: %s", err)
	}
}

func (s *Service) onEnvelope(id api.UUID, status storage.EnvelopeStorageReceptionStatus, newData bool) {
	// TODO make EnvelopePublish parametrizable with subject args so a
	// subscriber can listen to specific envelopes. This means supporting
	// subject params on request-less nrpc functions
	if err := s.coreEvents.EnvelopePublish(api.EnvelopeEvent{
		EnvelopeID: id,
		Status:     api.EnvelopeEvent_Status(status), // We know the enums are compatible
		NewData:    newData,
	}); err != nil {
		s.Log.Errorf("Error publising Envelope event: %s", err)
	}
}
