package director

import (
	"sync"

	"xbus.io/go-xbus/api"
)

// EnvelopeEventHandler is a function that handles a api.EnvelopeEvent
type EnvelopeEventHandler func(api.EnvelopeEvent)

// Subscription is returned by Watch
type Subscription struct {
	watcher *EnvelopeWatcher
	fn      EnvelopeEventHandler
	ids     []api.UUID
}

// Unsubscribe stops the watcher
func (s *Subscription) Unsubscribe() {
	s.watcher.Unsubscribe(s)
}

// EnvelopeWatcher can monitor status changes on several envelopes
type EnvelopeWatcher struct {
	subscriptions []*Subscription

	subscriptionIndex map[api.UUID][]*Subscription

	m sync.Mutex
}

// NewEnvelopeWatcher initialize a EnvelopeWatcher
func NewEnvelopeWatcher() *EnvelopeWatcher {
	return &EnvelopeWatcher{
		subscriptionIndex: make(map[api.UUID][]*Subscription),
	}
}

// PushEvent pushes an event
func (w *EnvelopeWatcher) PushEvent(event api.EnvelopeEvent) {
	w.m.Lock()
	var subs = w.subscriptionIndex[event.EnvelopeID]
	w.m.Unlock()

	for _, s := range subs {
		s.fn(event)
	}
}

// Watch starts watching for envelope status change and calls the given function
// on each event
func (w *EnvelopeWatcher) Watch(handler EnvelopeEventHandler, ids ...api.UUID) *Subscription {
	w.m.Lock()
	sub := Subscription{watcher: w, fn: handler, ids: ids}
	w.subscriptions = append(w.subscriptions, &sub)
	for _, id := range ids {
		w.subscriptionIndex[id] = append(w.subscriptionIndex[id], &sub)
	}
	w.m.Unlock()
	return &sub
}

func removeSubscription(slice []*Subscription, sub *Subscription) []*Subscription {
	for i, s := range slice {
		if s == sub {
			return append(slice[:i], slice[i+1:]...)
		}
	}
	return slice
}

// Unsubscribe disable a watcher
func (w *EnvelopeWatcher) Unsubscribe(sub *Subscription) {
	w.m.Lock()
	w.subscriptions = removeSubscription(w.subscriptions, sub)
	for _, id := range sub.ids {
		w.subscriptionIndex[id] = removeSubscription(w.subscriptionIndex[id], sub)
	}
	w.m.Unlock()
}
