// Copyright 2016 Orus.io. All rights reserved.

package envelopestorage

import (
	"context"
	"fmt"
	"time"

	nrpc "github.com/nats-rpc/gogo-nrpc"

	"xbus.io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

const (
	// ServiceName name of the service
	ServiceName = "envelopestorage"

	// QueueName queue name for queuesubscribes
	QueueName = "xbus.services." + ServiceName
)

var (
	// EnvelopeStallDelay is the delay after which an envelope is marked stalled
	// when its emitter stops sending fragments.
	// TODO in a later version, this value should be envelope-dependant and be
	// set by the emitter itself (within reasonnable limits)
	EnvelopeStallDelay = 30 * time.Second
)

// Options is used to create a 'envelopestorage' service
type Options struct {
	services.Options

	EnvelopeStorage storage.EnvelopeStorage
	JobStorage      storage.JobStorage
}

// NewService creates a Service
func (o Options) NewService() *Service {
	return &Service{
		BaseService: services.NewBaseService(ServiceName, o.Options),
		o:           o,
	}
}

// Service is in charge of emitting process status events to various subjects
type Service struct {
	services.BaseService

	workers *nrpc.WorkerPool

	o Options
}

// Startup starts the service
func (s *Service) Startup() error {
	// TODO make the pool settings changeable by the xbusd configuration
	s.workers = nrpc.NewWorkerPool(context.Background(), 100, 100, 1*time.Minute)

	envelopeStorageHandler := api.NewEnvelopeStorageConcurrentHandler(
		s.workers, s.Conn, envelopeStorageServer{s},
	)
	return s.SubscribeAll(
		s.HandlerQueueSubscribe(envelopeStorageHandler, QueueName),
	)
}

// Shutdown stops the service
func (s *Service) Shutdown() error {
	err := s.BaseService.Shutdown()
	s.workers.Close(time.Second)
	return err
}

type envelopeStorageServer struct {
	*Service
}

func (s envelopeStorageServer) Retrieve(
	ctx context.Context, clientKind string, clientID string,
	req api.EnvelopePosition, send func(api.EnvelopeFragmentAndPosition),
) error {
	// TODO verify credentials
	if clientKind == "actor" {
		actorID, err := api.UUIDFromString(clientID)
		if err != nil {
			return err
		}
		var allowed = false
		for _, job := range s.o.JobStorage.ListRunningJobByEnvelopeID(req.EnvelopeID) {
			if job.Node.ActorID == actorID {
				allowed = true
			}
		}
		if !allowed {
			return fmt.Errorf("Permission denied")
		}
	}
	pos := storage.DecodePosition(req)
	for !pos.Complete {
		if ctx.Err() != nil {
			return ctx.Err()
		}
		fragment, newpos, err := s.o.EnvelopeStorage.ReadEnvelope(
			req.EnvelopeID, pos, 1024*512)
		if err != nil {
			s.Log.Noticef("Error reading envelope %s: Error :-(", req.EnvelopeID)
			panic(err)
		}
		// TODO add a blocking version of readenvelope that waits until
		// there is something to return
		if newpos.Equals(pos) {
			s.o.EnvelopeStorage.CheckStalled(req.EnvelopeID, EnvelopeStallDelay)
			time.Sleep(1 * time.Second)
		} else {
			pos = newpos
			send(api.EnvelopeFragmentAndPosition{
				Fragment: &fragment,
				Position: storage.EncodePosition(req.EnvelopeID, pos),
			})
		}
	}
	return nil
}
