// Copyright 2019 Orus.io. All rights reserved.

package metrics

import (
	"context"
	"sync"

	"xbus.io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/services/tools"
	"github.com/prometheus/client_golang/prometheus"
)

func min(a, b uint) uint {
	if a < b {
		return a
	}
	return b
}

const (
	// MetricsQueue is the name we use to queueSub to a queue with the metrics
	MetricsQueue = "metrics_queue"
)

// Options is used to create a Service
type Options struct {
	services.Options
}

// NewService returns a new Service
func (o Options) NewService() *Service {
	return &Service{
		BaseService: services.NewBaseService("metrics", o.Options),
	}
}

// New returns a new Service
func New(options Options) *Service {
	return options.NewService()
}

// Service is the metrics service implementation
// The metrics service is responsible for keeping track of global metrics
// of the bus
type Service struct {
	services.BaseService
	Opts Options

	ctx      context.Context
	stop     func()
	wg       sync.WaitGroup
	api      *api.Client
	actorMap *tools.ActorMap
}

// Startup ...
func (s *Service) Startup() error {
	s.ctx, s.stop = context.WithCancel(context.Background())
	s.api = api.NewClient(s.Conn)
	if m, err := tools.NewActorMap(s.ctx, s.api, s.Log); err == nil {
		s.actorMap = m
	} else {
		return err
	}
	s.wg.Add(1)
	go s.run()
	return nil
}

func (s *Service) run() {
	defer s.wg.Done()
	for _, actorState := range s.actorMap.List() {
		updateActorOnlineStatus(actorState)
	}
	for {
		select {
		case actorState := <-s.actorMap.Events:
			updateActorOnlineStatus(actorState)
		case <-s.ctx.Done():
			return
		}
	}
}

// Shutdown the service
func (s *Service) Shutdown() error {
	s.stop()
	s.wg.Wait()
	s.actorMap.Close()
	return nil
}

// metrics
var actorOnlineStatus = prometheus.NewGaugeVec(prometheus.GaugeOpts{
	Name: "xbus_server_actor_online_status",
	Help: "Consumer/worker online status. 0 if the actor is offline, 1 if the actor is online",
}, []string{"name"})

func updateActorOnlineStatus(actorState tools.ActorState) {
	if actorState.Kind != api.Actor_EMITTER {
		var value float64
		if actorState.Online {
			value = 1.0
		}
		actorOnlineStatus.WithLabelValues(actorState.Name).Set(value)
	}
}

func init() {
	prometheus.MustRegister(actorOnlineStatus)
}
