// Copyright 2016 Orus.io. All rights reserved.

package processstate

import (
	"context"
	"fmt"

	"xbus.io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

const (
	// ServiceName name of the service
	ServiceName = "processstate"

	// QueueName queue name for queuesubscribes
	QueueName = "xbus.services." + ServiceName
)

// Options is used to create a 'processstate' service
type Options struct {
	services.Options

	LogStorage      storage.LogStorage
	ProcessManager  *process.Manager
	EnvelopeStorage storage.EnvelopeStorage
	ActorStorage    storage.ActorStorage
}

// NewService creates a Service
func (o Options) NewService() *Service {
	return &Service{
		BaseService: services.NewBaseService(ServiceName, o.Options),
		o:           o,
	}
}

// Service is in charge of emitting process status events to various subjects
type Service struct {
	services.BaseService

	o Options

	api          *api.Client
	processState *api.ProcessStateHandler
}

// Startup starts the service
func (s *Service) Startup() error {
	s.processState = api.NewProcessStateHandler(
		context.Background(), s.Conn, s,
	)
	s.processState.SetEncodings([]string{"protobuf", "json"})

	s.api = api.NewClient(s.Conn)
	sub, err := s.api.CoreEvents.ProcessSubscribe(s.handleCoreEventsProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	return s.SubscribeAll(
		s.HandlerQueueSubscribe(s.processState, QueueName),
	)
}

func (s *Service) handleCoreEventsProcess(evt api.CoreEvent) {
	s.TaskAdd()
	defer s.TaskDone()

	envelopeState := api.EmitterEnvelopeState{
		ProcessID: evt.Process.ID,
		Status:    evt.Process.Status,
	}

	for _, entry := range s.o.LogStorage.GetProcessMessages(
		evt.Process.ID, api.LogLevel_ERROR,
	) {
		s.Log.Noticef("%+v", entry)
		envelopeState.Errors = append(envelopeState.Errors, entry.Message)
	}

	// load the trigger envelope head
	envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
		evt.Process.TriggerEnvelopeID, storage.EnvelopeStartPos(), 100*1024,
	)
	if err != nil {
		s.Log.Errorf("Could not read process %s trigger envelope %s. Got %s",
			evt.Process.ID, evt.Process.TriggerEnvelopeID, err,
		)
	} else {
		envelopeState.TriggerEnvelope = &api.EnvelopeFragmentAndPosition{
			Fragment: &envelope,
			Position: storage.EncodePosition(evt.Process.TriggerEnvelopeID, position),
		}
	}

	if evt.Process.Status == api.Process_DONE {
		// Check for a response
		p, err := s.o.ProcessManager.Get(evt.Process.ID)
		if err != nil {
			panic(err)
		}
		responseID := p.GetResponse()
		if !responseID.IsZero() {
			envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
				responseID, storage.EnvelopeStartPos(), 512*1024,
			)
			if err != nil {
				s.Log.Errorf("Could not read process %s response envelope %s. Got %s",
					evt.Process.ID, responseID, err,
				)
			}
			envelopeState.Response = &api.EnvelopeFragmentAndPosition{
				Fragment: &envelope,
				Position: storage.EncodePosition(responseID, position),
			}
		}
	}

	s.Log.Noticef("Sending envelope state: %v", envelopeState)
	if err := s.processState.EnvelopeStatesPublish(
		evt.Process.TriggerEmitterID.String(),
		evt.Process.TriggerEnvelopeID.String(),
		envelopeState,
	); err != nil {
		s.Log.Errorf("Error sending process state: %s", err)
	}
}

// GetEnvelopeState implements ProcessState.GetEnvelopeState
func (s *Service) GetEnvelopeState(
	ctx context.Context, emitterID string, req api.GetEnvelopeStateRequest,
) (
	api.EmitterEnvelopeState, error,
) {
	emitterUUID, err := api.UUIDFromString(emitterID)
	if err != nil {
		return api.EmitterEnvelopeState{}, fmt.Errorf("Invalid emitterID: %s", err)
	}
	emitter, ok := s.o.ActorStorage.Get(emitterUUID)
	if !ok {
		return api.EmitterEnvelopeState{}, fmt.Errorf("Unknown emitter: %s", err)
	}
	process, err := s.o.ProcessManager.GetByTrigger(emitter, req.ID)

	envelopeState := api.EmitterEnvelopeState{
		ProcessID:        process.ID,
		ProcessCreatedAt: &process.CreatedAt,
		Status:           process.GetStatus(),
	}

	// load the trigger envelope head
	envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
		process.TriggerEnvelopeID, storage.EnvelopeStartPos(), 100*1024,
	)
	if err != nil {
		s.Log.Errorf("Could not read process %s trigger envelope %s. Got %s",
			process.ID, process.TriggerEnvelopeID, err,
		)
	} else {
		envelopeState.TriggerEnvelope = &api.EnvelopeFragmentAndPosition{
			Fragment: &envelope,
			Position: storage.EncodePosition(process.TriggerEnvelopeID, position),
		}
	}

	if envelopeState.Status == api.Process_DONE {
		// Check for a response
		responseID := process.GetResponse()
		if !responseID.IsZero() {
			envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
				responseID, storage.EnvelopeStartPos(), 100*1024,
			)
			if err != nil {
				s.Log.Errorf("Could not read process %s response envelope %s. Got %s",
					process.ID, responseID, err,
				)
			}
			envelopeState.Response = &api.EnvelopeFragmentAndPosition{
				Fragment: &envelope,
				Position: storage.EncodePosition(responseID, position),
			}
		}
	}

	if envelopeState.Status == api.Process_ERROR {
		for _, entry := range s.o.LogStorage.GetProcessMessages(
			envelopeState.ProcessID, api.LogLevel_ERROR,
		) {
			s.Log.Noticef("%+v", entry)
			envelopeState.Errors = append(envelopeState.Errors, entry.Message)
		}
	}

	return envelopeState, nil
}

// AckResult mark a process result as acked
func (s *Service) AckResult(
	ctx context.Context, emitterID string, req api.AckResultRequest,
) error {
	emitterUUID, err := api.UUIDFromString(emitterID)
	if err != nil {
		return fmt.Errorf("Invalid emitterID: %s", err)
	}
	_, ok := s.o.ActorStorage.Get(emitterUUID)
	if !ok {
		return fmt.Errorf("Unknown emitter: %s", err)
	}
	process, err := s.o.ProcessManager.Get(req.ProcessID)
	if err != nil {
		return err
	}
	if process.TriggerEmitterID != emitterUUID {
		return fmt.Errorf("This emitter has no such process ID")
	}
	s.o.ProcessManager.ProcessStorage.AckResult(req.ProcessID)

	return nil
}

// List the process of this emitter
func (s *Service) List(
	ctx context.Context, emitterID string, req api.ProcessFilter,
	send func(api.EmitterEnvelopeStateList),
) error {
	s.TaskAdd()
	defer s.TaskDone()

	req.TriggerEmitterID = []api.UUID{
		api.MustUUIDFromString(emitterID),
	}

	pList := s.o.ProcessManager.ProcessStorage.Query(req)
	select {
	case <-ctx.Done():
		return nil
	default:
	}
	var batch []*api.EmitterEnvelopeState
	for _, p := range pList {
		state := s.o.ProcessManager.ProcessStorage.GetState(p.ID)
		select {
		case <-ctx.Done():
			return nil
		default:
		}
		envelopeState := api.EmitterEnvelopeState{
			ProcessID: p.ID,
			Status:    state.Status,
		}
		envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
			p.TriggerEnvelopeID, storage.EnvelopeStartPos(), 512*1024,
		)
		if err != nil {
			s.Log.Errorf("Could not read process %s trigger envelope %s. Got %s",
				p.ID, p.TriggerEnvelopeID, err,
			)
		} else {
			envelopeState.TriggerEnvelope = &api.EnvelopeFragmentAndPosition{
				Fragment: &envelope,
				Position: storage.EncodePosition(p.TriggerEnvelopeID, position),
			}
		}
		if envelopeState.Status == api.Process_DONE {
			// Check for a response
			responseID := state.ResponseEnvelopeID
			if !responseID.IsZero() {
				envelope, position, err := s.o.EnvelopeStorage.ReadEnvelope(
					responseID, storage.EnvelopeStartPos(), 512*1024,
				)
				if err != nil {
					s.Log.Errorf("Could not read process %s response envelope %s. Got %s",
						p.ID, responseID, err,
					)
				}
				envelopeState.Response = &api.EnvelopeFragmentAndPosition{
					Fragment: &envelope,
					Position: storage.EncodePosition(responseID, position),
				}
			}
		}
		if envelopeState.Status == api.Process_ERROR {
			for _, entry := range s.o.LogStorage.GetProcessMessages(
				envelopeState.ProcessID, api.LogLevel_ERROR,
			) {
				s.Log.Noticef("%+v", entry)
				envelopeState.Errors = append(envelopeState.Errors, entry.Message)
			}
		}
		batch = append(batch, &envelopeState)
		if len(batch) == 200 {
			send(api.EmitterEnvelopeStateList{List: batch})
			batch = nil
		}
	}
	if batch != nil {
		send(api.EmitterEnvelopeStateList{List: batch})
	}
	return nil
}
