// Copyright 2016 Orus.io. All rights reserved.

// Package ravencapture provides the 'core' service, which must be started once and only once
// in every server process
package ravencapture

import (
	"context"
	"fmt"
	"strconv"
	"time"

	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/api/control"
	"bitbucket.org/orus-io/xbus/services"
	raven "github.com/getsentry/raven-go"
	yaml "gopkg.in/yaml.v2"
)

func convertMapsKeyToString(m map[interface{}]interface{}) map[string]interface{} {
	var res = make(map[string]interface{})

	for key, value := range m {
		if mv, ok := value.(map[interface{}]interface{}); ok {
			value = convertMapsKeyToString(mv)
		}
		switch k := key.(type) {
		case string:
			res[k] = value
		default:
			res[fmt.Sprintf("%v", k)] = value
		}
	}

	return res
}

func makeJSONReady(value interface{}) interface{} {
	switch value := value.(type) {
	case map[interface{}]interface{}:
		var res = make(map[string]interface{})

		for key, item := range value {
			item = makeJSONReady(item)
			switch k := key.(type) {
			case string:
				res[k] = item
			default:
				res[fmt.Sprintf("%v", k)] = item
			}
		}

		return res
	case []interface{}:
		var res []interface{}
		for _, item := range value {
			item = makeJSONReady(item)
			res = append(res, item)
		}
		return res
	default:
		return value
	}
}

func extractVar(value interface{}, path ...string) interface{} {
	if len(path) == 0 {
		return value
	}
	key := path[0]
	subPath := path[1:]
	switch value := value.(type) {
	case map[string]interface{}:
		return extractVar(value[key], subPath...)
	case map[interface{}]interface{}:
		return extractVar(value[key], subPath...)
	case []interface{}:
		iKey, err := strconv.Atoi(key)
		if err != nil {
			return nil
		}
		return value[iKey]
	default:
		return nil
	}
}

func extractVarString(value interface{}, path ...string) string {
	s, _ := extractVar(value, path...).(string)
	return s
}

// Options is used to create a 'ravencapture' service
type Options struct {
	services.Options

	RavenClient *raven.Client
}

// NewService creates a Emitter
func (o Options) NewService() *Service {
	return &Service{
		BaseService: services.NewBaseService("ravencapture", o.Options),
		o:           o,
		actorNames:  make(map[api.UUID]string),
	}
}

// Service is in charge of emitting envelopes to Actors
type Service struct {
	services.BaseService

	o Options

	actorNames map[api.UUID]string
}

func (s *Service) refreshActorNames() {
	// We may prefer to use a ActorMap, but I have a feeling its overkill
	// as the actor names are not likely to change
	client := api.NewClient(s.Conn)
	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()
	if err := client.Sentinel.ListActors(ctx,
		func(ctx context.Context, reply api.ClientGetActorsReply) {
			for _, actor := range reply.Actors {
				s.actorNames[actor.ID] = actor.Name
			}
		},
	); err != nil {
		s.Log.Errorf("Failed to refresh the actor names list: %s", err)
	}
}

func (s *Service) getActorName(id api.UUID) string {
	if name, ok := s.actorNames[id]; ok {
		return name
	}
	s.refreshActorNames()
	if name, ok := s.actorNames[id]; ok {
		return name
	}
	return id.String()
}

// Startup starts the service
func (s *Service) Startup() error {
	s.Log.Noticef("Startup")
	api := api.NewClient(s.Conn)
	sub, err := api.CoreEvents.ProcessSubscribe(s.handleCoreEventsProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)
	return nil
}

func (s *Service) handleCoreEventsProcess(evt api.CoreEvent) {
	if evt.Process.Status == api.Process_ERROR {
		c := control.NewProcessClient(s.Conn)

		var buf []byte

		if err := c.ExportStream(
			context.Background(),
			control.ProcessExportRequest{
				ProcessIDs:        []api.UUID{evt.Process.ID},
				ExportEnvelope:    true,
				ExportJobs:        false,
				ExportLogs:        true,
				ExportProcessLogs: true,
				EnvelopeSizeLimit: 1024 * 10,
				LogLevel:          api.LogLevel_NOTICE,
			},
			func(ctx context.Context, res control.OctetStreamList) {
				for _, chunk := range res.GetChunks() {
					buf = append(buf, chunk.Data...)
				}
			},
		); err != nil {
			s.Log.Errorf("Could not export process %s: %s", evt.Process.ID, err)
			return
		}

		var data []map[string]interface{}

		if err := yaml.Unmarshal(buf, &data); err != nil {
			s.Log.Errorf("Process export could not be parsed yaml: %s", err)
			return
		}

		if len(data) != 1 {
			s.Log.Errorf("Expected 1 process in the export, got %d", len(data))
			return
		}

		type processlogs struct {
			Logs []api.LogEntry
		}
		var logs []processlogs

		if err := yaml.Unmarshal(buf, &logs); err != nil {
			fmt.Println(err)
			s.Log.Noticef("Could not read logs in the export: %s", err)
			return
		}
		fmt.Println(logs)

		var (
			tags   = make(map[string]string)
			extra  = make(map[string]interface{})
			reason = "N/A"
		)

		tags["xbus.emitter"] = extractVarString(data[0], "trigger", "emitter", "name")

		if len(logs) == 1 {
			for _, log := range logs[0].Logs {
				if log.Message.Level == api.LogLevel_ERROR {
					tags["xbus.actor"] = s.getActorName(log.ActorID)
					reason = log.Message.Text
					break
				}
			}
		}
		if reason == "N/A" {
			if value := extractVarString(data[0], "state", "statusreason"); value != "" {
				reason = value
			} else {
				s.Log.Errorf("Could not extract StatusReason")
			}
		}

		if value := extractVarString(data[0], "pipeline", "pipelineinfo", "name"); value != "" {
			tags["xbus.pipeline.name"] = value
		}
		if value := extractVarString(data[0], "pipeline", "pipelineinfo", "version"); value != "" {
			tags["xbus.pipeline.version"] = value
		}

		for name, data := range data[0] {
			data = makeJSONReady(data)
			if name == "id" {
				name = "process_id"
			}
			name = "xbus." + name
			extra[name] = data
		}

		text := "[process]"
		if tags["xbus.pipeline.name"] != "" {
			text += fmt.Sprintf("[%s-%s]",
				tags["xbus.pipeline.name"], tags["xbus.pipeline.version"],
			)
		} else {
			text += fmt.Sprintf("[from %s]", tags["xbus.emitter"])
		}
		text += " " + reason
		message := raven.Message{
			Message: text,
			Params:  []interface{}{evt.Process.ID},
		}
		packet := raven.NewPacketWithExtra(
			fmt.Sprintf(message.Message, message.Params...),
			extra,
			&message)

		_, ch := s.o.RavenClient.Capture(packet, tags)
		go func() {
			if err := <-ch; err != nil {
				s.Log.Errorf("Sending to sentry failed: %s", err)
			}
		}()
	}
}
