package storage

import (
	"fmt"
	"sort"
	"strings"

	"gopkg.in/yaml.v2"

	"xbus.io/go-xbus/api"
)

func (s ReceptionStatus) String() string {
	switch s {
	case ReceptionPending:
		return "pending"
	case ReceptionOngoing:
		return "ongoing"
	case ReceptionDone:
		return "done"
	case ReceptionClosed:
		return "closed"
	default:
		return ""
	}
}

// UnmarshalText parses a ReceptionStatus
func (s *ReceptionStatus) UnmarshalText(data []byte) error {
	switch string(data) {
	case "pending":
		*s = ReceptionPending
	case "ongoing":
		*s = ReceptionOngoing
	case "done":
		*s = ReceptionDone
	case "closed":
		*s = ReceptionClosed
	default:
		return fmt.Errorf("Invalid ReceptionStatus: %s", string(data))
	}
	return nil
}

// MarshalText serialize a ReceptionStatus to text
func (s ReceptionStatus) MarshalText() ([]byte, error) {
	return []byte(s.String()), nil
}

// String serialize a ProcessNode
func (n ProcessNode) String() string {
	return n.ActorID.String() + "/" + n.NodeID
}

// ProcessNodeFromString deserialize a ProcessNode
func ProcessNodeFromString(s string) (ProcessNode, error) {
	var (
		n   ProcessNode
		err error
	)
	splitted := strings.Split(s, "/")
	if len(splitted) != 2 {
		return n, fmt.Errorf("Invalid processnode: %s. Expects '<actoruuid>/<nodeid>'", s)
	}
	n.ActorID, err = api.UUIDFromString(splitted[0])
	if err != nil {
		return n, err
	}
	n.NodeID = splitted[1]
	return n, nil
}

// MarshalText implements encoding.TextMarshaler for ProcessNode
func (n ProcessNode) MarshalText() ([]byte, error) {
	return []byte(n.String()), nil
}

// UnmarshalText implements encoding.TextUnmarshaler for ProcessNode
func (n *ProcessNode) UnmarshalText(text []byte) (err error) {
	*n, err = ProcessNodeFromString(string(text))
	return
}

// Fmt returns a formatted string of the state
func (s ProcessState) Fmt() string {
	b, err := yaml.Marshal(s)
	if err != nil {
		panic(err)
	}
	return string(b)
}

type processNodeStateMapSlice yaml.MapSlice

func (ms processNodeStateMapSlice) Len() int {
	return len(ms)
}

func (ms processNodeStateMapSlice) Less(i, j int) bool {
	return ms[i].Key.(string) < ms[j].Key.(string)
}

func (ms processNodeStateMapSlice) Swap(i, j int) {
	ms[i], ms[j] = ms[j], ms[i]
}

// MarshalYAML as an ordered dict
func (m ProcessNodeStateMap) MarshalYAML() (interface{}, error) {
	var items = make(processNodeStateMapSlice, len(m))
	i := 0
	for key, value := range m {
		items[i].Key = key.String()
		items[i].Value = value
		i++
	}
	sort.Sort(items)
	return items, nil
}
