// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"database/sql"
	"errors"
	"fmt"
	"strings"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	xbus "xbus.io/go-xbus"
	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

// NewEnvelopeStorage initialize a EnvelopeStorage
func NewEnvelopeStorage(model *model.Model, db *yago.DB, logger xbus.Logger) *EnvelopeStorage {
	return &EnvelopeStorage{
		Base{
			model:  model,
			db:     db,
			logger: xbus.NewPrefixedLogger("[storage/sql/envelop]", logger),
		},
		nil}
}

// EnvelopeStorage implements the storage.EnvelopeStorage interface
type EnvelopeStorage struct {
	Base
	eventHandler storage.EnvelopeStorageEventHandler
}

// SetEventHandler set the current handler
func (es *EnvelopeStorage) SetEventHandler(handler storage.EnvelopeStorageEventHandler) {
	es.eventHandler = handler
}

func (es EnvelopeStorage) emitEvent(
	id api.UUID, status storage.EnvelopeStorageReceptionStatus, newdata bool) {
	if es.eventHandler != nil {
		es.eventHandler(id, status, newdata)
	}
}

// GetEventTypes returns the eventtypes of an envelope, or nil if unknown
// exists is false if the envelope is unknown, true if exists.
// will panic on any SQL error.
func (es EnvelopeStorage) GetEventTypes(id api.UUID) (types []string, exists bool) {
	tx := es.MustBegin()
	defer tx.Rollback()

	var (
		dbEnvelope model.Envelope
	)

	err := tx.Query(es.model.Envelope).Get(&dbEnvelope, uuid.UUID(id))
	if err == yago.ErrRecordNotFound {
		return nil, false
	}
	if err != nil {
		panic(err)
	}
	if !dbEnvelope.TypesKnown {
		return nil, true
	}
	rows, err := tx.Query(es.model.Event).
		Select(es.model.Event.EventType).
		Where(es.model.Event.EnvelopeID.Eq(uuid.UUID(id))).
		SQLQuery()
	if err != nil {
		panic(err)
	}
	var eventTypes []string

	for rows.Next() {
		var eventType string
		must(
			rows.Scan(&eventType))
		eventTypes = append(eventTypes, eventType)
	}
	return eventTypes, true
}

// StoreEnvelope stores an envelope chunk and returns the envelope state in the
// storage
func (es EnvelopeStorage) StoreEnvelope(envelope api.Envelope) (storage.EnvelopeState, error) {
	var (
		dbEnvelope model.Envelope
		state      storage.EnvelopeState
	)

	tx := es.MustBegin()
	defer tx.RollbackIfOpened()
	// Load the envelope or create it if it is a new one
	err := tx.Query(es.model.Envelope).Get(&dbEnvelope, uuid.UUID(envelope.ID))
	if err == yago.ErrRecordNotFound {
		dbEnvelope = model.Envelope{
			Base: model.Base{
				ID: uuid.UUID(envelope.ID),
			},
			TypesKnown: false,
			Status:     storage.EnvelopeStorageReceptionReceiving,
		}
		err = tx.Insert(&dbEnvelope)
	}
	if err != nil {
		panic(err)
	}

	// Load the current state of the envelope

	state.ID = envelope.ID
	state.TypesKnown = dbEnvelope.TypesKnown

	switch dbEnvelope.Status {
	case storage.EnvelopeStorageReceptionComplete:
		err = fmt.Errorf("Envelope '%s' is already complete", envelope.ID)
		es.logger.Errorf(err.Error())
		return storage.EnvelopeState{}, err
	case storage.EnvelopeStorageReceptionError:
		state.Status = storage.EnvelopeStorageReceptionError
		state.Err = errors.New(dbEnvelope.Error)
		return state, nil
	case storage.EnvelopeStorageReceptionReceiving:
		state.Status = storage.EnvelopeStorageReceptionReceiving
	case storage.EnvelopeStorageReceptionStalled:
	default:
		return storage.EnvelopeState{},
			fmt.Errorf("Invalid status: %s", dbEnvelope.Status)
	}

	// Check if the event ids list is known or in the envelope
	var eventIDs []api.UUID
	if dbEnvelope.EventIDs != "" {
		for _, sid := range strings.Split(dbEnvelope.EventIDs, ",") {
			if sid == "" {
				continue
			}
			eventIDs = append(eventIDs, api.MustUUIDFromString(sid))
		}
	} else if len(envelope.EventIDs) != 0 {
		var sids []string
		for _, id := range envelope.EventIDs {
			sids = append(sids, id.String())
		}
		dbEnvelope.EventIDs = strings.Join(sids, ",")
		eventIDs = envelope.EventIDs
	}

	// Load the Event & their validators

	var (
		dbEvents []model.Event
	)

	must(
		tx.Query(es.model.Event).
			Where(es.model.Event.EnvelopeID.Eq(dbEnvelope.ID)).
			All(&dbEvents))

	var (
		dbEventByID     = make(map[uuid.UUID]model.Event)
		eventValidators = make(map[uuid.UUID]tools.EventValidator)
		eventStates     = make(map[uuid.UUID]storage.EventState)
		newData         = false
	)

	for _, dbEvent := range dbEvents {
		var availableItems uint64
		if dbEvent.Validator != "" {
			validator, err := tools.LoadEventValidator(dbEvent.Validator)
			if err != nil {
				panic(err)
			}
			eventValidators[dbEvent.EventID] = validator
			availableItems = validator.CurrentItemCount
		}
		dbEventByID[dbEvent.EventID] = dbEvent
		eventStates[dbEvent.EventID] = storage.EventState{
			ID:             api.UUID(dbEvent.EventID),
			Type:           dbEvent.EventType,
			AvailableItems: availableItems,
			TotalItems:     dbEvent.ItemCount,
			Status:         dbEvent.Status,
		}
	}

	// Now load the incoming events into the database.
	// If the batch n° is consistent with the validator current
	// state, we feed it immediately to it, followed by any
	// waiting batch if their batch n° match.
	// If not, only store the items to the db.

	for _, event := range envelope.Events {
		dbEvent, ok := dbEventByID[uuid.UUID(event.ID)]
		if !ok {
			dbEvent = model.Event{
				EventID:    uuid.UUID(event.ID),
				EnvelopeID: dbEnvelope.ID,
				Status:     storage.EnvelopeStorageReceptionReceiving,
				ItemCount:  event.ItemCount,
				Checksum:   event.Checksum,
			}
			if event.Type != "" {
				dbEvent.EventType = event.Type
			}
			must(
				tx.Insert(&dbEvent))
		}
		eventState, ok := eventStates[dbEvent.EventID]
		if !ok {
			eventState = storage.EventState{
				ID:             api.UUID(dbEvent.EventID),
				Type:           event.Type,
				AvailableItems: 0,
				TotalItems:     event.ItemCount,
				Status:         storage.EnvelopeStorageReceptionReceiving,
			}
		}

		// if no item in the event chunk, we need to write something
		// in the items table so batch numbers remain consistent
		// and ReadEnvelope can see empty batches and jump to non-empty ones
		if len(event.Items) == 0 {
			dbItem := model.Item{
				EventID: dbEvent.ID,
				Batch:   event.Index,
				Index:   -1,
			}
			must(
				tx.Insert(&dbItem))
		} else {
			newData = true

			// else store items in the database
			for i, item := range event.Items {
				dbItem := model.Item{
					EventID: dbEvent.ID,
					Batch:   event.Index,
					Index:   int64(i),
					Data:    item,
				}
				must(
					tx.Insert(&dbItem))
			}
		}

		// if possible, update the validation state
		validator, _ := eventValidators[uuid.UUID(event.ID)]
		if event.Index == validator.CurrentIndex+1 {
			validator, err = validator.Update(event)
			must(err)

			validatorStr, err := tools.DumpEventValidator(validator)
			must(err)

			eventState.AvailableItems = validator.CurrentItemCount
			if validator.Complete {
				if validator.Valid {
					eventState.Status = storage.EnvelopeStorageReceptionComplete
				} else if validator.Error != nil {
					eventState.Status = storage.EnvelopeStorageReceptionError
					eventState.Err = validator.Error
				}
			}

			dbEvent.Validator = validatorStr
			dbEvent.ItemCount = validator.ItemCount
			dbEvent.Checksum = validator.Checksum
			dbEvent.Status = eventState.Status
			eventValidators[uuid.UUID(event.ID)] = validator
			dbEventByID[uuid.UUID(event.ID)] = dbEvent

			// TODO look for more items in the database and feed them
			// to the validator

			must(
				tx.Update(&dbEvent, "Validator", "ItemCount", "Checksum", "Status"))

		}

		eventStates[dbEvent.EventID] = eventState
	}

	for _, eventState := range eventStates {
		state.Events = append(state.Events, eventState)
	}

	if !state.TypesKnown && len(eventIDs) != 0 {
		if len(eventIDs) == len(state.Events) {
			foundUnknownType := false
			for _, eventID := range eventIDs {
				state, _ := eventStates[uuid.UUID(eventID)]
				if state.Type == "" {
					foundUnknownType = true
					break
				}
			}
			if !foundUnknownType {
				state.TypesKnown = true
				dbEnvelope.TypesKnown = true
			}
		}
	}

	if state.TypesKnown {
		foundNotComplete := false
		foundErrors := false
		for _, eventState := range eventStates {
			if eventState.Status != storage.EnvelopeStorageReceptionComplete {
				foundNotComplete = true
			}
			if eventState.Status == storage.EnvelopeStorageReceptionError {
				foundErrors = true
			}
		}
		if !foundNotComplete {
			state.Status = storage.EnvelopeStorageReceptionComplete
		} else if foundErrors {
			state.Status = storage.EnvelopeStorageReceptionError
		} else {
			state.Status = storage.EnvelopeStorageReceptionReceiving
		}
		dbEnvelope.Status = state.Status
	}

	if dbEnvelope.Status == storage.EnvelopeStorageReceptionStalled {
		dbEnvelope.Status = storage.EnvelopeStorageReceptionReceiving
	}
	must(tx.Update(&dbEnvelope))

	tx.MustCommit()

	es.emitEvent(state.ID, state.Status, newData)

	return state, nil
}

// CheckStalled mark the given envelope as stalled if no fragment was
// received in the given time _and_ the envelope is in state "Receiving"
// Returns true if the envelope is indeed stalled
// If the envelope is already marked stalled, the given time is ignored.
func (es EnvelopeStorage) CheckStalled(id api.UUID, delay time.Duration) bool {
	tx := es.MustBegin()
	defer tx.RollbackIfOpened()

	var dbEnvelope model.Envelope

	if err := tx.Query(es.model.Envelope).Get(&dbEnvelope, uuid.UUID(id)); err != nil {
		if err == yago.ErrRecordNotFound {
			return false
		}
		panic(err)
	}

	if dbEnvelope.Status == storage.EnvelopeStorageReceptionStalled {
		return true
	}
	if dbEnvelope.Status != storage.EnvelopeStorageReceptionReceiving {
		return false
	}
	row := tx.GetEngine().QueryRow(model.RawSQL(`
		UPDATE envelope
		SET status = 'stalled',
		    updated_at = $3
		WHERE status = 'receiving'
		  AND updated_at < $2
		  AND id=$1 RETURNING status;
	`, uuid.UUID(id), time.Now().Add(-delay), time.Now()))

	var status string

	err := row.Scan(&status)
	if sqlErr, ok := err.(qb.Error); ok && sqlErr.Orig == sql.ErrNoRows {
		// no match, we are not stalled
		return false
	} else if err != nil {
		fmt.Printf("%#v\n", err)
		panic(err)
	}

	if status != "stalled" {
		panic("Invalid status: " + status)
	}

	tx.MustCommit()

	// The status changed to stalled
	es.emitEvent(id, storage.EnvelopeStorageReceptionStalled, false)

	return true
}

// ReadEnvelope reads an envelope from the given position
func (es EnvelopeStorage) ReadEnvelope(id api.UUID, position storage.EnvelopeReadPos, maxSize int) (api.Envelope, storage.EnvelopeReadPos, error) {
	position = position.Clone()
	tx := es.MustBegin()
	defer tx.Rollback()

	var dbEnvelope model.Envelope
	// Load the envelope
	err := tx.Query(es.model.Envelope).Get(&dbEnvelope, uuid.UUID(id))
	if err == yago.ErrRecordNotFound {
		return api.Envelope{}, storage.EnvelopeReadPos{}, fmt.Errorf("No such envelope: %s", id)
	}
	must(err)

	if dbEnvelope.Status == storage.EnvelopeStorageReceptionError {
		return api.Envelope{}, position, fmt.Errorf("Envelope is in the error state: %s", dbEnvelope.Error)
	}

	// TODO if event types are not all known, return empty envelope
	// with unchanged position

	envelope := api.Envelope{
		ID: api.UUID(dbEnvelope.ID),
	}
	if dbEnvelope.EventIDs != "" {
		for _, sid := range strings.Split(dbEnvelope.EventIDs, ",") {
			if sid == "" {
				continue
			}
			envelope.EventIDs = append(envelope.EventIDs, api.MustUUIDFromString(sid))
		}
	}

	var dbEvents []model.Event

	must(
		tx.Query(es.model.Event).
			Where(es.model.Event.EnvelopeID.Eq(uuid.UUID(id))).
			All(&dbEvents))

	var estimateSize = 36 + len(dbEvents)*3

	allDone := true

	for _, dbEvent := range dbEvents {
		eventPos, _ := position.Events[api.UUID(dbEvent.EventID)]
		eventPos.Index++

		selectItems := es.model.Item.GetMapper().Table().
			Select(es.model.Item.Batch, es.model.Item.Index, es.model.Item.Data).
			Where(
				es.model.Item.EventID.Eq(dbEvent.ID),
				es.model.Item.Index.NotEq(-1),
			).
			OrderBy(es.model.Item.Batch.Column, es.model.Item.Index.Column)

		if eventPos.ItemCount != 0 {
			offset := eventPos.ItemCount - 1

			// XXX with huge amount of items, the offset cast to int can
			// theoraticaly overflow
			selectItems = selectItems.Offset(int(offset))
		}

		var emptyBatches []int
		{
			rows, err := tx.Query(es.model.Item).
				Select(es.model.Item.Batch).
				Where(
					es.model.Item.EventID.Eq(dbEvent.ID),
					es.model.Item.Index.Eq(-1)).
				OrderBy(es.model.Item.Batch).
				SQLQuery()

			must(err)

			for rows.Next() {
				var batch int
				if err := rows.Scan(&batch); err != nil {
					rows.Close()
					panic(err)
				}
				emptyBatches = append(emptyBatches, batch)
			}

			rows.Close()
		}

		rows, err := tx.GetEngine().Query(selectItems)

		if err != nil {
			panic(err)
		}

		var (
			currentBatch int
		)

		if eventPos.ItemCount != 0 {
			if !rows.Next() {
				panic("Data inconsistency. The last item sent cannot be found.")
			}
			var i int
			must(rows.Scan(&currentBatch, &i, &api.Item{}))
			for len(emptyBatches) != 0 && emptyBatches[0] < currentBatch {
				emptyBatches = emptyBatches[1:]
			}
		}

		event := api.Event{
			ID:        api.UUID(dbEvent.EventID),
			Index:     eventPos.Index,
			Type:      dbEvent.EventType,
			ItemCount: dbEvent.ItemCount,
			Checksum:  dbEvent.Checksum,
		}

		empty := true

		for rows.Next() {
			var (
				batch int
				index int
				item  []byte
			)
			if err := rows.Scan(&batch, &index, &item); err != nil {
				panic(err)
			}

			if batch-currentBatch > 1 {
				// There is a hole in the items list
				// Check for empty batches
				for len(emptyBatches) != 0 && emptyBatches[0] == currentBatch+1 {
					emptyBatches = emptyBatches[1:]
					currentBatch++
				}
				if batch-currentBatch > 1 {
					break
				}
			}

			currentBatch = batch

			if maxSize != 0 && estimateSize+len(item) >= maxSize {
				// Adding one more item exceed the maximum envelope size.
				// Stops here
				break
			}

			event.Items = append(event.Items, item)
			empty = false
			eventPos.ItemCount++
			estimateSize += len(item)
		}

		must(rows.Close())

		if eventPos.ItemCount != dbEvent.ItemCount {
			allDone = false
		}

		if !empty || position.Start || dbEnvelope.Status == storage.EnvelopeStorageReceptionComplete {
			envelope.Events = append(envelope.Events, event)
			position.Events[event.ID] = eventPos
		}
	}

	position.Start = false
	position.Complete = dbEnvelope.Status == storage.EnvelopeStorageReceptionComplete && allDone
	envelope.Last = position.Complete
	return envelope, position, nil
}

// Purge removes all envelopes selected by the given filter
func (es EnvelopeStorage) Purge(filter storage.EnvelopeFilter) {
	var batchSize = 1000

	allIDs := make(chan api.UUID, batchSize*2)
	purgeIDs := make(chan []api.UUID, 2)
	done := make(chan bool)
	errC := make(chan interface{})

	catchErrors := func() {
		if r := recover(); r != nil {
			errC <- r
		}
	}

	// read all the envelopes in transaction-independant batches,
	go func() {
		defer catchErrors()
		defer func() { close(allIDs) }()

		var offset = 0

		for func() (cont bool) {
			tx := es.MustBegin()
			defer tx.Rollback()

			q := tx.Query(es.model.Envelope).
				Select(es.model.Envelope.ID).
				OrderBy(es.model.Envelope.CreatedAt).
				LimitOffset(batchSize*10, offset)

			rows, err := q.SQLQuery()
			must(err)
			defer rows.Close()

			for rows.Next() {
				var id uuid.UUID
				must(rows.Scan(&id))
				allIDs <- api.UUID(id)
				cont = true
			}

			offset += batchSize * 10
			return
		}() {
		}
	}()

	// filter with the callback
	go func() {
		defer catchErrors()
		defer func() { close(purgeIDs) }()

		var inputBatch []api.UUID
		for id := range allIDs {
			inputBatch = append(inputBatch, id)
			if len(inputBatch) == batchSize {
				purgeIDs <- filter(inputBatch)
				inputBatch = nil
			}
		}

		if len(inputBatch) != 0 {
			purgeIDs <- filter(inputBatch)
		}
	}()

	// remove the envelopes that where not filtered out, by batches,
	// in independant transactions

	go func() {
		defer catchErrors()

		var input []uuid.UUID

		var removeN = func(n int) {
			tx := es.MustBegin()
			defer tx.RollbackIfOpened()

			var ids []interface{}

			for i := 0; i != n; i++ {
				ids = append(ids, uuid.UUID(input[i]))
			}

			input = input[n:]

			_, err := tx.GetEngine().Exec(
				es.model.Envelope.GetMapper().Table().Delete().
					Where(es.model.Envelope.ID.In(ids...)),
			)
			must(err)
			tx.MustCommit()
		}

		for batch := range purgeIDs {
			for _, id := range batch {
				input = append(input, uuid.UUID(id))
			}
			if len(input) >= batchSize {
				removeN(batchSize)
			}
		}
		if len(input) != 0 {
			removeN(len(input))
		}
		done <- true
	}()

	select {
	case <-done:
		// all done \o/
	case err := <-errC:
		panic(err)
	}
}
