// Copyright 2016 Orus.io. All rights reserved.

package sqldb

import (
	"math/rand"
	"os"
	"testing"

	// load postgres driver
	_ "github.com/lib/pq"

	// load sqlite driver
	_ "github.com/mattn/go-sqlite3"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	xbus "xbus.io/go-xbus"
	"bitbucket.org/orus-io/xbus/migration"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"
)

const testMaxDBConnections = 10

// CloseFunction is a function with no arg nor return values for closing a resource
type CloseFunction func()

type migrateLogger struct {
	log xbus.Logger
}

func (l migrateLogger) Printf(format string, v ...interface{}) {
	l.log.Noticef(format, v...)
}

func (migrateLogger) Verbose() bool {
	return false
}

// TestDB holds a test database and provides access to it until its 'Close'
// function is called
type TestDB struct {
	Driver   string
	Dsn      string
	LogFlags qb.LogFlags
	Logger   xbus.Logger

	t    testing.TB
	lock *testutils.PortLock
}

// Close releases the underlying database
func (tdb *TestDB) Close() {
	if tdb.lock != nil {
		if err := tdb.lock.Unlock(); err != nil {
			tdb.t.Errorf("Error releasing lock: %s", err)
		}
	}
}

// GetModelDB returns a new connection pool to the database
func (tdb *TestDB) GetModelDB(maxConn int) (*model.Model, *yago.DB) {
	m, db, err := model.NewDB(tdb.Driver, tdb.Dsn, maxConn, tdb.Logger, tdb.LogFlags)
	if err != nil {
		tdb.t.Fatalf("Cannot initialize test database: %s", err)
		return nil, nil
	}
	return m, db
}

var (
	dbinstanceid int64
	dbtablenames []string
)

// GetTestDB creates a test database and returns a TestDB pointing to it
func GetTestDB(t testing.TB) *TestDB {
	var (
		driver     = "sqlite3"
		droptables = false
	)

	dsn, ok := os.LookupEnv("XBUS_TEST_POSTGRES_DSN")
	if ok {
		driver = "postgres"
		droptables = true
	} else {
		dsn, ok = os.LookupEnv("XBUS_TEST_SQLITE_DSN")
		if !ok {
			os.Remove("test.db")
			dsn = "test.db"
		}
	}
	logFlags := qb.LDefault
	if v, _ := os.LookupEnv("XBUS_TEST_SQL_LOGQUERY"); v != "" {
		logFlags |= qb.LQuery
	}
	if v, _ := os.LookupEnv("XBUS_TEST_SQL_LOGBINDINGS"); v != "" {
		logFlags |= qb.LBindings
	}
	tdb := TestDB{
		Driver: driver,
		Dsn:    dsn,
		Logger: testutils.GetTestLogger(t),
		t:      t,
	}

	recreate, ok := func() (bool, bool) {
		_, db := tdb.GetModelDB(1)
		defer db.Close()

		var allGood bool

		defer func() {
			if !allGood {
				tdb.Close()
			}
		}()
		if driver == "postgres" {
			lock := testutils.NewPortLock(1764)
			err := lock.Lock()
			if err != nil {
				t.Errorf("Cannot get lock: %s", err)
				return false, false
			}
			tdb.lock = lock
		}
		if dbinstanceid != 0 {
			var id int64
			row := db.Engine.DB().DB.QueryRow("SELECT id FROM xbus_test_instance")
			if err := row.Scan(&id); err == nil && dbinstanceid == id {
				return false, true
			}
			dbinstanceid = 0
		}
		if droptables {
			tables := db.Metadata.GetQbMetadata().Tables()
			for i := len(tables) - 1; i >= 0; i-- {
				table := tables[i]
				drop := table.Drop(db.Engine.Dialect())
				if _, err := db.Engine.DB().Exec(drop); err != nil {
					t.Log(err)
				}
			}
			if _, err := db.Engine.DB().DB.Exec("DROP TABLE IF EXISTS account_session"); err != nil {
				t.Fatal(err)
			}
			if _, err := db.Engine.DB().DB.Exec("DROP TABLE IF EXISTS job"); err != nil {
				t.Fatal(err)
			}
			if _, err := db.Engine.DB().DB.Exec("DROP TABLE IF EXISTS passwords"); err != nil {
				t.Fatal(err)
			}
			if _, err := db.Engine.DB().DB.Exec("DROP TABLE IF EXISTS schema_migrations"); err != nil {
				t.Fatal(err)
			}
			if _, err := db.Engine.DB().DB.Exec("DROP TABLE IF EXISTS xbus_test_instance"); err != nil {
				t.Fatal(err)
			}
		}
		return true, true
	}()
	if !ok {
		return nil
	}

	if recreate {
		if !func() bool {
			_, db := tdb.GetModelDB(1)
			defer db.Close()

			m, _, err := migration.New(db)
			if err != nil {
				t.Errorf("Failed to init migration engine: %s", err)
				return false
			}
			defer m.Close()

			m.Log = migrateLogger{tdb.Logger}

			if err := m.Up(); err != nil {
				t.Error(err)
				return false
			}
			return true
		}() {
			return nil
		}

		_, db := tdb.GetModelDB(1)
		defer db.Close()

		if _, err := db.Engine.DB().DB.Exec(
			"CREATE TABLE xbus_test_instance (id BIGINT)"); err != nil {
			t.Error(err)
			return nil
		}

		dbinstanceid = rand.Int63()

		if _, err := db.Engine.DB().DB.Exec(
			"INSERT INTO xbus_test_instance (id) VALUES ($1)", dbinstanceid,
		); err != nil {
			dbinstanceid = 0
			t.Error(err)
			return nil
		}

		rows, err := db.Engine.DB().DB.Query(
			"SELECT tablename FROM pg_catalog.pg_tables WHERE schemaname = 'public'")
		if err != nil {
			dbinstanceid = 0
			t.Error(err)
			return nil
		}
		defer rows.Close()

		for rows.Next() {
			var name string
			if err := rows.Scan(&name); err != nil {
				dbinstanceid = 0
				t.Error(err)
				return nil
			}
			if name == "schema_migrations" || name == "xbus_test_instance" {
				continue
			}
			dbtablenames = append(dbtablenames, name)
		}

	} else {
		_, db := tdb.GetModelDB(1)
		defer db.Close()

		for _, name := range dbtablenames {
			if _, err := db.Engine.DB().DB.Exec(
				"ALTER TABLE " + name + " DISABLE TRIGGER ALL",
			); err != nil {
				t.Fatal(err)
				return nil
			}
		}
		for _, name := range dbtablenames {
			if _, err := db.Engine.DB().DB.Exec(
				"DELETE FROM " + name,
			); err != nil {
				t.Fatal(err)
				return nil
			}
		}
		for _, name := range dbtablenames {
			if _, err := db.Engine.DB().DB.Exec(
				"ALTER TABLE " + name + " ENABLE TRIGGER ALL",
			); err != nil {
				t.Fatal(err)
				return nil
			}
		}
	}

	return &tdb
}
