// Copyright 2016 Orus.io. All rights reserved.

package storage

import (
	"testing"
	"time"

	"github.com/stretchr/testify/assert"

	"xbus.io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// EnvelopeStorageFactory is any factory of EnvelopeStorage
type EnvelopeStorageFactory func(bt testing.TB) (storage.EnvelopeStorage, func())

// EnvelopeStorageTests is a test suite for EnvelopeStorage implementations
func EnvelopeStorageTests(t *testing.T, factory EnvelopeStorageFactory, supportUnordered bool) {

	// setup a little event handler mock
	var (
		handlerCalled       bool
		handlerParamID      api.UUID
		handlerParamStatus  storage.EnvelopeStorageReceptionStatus
		handlerParamNewData bool
	)

	handler := func(id api.UUID, status storage.EnvelopeStorageReceptionStatus, newData bool) {
		handlerCalled = true
		handlerParamID = id
		handlerParamStatus = status
		handlerParamNewData = newData
	}

	resetHandler := func() {
		handlerCalled = false
	}

	t.Run("Store/TypesUnknown", func(t *testing.T) {
		envelopeStorage, clean := factory(t)
		defer clean()

		env := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index:     1,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		status, err := envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, env.ID, status.ID)
		assert.False(t, status.TypesKnown)
		assert.Equal(t, 1, len(status.Events))
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving.String(), status.Status.String())
		if t.Failed() {
			return
		}

	})

	t.Run("DataLessFirstFragment", func(t *testing.T) {
		envelopeStorage, clean := factory(t)
		if envelopeStorage == nil {
			t.Fatal("on envelope storage")
		}
		defer clean()

		env1 := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			EventIDs: []api.UUID{
				api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
			},
		}

		env2 := api.Envelope{
			ID: env1.ID,
			Events: []api.Event{
				api.Event{
					ID:    api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index: 1,
					Type:  "xbus.test",
					Items: []api.Item{
						[]byte("Data"),
						[]byte("Data"),
					},
					ItemCount: 2,
					Checksum:  1307176161,
				},
			},
		}

		status, err := envelopeStorage.StoreEnvelope(env1)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

		status, err = envelopeStorage.StoreEnvelope(env2)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)
	})

	t.Run("Store/SingleEnvelope", func(t *testing.T) {
		envelopeStorage, clean := factory(t)
		defer clean()
		envelopeStorage.SetEventHandler(handler)

		env := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index:     1,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		eventID := env.Events[0].ID
		env.EventIDs = []api.UUID{eventID}

		resetHandler()
		status, err := envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, env.ID, status.ID)
		assert.True(t, status.TypesKnown)
		assert.Equal(t, 1, len(status.Events))
		if t.Failed() {
			return
		}
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.Equal(t, uint64(1), status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Events[0].Status)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)

		assert.True(t, handlerCalled)
		assert.Equal(t, env.ID, handlerParamID)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, handlerParamStatus)
		assert.True(t, handlerParamNewData)

		types, exists := envelopeStorage.GetEventTypes(env.ID)
		assert.True(t, exists)
		assert.Equal(t, []string{"xbus.test"}, types)
	})

	t.Run("Store/PairOfEnvelope", func(t *testing.T) {
		env1 := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:    api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index: 1,
					Type:  "xbus.test",
					Items: []api.Item{[]byte("Data")},
				},
			},
		}

		eventID := env1.Events[0].ID
		env1.EventIDs = []api.UUID{eventID}

		env2 := api.Envelope{
			ID: env1.ID,
			Events: []api.Event{
				api.Event{
					ID:        eventID,
					Index:     2,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 2,
					Checksum:  1307176161,
				},
			},
		}

		t.Run("Ordered", func(t *testing.T) {
			envelopeStorage, clean := factory(t)
			defer clean()
			envelopeStorage.SetEventHandler(handler)

			resetHandler()
			status, err := envelopeStorage.StoreEnvelope(env1)
			assert.Nil(t, err)
			assert.Equal(t, env1.ID, status.ID)
			assert.Equal(t, eventID, status.Events[0].ID)
			assert.True(t, status.TypesKnown)
			assert.Equal(t, uint64(1), status.Events[0].AvailableItems)
			assert.Equal(t, nil, status.Events[0].Err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Events[0].Status)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

			assert.True(t, handlerCalled)
			assert.Equal(t, env1.ID, handlerParamID)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, handlerParamStatus)
			assert.True(t, handlerParamNewData)

			resetHandler()
			status, err = envelopeStorage.StoreEnvelope(env2)
			assert.Nil(t, err)
			assert.Equal(t, env2.ID, status.ID)
			assert.Equal(t, 1, len(status.Events))
			if t.Failed() {
				return
			}
			assert.Equal(t, eventID, status.Events[0].ID)
			assert.Equal(t, uint64(2), status.Events[0].AvailableItems)
			assert.Equal(t, nil, status.Events[0].Err)
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Events[0].Status)
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)

			assert.True(t, handlerCalled)
			assert.Equal(t, env1.ID, handlerParamID)
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete, handlerParamStatus)
			assert.True(t, handlerParamNewData)

		})

		if supportUnordered {
			t.Run("Unordered", func(t *testing.T) {
				envelopeStorage, clean := factory(t)
				defer clean()

				status, err := envelopeStorage.StoreEnvelope(env2)
				assert.Nil(t, err)
				assert.Equal(t, env1.ID, status.ID)
				assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

				status, err = envelopeStorage.StoreEnvelope(env1)
				assert.Nil(t, err)
				assert.Equal(t, env2.ID, status.ID)
				assert.Equal(t, 1, len(status.Events))
				if t.Failed() {
					return
				}
				assert.Equal(t, eventID, status.Events[0].ID)
				assert.True(t, status.TypesKnown)
				assert.Equal(t, uint64(2), status.Events[0].AvailableItems)
				assert.Equal(t, nil, status.Events[0].Err)
				assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Events[0].Status)
				assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)
			})
		}

		t.Run("CheckStalled", func(t *testing.T) {
			envelopeStorage, clean := factory(t)
			defer clean()
			envelopeStorage.SetEventHandler(handler)

			status, err := envelopeStorage.StoreEnvelope(env1)
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

			assert.False(t, envelopeStorage.CheckStalled(env1.ID, 50*time.Millisecond))

			time.Sleep(1000 * time.Millisecond)

			resetHandler()
			assert.True(t, envelopeStorage.CheckStalled(env1.ID, time.Millisecond))

			assert.True(t, handlerCalled)
			assert.Equal(t, env1.ID, handlerParamID)
			assert.Equal(t, storage.EnvelopeStorageReceptionStalled, handlerParamStatus)
			assert.False(t, handlerParamNewData)

			resetHandler()
			assert.True(t, envelopeStorage.CheckStalled(env1.ID, time.Second))
			assert.False(t, handlerCalled)

			resetHandler()
			status, err = envelopeStorage.StoreEnvelope(env2)
			assert.Nil(t, err)
			assert.Equal(t, env2.ID, status.ID)
			assert.False(t, envelopeStorage.CheckStalled(env1.ID, time.Second))
		})
	})

	t.Run("Read/SingleEnvelope", func(t *testing.T) {
		envelopeStorage, clean := factory(t)
		defer clean()
		env := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index:     1,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		eventID := env.Events[0].ID
		env.EventIDs = []api.UUID{eventID}

		status, err := envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)

		out, pos, err := envelopeStorage.ReadEnvelope(env.ID, storage.EnvelopeStartPos(), 0)
		assert.Nil(t, err)
		assert.True(t, pos.Complete)
		assert.Equal(t, env.ID, out.ID)
		assert.Equal(t, 1, len(out.Events))
		if t.Failed() {
			return
		}
		outEvent := out.Events[0]
		assert.Equal(t, eventID, outEvent.ID)
		assert.Equal(t, uint64(1), outEvent.Index)
		assert.Equal(t, "xbus.test", outEvent.Type)
		assert.Equal(t, []api.Item{[]byte("Data")}, outEvent.Items)
		assert.Equal(t, uint64(1), outEvent.ItemCount)
		assert.Equal(t, api.Checksum(1859455373), outEvent.Checksum)
	})

	t.Run("Stream/PairOfEnvelope", func(t *testing.T) {
		envelopeStorage, clean := factory(t)
		defer clean()

		env := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:    api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index: 1,
					Type:  "xbus.test",
					Items: []api.Item{[]byte("Data")},
				},
			},
		}

		eventID := env.Events[0].ID
		env.EventIDs = []api.UUID{eventID}

		status, err := envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, env.ID, status.ID)
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.True(t, status.TypesKnown)
		assert.Equal(t, uint64(0), status.Events[0].TotalItems)
		assert.Equal(t, uint64(1), status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Events[0].Status)
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

		out, pos, err := envelopeStorage.ReadEnvelope(env.ID, storage.EnvelopeStartPos(), 0)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, 1, len(out.Events))
		assert.Equal(t, 1, len(out.Events[0].Items))
		assert.Equal(t, "Data", string(out.Events[0].Items[0]))

		env = api.Envelope{
			ID: env.ID,
			Events: []api.Event{
				api.Event{
					ID:        eventID,
					Index:     2,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data2")},
					ItemCount: 2,
					Checksum:  0x92b3a8e2,
				},
			},
		}

		env.EventIDs = []api.UUID{eventID}

		status, err = envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, env.ID, status.ID)
		assert.Equal(t, 1, len(status.Events))
		if t.Failed() {
			return
		}
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.Equal(t, uint64(2), status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Events[0].Status)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)

		out, pos, err = envelopeStorage.ReadEnvelope(env.ID, pos, 0)
		assert.Equal(t, 1, len(out.Events))
		assert.Equal(t, "Data2", string(out.Events[0].Items[0]))
	})

	t.Run("Stream/ItemLessFragments", func(t *testing.T) {
		envID := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df")
		eventID := api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882")

		fragments := []api.Envelope{
			api.Envelope{
				ID:       envID,
				EventIDs: []api.UUID{eventID},
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 1,
						Type:  "xbus.test",
					},
				},
			},
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 2,
						Type:  "xbus.test",
						Items: []api.Item{[]byte("item1")},
					},
				},
			},
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 3,
						Type:  "xbus.test",
					},
				},
			},
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:        eventID,
						Index:     4,
						ItemCount: 2,
						Type:      "xbus.test",
						Items:     []api.Item{[]byte("item2")},
						Checksum:  0x220c46c7,
					},
				},
			},
		}

		bigEnvFragments := GenBigEnvFragments()

		t.Run("Ordered", func(t *testing.T) {
			envelopeStorage, clean := factory(t)
			defer clean()

			status, err := envelopeStorage.StoreEnvelope(fragments[0])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

			status, err = envelopeStorage.StoreEnvelope(fragments[1])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

			status, err = envelopeStorage.StoreEnvelope(fragments[2])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

			status, err = envelopeStorage.StoreEnvelope(fragments[3])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)
			assert.Equal(t, "", status.Error())

			out, pos, err := envelopeStorage.ReadEnvelope(envID, storage.EnvelopeStartPos(), 0)
			assert.Nil(t, err)
			assert.True(t, pos.Complete)
			assert.Equal(t, 1, len(out.Events))
			if len(out.Events) == 1 {
				assert.Equal(t, []api.Item{api.Item("item1"), api.Item("item2")}, out.Events[0].Items)
			}
		})

		t.Run("Ordered/ReadBig", func(t *testing.T) {
			envelopeStorage, clean := factory(t)
			defer clean()

			for _, fragment := range bigEnvFragments[:4] {
				status, err := envelopeStorage.StoreEnvelope(fragment)
				assert.Nil(t, err)
				assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)
			}

			go func() {
				time.Sleep(100 * time.Millisecond)
				for _, fragment := range bigEnvFragments[4 : len(bigEnvFragments)-1] {
					status, err := envelopeStorage.StoreEnvelope(fragment)
					assert.Nil(t, err)
					assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)
				}
				status, err := envelopeStorage.StoreEnvelope(bigEnvFragments[len(bigEnvFragments)-1])
				assert.Nil(t, err)
				assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)
				assert.Equal(t, "", status.Error())

			}()

			pos := storage.EnvelopeStartPos()
			itemCount := 0
			for !pos.Complete {
				out, newPos, err := envelopeStorage.ReadEnvelope(envID, pos, 500)
				if err != nil {
					t.Fatal(err)
				}
				if newPos.Equals(pos) {
					time.Sleep(100 * time.Millisecond)
				}
				for _, ev := range out.Events {
					itemCount += len(ev.Items)
					if newPos.Events[eventID].ItemCount > 202 {
						t.Fatal("ReadEnvelope returning too many items")
					}
				}
				pos = newPos
			}
			assert.Equal(t, 202, itemCount)
		})

		if supportUnordered {
			t.Run("Unordered", func(t *testing.T) {
				envelopeStorage, clean := factory(t)
				defer clean()

				status, err := envelopeStorage.StoreEnvelope(fragments[2])
				assert.Nil(t, err)
				assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

				status, err = envelopeStorage.StoreEnvelope(fragments[0])
				assert.Nil(t, err)
				assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

				status, err = envelopeStorage.StoreEnvelope(fragments[3])
				assert.Nil(t, err)
				assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

				status, err = envelopeStorage.StoreEnvelope(fragments[1])
				assert.Nil(t, err)
				assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)

				out, pos, err := envelopeStorage.ReadEnvelope(envID, storage.EnvelopeStartPos(), 0)
				assert.Nil(t, err)
				assert.True(t, pos.Complete)
				assert.Equal(t, 1, len(out.Events))
				if len(out.Events) == 1 {
					assert.Equal(t, []api.Item{api.Item("item1"), api.Item("item2")}, out.Events[0].Items)
				}
			})
		}

	})

	t.Run("Stream/EarlyEventCompletion", func(t *testing.T) {
		envelopeStorage, clean := factory(t)
		defer clean()

		envID := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df")
		eventID := api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882")
		eventID2 := api.MustUUIDFromString("615d146a-02af-11e7-8cdf-d3cb88225aff")

		status, err := envelopeStorage.StoreEnvelope(
			api.Envelope{
				ID:       envID,
				EventIDs: []api.UUID{eventID, eventID2},
				Events: []api.Event{
					api.Event{
						ID:        eventID,
						Type:      "xbus.test",
						Index:     1,
						Items:     []api.Item{[]byte("Data")},
						ItemCount: 1,
						Checksum:  0x6ed5098d,
					},
				},
			})
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

		status, err = envelopeStorage.StoreEnvelope(
			api.Envelope{
				ID:       envID,
				EventIDs: []api.UUID{eventID, eventID2},
				Events: []api.Event{
					api.Event{
						ID:        eventID2,
						Type:      "xbus.test.2",
						Index:     1,
						Items:     []api.Item{[]byte("Data")},
						ItemCount: 1,
						Checksum:  0x15d47408,
					},
				},
				Last: true,
			})
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete.String(), status.Status.String())
	})

	t.Run("Purge", func(t *testing.T) {
		envelopeStorage, clean := factory(t)
		defer clean()
		env1 := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index:     1,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		env2 := api.Envelope{
			ID: api.MustUUIDFromString("7f107142-8977-11e7-bd71-635f1840d117"),
			Events: []api.Event{
				api.Event{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index:     1,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		env1.EventIDs = []api.UUID{env1.Events[0].ID}
		env2.EventIDs = []api.UUID{env2.Events[0].ID}

		_, err := envelopeStorage.StoreEnvelope(env1)
		assert.Nil(t, err)

		_, err = envelopeStorage.StoreEnvelope(env2)
		assert.Nil(t, err)

		_, exists := envelopeStorage.GetEventTypes(env1.ID)
		assert.True(t, exists)

		_, exists = envelopeStorage.GetEventTypes(env2.ID)
		assert.True(t, exists)

		envelopeStorage.Purge(func(ids []api.UUID) []api.UUID {
			var purge []api.UUID
			for _, id := range ids {
				if id == env1.ID {
					purge = append(purge, id)
				}
			}
			return purge
		})

		_, exists = envelopeStorage.GetEventTypes(env1.ID)
		assert.False(t, exists)
		_, exists = envelopeStorage.GetEventTypes(env2.ID)
		assert.True(t, exists)
	})

}

// GenBigEnvFragments creates fragments of a big envelope
func GenBigEnvFragments() []api.Envelope {
	envID := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df")
	eventID := api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882")
	eventID2 := api.MustUUIDFromString("615d146a-02af-11e7-8cdf-d3cb88225aff")
	bigEnvFragments := []api.Envelope{
		api.Envelope{
			ID:       envID,
			EventIDs: []api.UUID{eventID, eventID2},
			Events: []api.Event{
				api.Event{
					ID:    eventID,
					Index: 1,
					Type:  "xbus.test",
				},
				api.Event{
					ID:    eventID2,
					Index: 1,
					Type:  "xbus.test.2",
				},
			},
		},
		api.Envelope{
			ID: envID,
			Events: []api.Event{
				api.Event{
					ID:    eventID,
					Index: 2,
					Type:  "xbus.test",
				},
			},
		},
		api.Envelope{
			ID: envID,
			Events: []api.Event{
				api.Event{
					ID:    eventID,
					Index: 3,
					Type:  "xbus.test",
				},
			},
		},
		api.Envelope{
			ID: envID,
			Events: []api.Event{
				api.Event{
					ID:    eventID,
					Index: 4,
					Type:  "xbus.test",
				},
			},
		},
	}

	item := make([]byte, 100)

	for i := 0; i != 100; i++ {
		bigEnvFragments = append(bigEnvFragments, api.Envelope{
			ID: envID,
			Events: []api.Event{
				api.Event{
					ID:    eventID,
					Index: uint64(i + 5),
					Type:  "xbus.test",
					Items: []api.Item{item},
				},
			},
		})
	}
	for i := 0; i != 100; i++ {
		bigEnvFragments = append(bigEnvFragments, api.Envelope{
			ID: envID,
			Events: []api.Event{
				api.Event{
					ID:    eventID2,
					Index: uint64(i + 2),
					Type:  "xbus.test.2",
					Items: []api.Item{item},
				},
			},
		})
	}

	bigEnvFragments = append(bigEnvFragments,
		api.Envelope{
			ID: envID,
			Events: []api.Event{
				api.Event{
					ID:    eventID,
					Index: 105,
					Type:  "xbus.test",
				},
			},
		},
		api.Envelope{
			ID: envID,
			Events: []api.Event{
				api.Event{
					ID:        eventID,
					Index:     106,
					Type:      "xbus.test",
					Items:     []api.Item{item},
					ItemCount: 101,
					Checksum:  0xbdac9720,
				},
				api.Event{
					ID:        eventID2,
					Index:     102,
					Type:      "xbus.test.2",
					Items:     []api.Item{item},
					ItemCount: 101,
					Checksum:  0x560279ee,
				},
			},
		})
	return bigEnvFragments
}
