package storage

import (
	"testing"

	"github.com/stretchr/testify/assert"

	"xbus.io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

var simpleEnvelope = api.Envelope{
	ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
	Events: []api.Event{
		api.Event{
			ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
			Index:     1,
			Type:      "xbus.test",
			Items:     []api.Item{[]byte("Data")},
			ItemCount: 1,
			Checksum:  1859455373,
		},
	},
}

// EnvelopeStorageBench is a benchmark suite for a EnvelopeStorage implementation
func EnvelopeStorageBench(b *testing.B, factory EnvelopeStorageFactory) {
	b.Run("InitClearStorage", func(b *testing.B) {
		for i := 0; i != b.N; i++ {
			_, clean := factory(b)
			clean()
		}
	})
	b.Run("StoreAndReadOneSimpleEnvelope", func(b *testing.B) {
		b.StopTimer()
		for i := 0; i != b.N; i++ {
			es, clean := factory(b)
			b.StartTimer()
			es.StoreEnvelope(simpleEnvelope)
			es.ReadEnvelope(simpleEnvelope.ID, storage.EnvelopeStartPos(), 1024*1024)
			b.StopTimer()
			clean()
		}
	})
	b.Run("StoreAndReadManySimpleEnvelope", func(b *testing.B) {
		var err error
		es, clean := factory(b)
		defer clean()
		for i := 0; i != b.N; i++ {
			simpleEnvelope.ID, err = api.UUIDV4()
			if err != nil {
				panic(err)
			}
			es.StoreEnvelope(simpleEnvelope)
			es.ReadEnvelope(simpleEnvelope.ID, storage.EnvelopeStartPos(), 1024*1024)
		}
	})
	b.Run("Ordered/ReadBig", func(b *testing.B) {
		envelopeStorage, clean := factory(b)
		defer clean()

		bigEnvFragments := GenBigEnvFragments()
		envID := bigEnvFragments[0].ID
		eventID := bigEnvFragments[0].EventIDs[0]
		for _, fragment := range bigEnvFragments[:len(bigEnvFragments)-1] {
			status, err := envelopeStorage.StoreEnvelope(fragment)
			assert.Nil(b, err)
			assert.Equal(b, storage.EnvelopeStorageReceptionReceiving, status.Status)
		}

		status, err := envelopeStorage.StoreEnvelope(bigEnvFragments[len(bigEnvFragments)-1])
		assert.Nil(b, err)
		assert.Equal(b, storage.EnvelopeStorageReceptionComplete, status.Status)
		assert.Equal(b, "", status.Error())

		b.ResetTimer()
		for i := 0; i != b.N; i++ {
			pos := storage.EnvelopeStartPos()
			itemCount := 0
			for !pos.Complete {
				out, newPos, err := envelopeStorage.ReadEnvelope(envID, pos, 500)
				if err != nil {
					b.Fatal(err)
				}
				for _, ev := range out.Events {
					itemCount += len(ev.Items)
					if newPos.Events[eventID].ItemCount > 202 {
						b.Fatal("ReadEnvelope returning too many items")
					}
				}
				pos = newPos
			}
			assert.Equal(b, 202, itemCount)
		}
	})
	b.Run("Ordered/WriteBig", func(b *testing.B) {
		b.StopTimer()
		bigEnvFragments := GenBigEnvFragments()

		for i := 0; i != b.N; i++ {
			envelopeStorage, clean := factory(b)
			b.StartTimer()

			for _, fragment := range bigEnvFragments[:len(bigEnvFragments)-1] {
				status, err := envelopeStorage.StoreEnvelope(fragment)
				assert.Nil(b, err)
				assert.Equal(b, storage.EnvelopeStorageReceptionReceiving, status.Status)
			}

			status, err := envelopeStorage.StoreEnvelope(bigEnvFragments[len(bigEnvFragments)-1])
			assert.Nil(b, err)
			assert.Equal(b, storage.EnvelopeStorageReceptionComplete, status.Status)
			assert.Equal(b, "", status.Error())

			b.StopTimer()
			clean()
		}
	})
}
