package storage

import (
	"testing"

	"github.com/stretchr/testify/assert"

	"xbus.io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// ProcessStorageFactory is any factory of ProcessStorage
type ProcessStorageFactory func(bt testing.TB) (storage.ProcessStorage, func())

// ProcessStorageTests is a test suite for ProcessStorage implementations
func ProcessStorageTests(t *testing.T, factory ProcessStorageFactory) {
	var uuids []api.UUID
	for i := 0; i != 3; i++ {
		value, err := api.UUIDV4()
		if err != nil {
			t.Fatal(err)
		}
		uuids = append(uuids, value)
	}

	t.Run("Process Life Cycle", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		p := storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
		}

		id, err := ps.CreateProcess(p)
		if err != nil {
			t.Fatal(err)
		}

		t.Run("Get", func(t *testing.T) {
			p1, ok := ps.GetProcess(id)
			if !ok {
				t.Fatal("Could not get the created process")
			}
			assert.False(t, p1.ID.IsZero())
			assert.Equal(t, uuids[0], p1.TriggerEmitterID)
			assert.Equal(t, uuids[1], p1.TriggerEnvelopeID)
		})
		t.Run("SetPipeline", func(t *testing.T) {
			ps.SetPipeline(id, uuids[2], storage.NodeOutputRef{NodeID: "source", Output: "default"})
			p1, ok := ps.GetProcess(id)
			if !ok {
				t.Fatal("Could not get the created process")
			}
			assert.Equal(t, uuids[2], p1.PipelineID)
		})
		t.Run("PurgeNonTerminated", func(t *testing.T) {
			ids, err := ps.PurgeProcess([]api.UUID{id})
			if err != nil {
				t.Fatal(err)
			}
			if len(ids) != 0 {
				t.Fatalf("Should not purge any process, purged: %s", ids)
			}
		})
		t.Run("SetStatus", func(t *testing.T) {
			if err := ps.SetStatus(id, api.Process_ERROR, ""); err != nil {
				t.Fatal(err)
			}
			assert.Equal(t, api.Process_ERROR.String(), ps.GetState(id).Status.String())
		})
		t.Run("Purge", func(t *testing.T) {
			ids, err := ps.PurgeProcess([]api.UUID{id})
			if err != nil {
				t.Fatalf("Expected no error, got: %s", err)
			}
			assert.Equal(t, []api.UUID{id}, ids)
			assert.Equal(t, 0, len(ps.Query(api.ProcessFilter{ID: []api.UUID{id}})))
		})
	})

	t.Run("ResultAck", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		p := storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
		}

		id, err := ps.CreateProcess(p)
		if err != nil {
			t.Fatal(err)
		}

		ps.AckResult(id)

		logEntries := ps.GetUnprocessedProcessLogEntries()

		assert.Equal(t, 1, len(logEntries))
		assert.Equal(t, id, logEntries[0].ProcessID)
		assert.Equal(t, true, logEntries[0].ResultAck)
	})

	t.Run("ProcessState", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		p := storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
		}

		id, err := ps.CreateProcess(p)
		if err != nil {
			t.Fatal(err)
		}

		t.Run("Save ResultAcked", func(t *testing.T) {
			func() {
				updater := ps.GetStateUpdater(id)
				defer updater.Unlock()

				state, err := updater.Get()
				if err != nil {
					t.Fatal(err)
				}
				assert.False(t, state.ResultAcked)
				state.ResultAcked = true
				if err := updater.Set(state); err != nil {
					t.Fatal(err)
				}
			}()

			state := ps.GetState(id)
			assert.True(t, state.ResultAcked)
		})

	})

	t.Run("Query", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		p := storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
		}

		id, err := ps.CreateProcess(p)
		if err != nil {
			t.Fatal(err)
		}

		t.Run("Query by ResultAcked", func(t *testing.T) {
			// Make sure we do have a state
			// XXX this should not be necessary but the current sql implementation
			// is a little weak
			func() {
				updater := ps.GetStateUpdater(id)
				defer updater.Unlock()

				state, err := updater.Get()
				if err != nil {
					t.Fatal(err)
				}
				assert.False(t, state.ResultAcked)
				if err := updater.Set(state); err != nil {
					t.Fatal(err)
				}
			}()

			t.Run("No filter", func(t *testing.T) {
				list := ps.Query(api.ProcessFilter{
					ResultAcked: []bool{},
				})

				if len(list) != 1 {
					t.Fatal("no process found")
				}

				assert.Equal(t, id, list[0].ID)
			})

			t.Run("Filter false", func(t *testing.T) {
				list := ps.Query(api.ProcessFilter{
					ResultAcked: []bool{false},
				})

				if len(list) != 1 {
					t.Fatal("no process found")
				}

				assert.Equal(t, id, list[0].ID)
			})

			t.Run("Filter true", func(t *testing.T) {
				list := ps.Query(api.ProcessFilter{
					ResultAcked: []bool{true},
				})
				assert.Equal(t, 0, len(list))
			})

		})
	})
}
