package xbus

import (
	"context"
	"errors"
	"fmt"
	"time"

	api "xbus.io/go-xbus/api"
	"xbus.io/go-xbus/envelope"
)

var (
	// ErrNoSuchInput when invalid input name is given
	ErrNoSuchInput = errors.New("No such input")
)

// APCLogger is the a logger embedded in a APC
type APCLogger interface {
	Noticef(string, ...interface{})
	Warningf(string, ...interface{})
	Errorf(string, ...interface{})
	EnvNoticef(*api.Envelope, string, ...interface{})
	EnvWarningf(*api.Envelope, string, ...interface{})
	EnvErrorf(*api.Envelope, string, ...interface{})
}

// ActorProcessingContext is passed to the actor Process handlers
type ActorProcessingContext struct {
	Request      *api.ActorProcessRequest
	Context      context.Context
	send         func(api.ActorProcessingState)
	Actor        *Actor
	Log          APCLogger
	LocalLog     Logger
	currentState api.ActorProcessingState
	detached     bool
	done         bool

	receiveEnvelope func(
		ctx context.Context, clientKind, clientID string,
		envelopeID api.UUID,
		head *api.Envelope, position *api.EnvelopePosition,
	) (*envelope.BuilderFromFragment, error)

	openOutput func(
		context *api.ProcessingContext, output string, eventTypes ...string,
	) *EnvelopeSender
	closeOutput func(
		context *api.ProcessingContext, output string,
	) (api.EnvelopeAck, error)

	sendFragment func(
		ctx *api.ProcessingContext, output string, env api.Envelope,
	) (api.EnvelopeAck_ReceptionStatus, error)

	processingEnd func(
		context *api.ProcessingContext,
		status api.ActorProcessingState_Status,
		messages []api.LogMessage,
	) error
}

// ActorAPCLogger is the a logger embedded in a APC
type ActorAPCLogger struct {
	apc *ActorProcessingContext
}

// Noticef logs a notice message
func (l *ActorAPCLogger) Noticef(format string, args ...interface{}) {
	l.apc.Actor.Log.EnvNoticef(l.apc.Request.Context, nil, format, args...)
}

// Warningf logs a warning message
func (l *ActorAPCLogger) Warningf(format string, args ...interface{}) {
	l.apc.Actor.Log.EnvWarningf(l.apc.Request.Context, nil, format, args...)
}

// Errorf logs a error message
func (l *ActorAPCLogger) Errorf(format string, args ...interface{}) {
	l.apc.Actor.Log.EnvErrorf(l.apc.Request.Context, nil, format, args...)
}

// EnvNoticef logs a notice message
func (l *ActorAPCLogger) EnvNoticef(env *api.Envelope, format string, args ...interface{}) {
	l.apc.Actor.Log.EnvNoticef(l.apc.Request.Context, env, format, args...)
}

// EnvWarningf logs a warning message
func (l *ActorAPCLogger) EnvWarningf(env *api.Envelope, format string, args ...interface{}) {
	l.apc.Actor.Log.EnvWarningf(l.apc.Request.Context, env, format, args...)
}

// EnvErrorf logs a error message
func (l *ActorAPCLogger) EnvErrorf(env *api.Envelope, format string, args ...interface{}) {
	l.apc.Actor.Log.EnvErrorf(l.apc.Request.Context, env, format, args...)
}

// GetInput returns a input by its name
func (apc *ActorProcessingContext) GetInput(name string) *api.ActorProcessRequest_Input {
	for _, input := range apc.Request.GetInputs() {
		if input.Name == name {
			return input
		}
	}
	return nil
}

// GetEnvelopeBuilder returns a builder of the envelope of a given input
func (apc *ActorProcessingContext) GetEnvelopeBuilder(
	inputName string,
) (*envelope.BuilderFromFragment, error) {

	input := apc.GetInput(inputName)
	if input == nil {
		return nil, ErrNoSuchInput
	}
	return apc.receiveEnvelope(
		apc.Context, "actor", apc.Actor.ID.String(),
		input.Envelope.ID,
		input.Envelope, input.Position,
	)
}

// GetEnvelope returns the Envelope of a given input
func (apc *ActorProcessingContext) GetEnvelope(
	inputName string,
) (envelope.Envelope, error) {
	builder, err := apc.GetEnvelopeBuilder(inputName)
	if err != nil {
		return nil, err
	}

	return builder.Envelope(), nil
}

// ReadEnvelope returns a EnvelopeReceiver that reads the envelope of the
// given input
// Deprecated: use GetEnvelope, or GetEnvelopeBuilder if you need more control
func (apc *ActorProcessingContext) ReadEnvelope(
	inputName string,
) (*EnvelopeReceiver, error) {
	builder, err := apc.GetEnvelopeBuilder(inputName)
	if err != nil {
		return nil, err
	}

	return NewEnvelopeReceiverFromBuilder(inputName, builder)
}

// CompleteEnvelope returns a complete envelope.
func (apc *ActorProcessingContext) CompleteEnvelope(
	inputName string, timeout time.Duration,
) (*api.Envelope, error) {
	e, err := apc.GetEnvelope(inputName)
	if err != nil {
		return nil, err
	}
	return envelope.Complete(e, timeout)
}

// Send sends an envelope to a given output
func (apc *ActorProcessingContext) Send(
	outputName string, e envelope.Envelope,
) error {
	return SendEnvelope(
		apc.Context, e,
		api.EnvelopeMaxSize/3, EnvelopeSenderDefaultDelay, DefaultMaxChunksByFragment,
		func(fragment *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
			return apc.sendFragment(apc.Request.Context, outputName, *fragment)
		})
}

// OpenOutput returns a EnvelopeSender that sends to a given output
// Deprecated: use Send
func (apc *ActorProcessingContext) OpenOutput(
	outputName string, eventTypes ...string,
) *EnvelopeSender {
	return apc.openOutput(
		apc.Request.Context, outputName, eventTypes...)
}

// CloseOutput closes an output without sending anything to it
func (apc *ActorProcessingContext) CloseOutput(outputName string) error {
	_, err := apc.closeOutput(apc.Request.Context, outputName)
	return err
}

// ForwardEnvelope sends an existing envelope to an output
func (apc *ActorProcessingContext) ForwardEnvelope(
	outputName string, envelopeID api.UUID,
) error {
	newEnv := api.Envelope{
		ID:   envelopeID,
		Last: true,
	}
	_, err := apc.sendFragment(
		apc.Request.Context,
		outputName,
		newEnv)
	return err
}

// ForwardInput sends the envelope received on an input to a destination output
func (apc *ActorProcessingContext) ForwardInput(inputName, outputName string) error {
	input := apc.GetInput(inputName)
	if input == nil {
		return ErrNoSuchInput
	}
	return apc.ForwardEnvelope(outputName, input.Envelope.ID)
}

// IsDone returns true if processing is marked done
func (apc *ActorProcessingContext) IsDone() bool {
	return apc.done
}

// IsDetached returns true if processing is marked detached
func (apc *ActorProcessingContext) IsDetached() bool {
	return apc.detached
}

// Detach signals that when the handler finishes, the processing should
// not be automatically marked done successfully
func (apc *ActorProcessingContext) Detach() {
	apc.detached = true
}

// Done informs the bus that the processing is over
func (apc *ActorProcessingContext) Done() {
	// TODO handle errors
	apc.currentState.Status = api.ActorProcessingState_SUCCESS
	apc.done = true
	if apc.detached {
		apc.processingEnd(apc.Request.Context, api.ActorProcessingState_SUCCESS, nil)
	} else {
		apc.send(apc.currentState)
	}
}

// ActorLeaving informs the bus that the actor will be offline just after
// handling the request
func (apc *ActorProcessingContext) ActorLeaving() {
	apc.currentState.ActorLeaving = true
}

// Error informs the bus that the processing ends with error
func (apc *ActorProcessingContext) Error(err error) {
	apc.currentState.Status = api.ActorProcessingState_ERROR
	apc.currentState.Messages = []api.LogMessage{{
		Time:  time.Now(),
		Level: api.LogLevel_ERROR,
		Text:  err.Error(),
	}}
	apc.done = true
	if apc.detached {
		apc.processingEnd(apc.Request.Context, api.ActorProcessingState_ERROR,
			[]api.LogMessage{
				{Time: time.Now(),
					Level: api.LogLevel_ERROR,
					Text:  fmt.Sprintf("Error processing: %s", err)},
			})
	} else {
		apc.send(apc.currentState)
	}
}

// End sends the 'processing end' message
func (apc *ActorProcessingContext) End(
	status api.ActorProcessingState_Status, msgs []api.LogMessage,
) {
	apc.currentState.Status = status
	apc.currentState.Messages = msgs
	apc.done = true
	if apc.detached {
		apc.processingEnd(apc.Request.Context, status, msgs)
	} else {
		apc.send(apc.currentState)
	}
}

// ApcTestSetReceiveEnvelope injects a custom ReceiveEnvelope function into a ActorProcessingContext
// FOR TESTING PURPOSE ONLY
func ApcTestSetReceiveEnvelope(apc *ActorProcessingContext,
	receiveEnvelope func(
		ctx context.Context, clientKind, clientID string,
		envelopeID api.UUID,
		head *api.Envelope, position *api.EnvelopePosition,
	) (*envelope.BuilderFromFragment, error),
) {
	apc.receiveEnvelope = receiveEnvelope
}

// ApcTestSetSend injects a custom Send function into a ActorProcessingContext.
// FOR TESTING PURPOSE ONLY
func ApcTestSetSend(apc *ActorProcessingContext,
	send func(api.ActorProcessingState),
) {
	apc.send = send
}

// ApcTestSetOpenOutput injects a custom OpenOutput function into a
// ActorProcessingContext. FOR TESTING PURPOSE ONLY
func ApcTestSetOpenOutput(
	apc *ActorProcessingContext,
	openOutput func(
		context *api.ProcessingContext, output string, eventTypes ...string,
	) *EnvelopeSender,
) {
	apc.openOutput = openOutput
}

// ApcTestSetCloseOutput injects a custom CloseOutput function into a
// ActorProcessingContext. FOR TESTING PURPOSE ONLY
func ApcTestSetCloseOutput(
	apc *ActorProcessingContext,
	closeOutput func(
		context *api.ProcessingContext, output string,
	) (api.EnvelopeAck, error),
) {
	apc.closeOutput = closeOutput
}

// ApcTestSetSendFragment injects a custom SendEnvelope function into a
// ActorProcessingContext. FOR TESTING PURPOSE ONLY
func ApcTestSetSendFragment(
	apc *ActorProcessingContext,
	sendFragment func(
		ctx *api.ProcessingContext, output string, env api.Envelope,
	) (api.EnvelopeAck_ReceptionStatus, error),
) {
	apc.sendFragment = sendFragment
}

// ApcTestSetProcessingEnd injects a custom ProcessingEnd function into a
// ActorProcessingContext. FOR TESTING PURPOSE ONLY
func ApcTestSetProcessingEnd(
	apc *ActorProcessingContext,
	processingEnd func(
		context *api.ProcessingContext,
		status api.ActorProcessingState_Status,
		messages []api.LogMessage,
	) error,
) {
	apc.processingEnd = processingEnd
}
