package api

import (
	"bytes"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"strings"
	"unicode/utf8"

	"github.com/gogo/protobuf/jsonpb"
	"github.com/gogo/protobuf/proto"
)

/* Custom json marshaler/unmarshaler for Envelope and all the types that uses
it (OutputRequest, ActorProcessRequest) */

//ItemEncoding The idem encoding in a json payload
//go:generate enum-marshaler -type ItemEncoding -prefix Item -lower -tests
type ItemEncoding int

const (
	// ItemBinary is for base64 encoded data
	ItemBinary ItemEncoding = iota
	// ItemText is for text encoded data
	ItemText
	// ItemJSON is for embedding json directly
	ItemJSON
)

// ParseItemEncodingList parses a comma separated list of item encodings
// Space characters are trimmed from each item
func ParseItemEncodingList(s string) ([]ItemEncoding, error) {
	var itemEncodings []ItemEncoding
	for _, encoding := range strings.Split(s, ",") {
		encoding = strings.Trim(encoding, " ")
		if encoding == "" {
			continue
		}
		var item ItemEncoding
		if err := item.UnmarshalText([]byte(encoding)); err != nil {
			return nil, err
		}
		itemEncodings = append(itemEncodings, item)
	}
	return itemEncodings, nil
}

// JSONRawEvent is a Event with raw items
// It serves as a temporary decoding structure
type JSONRawEvent struct {
	ID           UUID              `json:"id"`
	Type         string            `json:"type,omitempty"`
	Index        json.RawMessage   `json:"index,omitempty"`
	ItemCount    json.RawMessage   `json:"itemCount,omitempty"`
	Checksum     Checksum          `json:"checksum,omitempty"`
	Items        []json.RawMessage `json:"items,omitempty"`
	ItemEncoding ItemEncoding      `json:"encoding"`
}

// JSONUnmarshalItem unmarshals a single item
func JSONUnmarshalItem(encoding ItemEncoding, raw json.RawMessage) (Item, error) {
	switch encoding {
	case ItemBinary:
		// a json encoded string containing base64 encoded data, let's decode it
		var s string
		if err := json.Unmarshal(raw, &s); err != nil {
			return nil, err
		}
		return base64.StdEncoding.DecodeString(s)

	case ItemText:
		// a single json encoded string
		var s string
		if err := json.Unmarshal(raw, &s); err != nil {
			return nil, err
		}
		return Item(s), nil

	case ItemJSON:
		// raw json, the very data we want. Nothing to do
		return Item(raw), nil
	default:
		return nil, fmt.Errorf("Unsupported encoding: %s", encoding)
	}
}

// UnmarshalJSON decodes the items depending on the declared encoding
func (event *Event) UnmarshalJSON(data []byte) error {
	var raw JSONRawEvent
	if err := json.Unmarshal(data, &raw); err != nil {
		return err
	}
	for _, rawItem := range raw.Items {
		item, err := JSONUnmarshalItem(raw.ItemEncoding, rawItem)
		if err != nil {
			return err
		}
		event.Items = append(event.Items, item)
	}

	if strings.HasPrefix(string(raw.Index), `"`) {
		raw.Index = raw.Index[1 : len(raw.Index)-1]
	}
	if strings.HasPrefix(string(raw.ItemCount), `"`) {
		raw.ItemCount = raw.ItemCount[1 : len(raw.ItemCount)-1]
	}
	event.ID = raw.ID
	event.Type = raw.Type
	event.Checksum = raw.Checksum
	if len(raw.Index) != 0 {
		if err := json.Unmarshal(raw.Index, &event.Index); err != nil {
			return err
		}
	}
	if len(raw.ItemCount) != 0 {
		if err := json.Unmarshal(raw.ItemCount, &event.ItemCount); err != nil {
			return err
		}
	}
	return nil
}

// UnmarshalJSONPB ...
func (event *Event) UnmarshalJSONPB(u *jsonpb.Unmarshaler, data []byte) error {
	return event.UnmarshalJSON(data)
}

// JSONMarshal encodes api messages to JSON. The encodings list controls how
// the api.Item will be encoded
func JSONMarshal(encodings []ItemEncoding, pb proto.Message) ([]byte, error) {
	m := jsonpb.Marshaler{}
	setMarshalerEncodings(&m, encodings)
	defer unsetMarshalerEncodings(&m)

	var buf bytes.Buffer
	if err := m.Marshal(&buf, pb); err != nil {
		return nil, err
	}
	return buf.Bytes(), nil
}

var (
	marshalerEncodings = make(map[*jsonpb.Marshaler][]ItemEncoding)
)

func getMarshalerEncodings(m *jsonpb.Marshaler) []ItemEncoding {
	encodings, _ := marshalerEncodings[m]
	return encodings
}

func setMarshalerEncodings(m *jsonpb.Marshaler, v []ItemEncoding) {
	marshalerEncodings[m] = v
}

func unsetMarshalerEncodings(m *jsonpb.Marshaler) {
	delete(marshalerEncodings, m)
}

// MarshalJSONEncoded encodes a Item to JSON
func (item Item) MarshalJSONEncoded(encoding ItemEncoding) ([]byte, error) {
	switch encoding {
	case ItemBinary:
		s := base64.StdEncoding.EncodeToString(item)
		b, err := json.Marshal(s)
		if err != nil {
			return nil, err
		}
		return json.RawMessage(b), nil

	case ItemText:
		// Check if it is valid utf-8
		if !utf8.Valid(item) {
			return nil, fmt.Errorf("Not a valid UTF-8 string")
		}
		// marshal and return it
		s := string(item)
		b, err := json.Marshal(s)
		if err != nil {
			return nil, err
		}
		return json.RawMessage(b), nil

	case ItemJSON:
		// Check if we have valid json
		if err := json.Compact(bytes.NewBuffer(nil), item); err != nil {
			return nil, fmt.Errorf("Not a valid JSON item")
		}
		// return it
		return json.RawMessage(item), nil

	default:
		return nil, fmt.Errorf("unsupported encoding: %s", encoding)
	}
}

// MarshalJSONPB encode a Event with an additional "encoding" that is
// the actual encoding of the items
func (event Event) MarshalJSONPB(m *jsonpb.Marshaler) ([]byte, error) {
	encodings := getMarshalerEncodings(m)
	if len(encodings) == 0 {
		encodings = []ItemEncoding{ItemBinary}
	}

	defer setMarshalerEncodings(m, encodings)

encodingLoop:
	for _, encoding := range encodings {
		setMarshalerEncodings(m, []ItemEncoding{encoding})
		var items []json.RawMessage
		for _, item := range event.Items {
			s, err := item.MarshalJSONEncoded(encoding)
			if err != nil {
				continue encodingLoop
			}
			items = append(items, json.RawMessage(s))
		}
		m := map[string]interface{}{
			"id":        event.ID,
			"type":      event.Type,
			"index":     event.Index,
			"itemCount": event.ItemCount,
			"checksum":  event.Checksum,
			"encoding":  encoding,
		}
		if len(items) > 0 {
			m["items"] = items
		}
		return json.Marshal(m)
	}

	return nil, fmt.Errorf(
		"No suitable encoding for event %s in %v",
		event.ID, encodings)

}
