package xbus

import (
	"context"
	"errors"
	"fmt"
	"sync"

	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/envelope"
)

var (
	// ErrEmissionCanceled is set on an Envelope if its emission was canceled
	ErrEmissionCanceled = errors.New("Emission canceled")
)

// ErrProcessFailed is returned when the process is in ERROR state
type ErrProcessFailed struct {
	msg    string
	errors []api.LogMessage
}

// NewErrProcessFailed crates a ErrProcessFailed error
func NewErrProcessFailed(errors []api.LogMessage) ErrProcessFailed {
	err := ErrProcessFailed{
		"Process failed: ",
		errors,
	}

	for i, lm := range errors {
		err.msg += lm.Text
		if i != 0 {
			err.msg += "; "
		}
	}

	return err
}

func (e ErrProcessFailed) Error() string {
	return e.msg
}

// SendFragment send a fragment to the server
func (a *Actor) SendFragment(
	context *api.ProcessingContext,
	output string,
	fragment *api.Envelope,
) (
	api.EnvelopeAck_ReceptionStatus, error,
) {
	ack, err := a.Client.sendEnvelopeFragment(a.ID, context, output, fragment)
	return ack.GetStatus(), err
}

// CustomEmission ...
func (a *Actor) CustomEmission(
	ctx context.Context, envelope envelope.Envelope, request bool,
) (*Emission, error) {
	return NewEmission(ctx, a, envelope, request)
}

// Emit sends an envelope. It returns when the envelope is completely sent. If
// the envelope is opened, it will block until its complete. See CustomEmission
// for having more options.
func (a *Actor) Emit(
	ctx context.Context, envelope envelope.Envelope,
) (*Emission, error) {
	emission, err := NewEmission(ctx, a, envelope, false)
	if err != nil {
		return nil, err
	}
	if err := emission.Wait(); err != nil {
		return nil, err
	}
	return emission, nil
}

// Request sends an envelope and wait for the reply
func (a *Actor) Request(
	ctx context.Context, envelope envelope.Envelope,
) (envelope.Envelope, error) {
	emission, err := a.CustomEmission(ctx, envelope, true)
	if err != nil {
		return nil, err
	}
	if err := emission.Wait(); err != nil {
		return nil, err
	}
	if err := emission.WaitCompletion(); err != nil {
		return nil, err
	}
	return emission.GetResponse()
}

// Emission is an ongoing emission and provides eeasy APIs to wait for the reply
// it controls:
// - splitting the envelope in fragments as long as its valid, until we
//   got them all
// - sending fragments as long as the server accepts them, until all of
//   them are sent
// - receiving the process state info
type Emission struct {
	ctx      context.Context
	cancel   func()
	actor    *Actor
	envelope envelope.Envelope
	output   string

	emitResultC chan error
	wg          sync.WaitGroup

	finalState *api.EmitterEnvelopeState
}

// NewEmission start a emission
func NewEmission(
	ctx context.Context, a *Actor, envelope envelope.Envelope, request bool,
) (*Emission, error) {

	output := api.DefaultOutput
	if request {
		output = "request"
	}
	emission := Emission{
		actor:       a,
		envelope:    envelope,
		output:      output,
		emitResultC: make(chan error, 1),
	}
	emission.ctx, emission.cancel = context.WithCancel(ctx)
	emission.wg.Add(1)

	go func() {
		defer emission.wg.Done()
		emission.emitResultC <- SendEnvelope(
			emission.ctx, emission.envelope,
			768*1024, EnvelopeSenderDefaultDelay, DefaultMaxChunksByFragment,
			emission.send)
		close(emission.emitResultC)
	}()
	return &emission, nil
}

func (e *Emission) send(fragment *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
	return e.actor.SendFragment(nil, e.output, fragment)
}

// Wait for the envelope emission completion
func (e *Emission) Wait() error {
	select {
	case <-e.ctx.Done():
		e.envelope.CloseWithError(e.ctx.Err())
		return e.ctx.Err()
	case err := <-e.emitResultC:
		return err
	}
}

// WaitCompletion waits for the Process to be completed
func (e *Emission) WaitCompletion() error {

	// The envelope is completed, now wait for the process to end
	stateChan, sub, err := e.actor.Client.API.ProcessState.EnvelopeStatesSubscribeChan(
		e.actor.ID.String(), e.envelope.ID().String(),
	)

	if err != nil {
		return err
	}
	defer func() {
		sub.Unsubscribe()

		// make sure nothing remains in the chan that would block a routine upstream
		// TODO this should be handled by the subscription itself, which needs
		// some improvements in nrpc
		select {
		case <-stateChan:
		default:
		}
	}()

	// get the current state before waiting for a change
	state, err := e.actor.Client.API.ProcessState.GetEnvelopeState(
		e.actor.ID.String(),
		api.GetEnvelopeStateRequest{ID: e.envelope.ID()},
	)

	if state.Status == api.Process_DONE {
		e.finalState = &state
		return nil
	}
	if state.Status == api.Process_ERROR {
		e.finalState = &state
		return NewErrProcessFailed(state.Errors)
	}

	for {
		select {
		case state := <-stateChan:
			if state.Status == api.Process_DONE {
				e.finalState = &state
				return nil
			}
			if state.Status == api.Process_ERROR {
				e.finalState = &state
				return NewErrProcessFailed(state.Errors)
			}
		case <-e.ctx.Done():
			return e.ctx.Err()
		}
	}
}

// GetResponse returns the response if available
func (e *Emission) GetResponse() (envelope.Envelope, error) {
	if e.finalState == nil {
		return nil, fmt.Errorf("Process is not terminated")
	}
	if e.finalState.Response != nil {
		builder, err := e.actor.Client.ReceiveEnvelope(
			e.ctx, "actor", e.actor.ID.String(),
			e.finalState.Response.GetFragment().ID,
			e.finalState.Response.GetFragment(),
			e.finalState.Response.GetPosition(),
		)
		if err != nil {
			return nil, err
		}
		return builder.Envelope(), nil
	}
	return nil, fmt.Errorf("Process has no response attached")
}

// CancelProcess cancels the process
/*
func (e *Emission) CancelProcess() {

} */

// Close releases all underlying resources
// if still sending, the envelope should be canceled first, possibly
func (e *Emission) Close() {
	e.cancel()
	e.wg.Wait()
}
