package envelope

import (
	"fmt"
	"io"
	"sync"

	"xbus.io/go-xbus/api"
)

// NewBuilder creates a new Builder
func NewBuilder(id api.UUID) *Builder {
	builder := Builder{id: id}
	builder.cond = sync.NewCond(&builder.mut)
	return &builder
}

// Builder is the default Envelope implementation
type Builder struct {
	id api.UUID

	mut  sync.Mutex
	cond *sync.Cond

	err            error
	messages       []Message
	currentMessage int
}

// ID returns the envelope ID
func (e *Builder) ID() api.UUID {
	return e.id
}

// Opened returns true if the builder stills accept new messages
func (e *Builder) Opened() (opened bool) {
	e.mut.Lock()
	opened = e.err == nil
	e.mut.Unlock()
	return
}

// Add adds a message to the envelope. The envelope must be opened.
func (e *Builder) Add(m Message) error {
	e.mut.Lock()
	defer e.mut.Unlock()

	if e.err != nil {
		return e.err
	}

	e.messages = append(e.messages, m)

	e.cond.Broadcast()
	return nil
}

// NextMessage returns the next available message
func (e *Builder) NextMessage(wait bool) (Message, error) {
	e.mut.Lock()
	defer e.mut.Unlock()

	for wait && e.err == nil && len(e.messages) == e.currentMessage {
		e.cond.Wait()
	}
	if e.err != nil && e.err != ErrClosedEnvelope {
		return nil, e.err
	} else if e.currentMessage == len(e.messages) {
		if e.err == ErrClosedEnvelope {
			return nil, io.EOF
		}
		return nil, nil
	}
	m := e.messages[e.currentMessage]
	e.currentMessage++
	return m, nil
}

// Clone the builder. Must be closed to make any sens (for now)
func (e *Builder) Clone() (Envelope, error) {
	if e.err != nil {
		return nil, e.err
	}
	ne := Builder{id: e.id}
	ne.cond = sync.NewCond(&ne.mut)
	for _, m := range e.messages {
		ne.messages = append(ne.messages, m.Clone())
	}
	return &ne, nil
}

// Close closes properly the envelope
func (e *Builder) Close() (err error) {
	e.mut.Lock()
	err = e.err
	e.err = ErrClosedEnvelope
	e.cond.Broadcast()
	e.mut.Unlock()
	return
}

// CloseWithError closes the envelope with an error. If the error is not
// ErrClosedEnvelope then all the messages are set in error state
func (e *Builder) CloseWithError(err error) (outErr error) {
	e.mut.Lock()
	defer e.mut.Unlock()

	if e.err != nil {
		return e.err
	}

	if err == ErrClosedEnvelope {
		err = fmt.Errorf("Attempted to CloseWithError(ErrClosedEnvelope)")
		outErr = err
	}

	for _, m := range e.messages {
		if mc, ok := m.(MessageCloser); ok {
			mc.CloseWithError(err)
		}
	}

	e.err = err
	e.mut.Unlock()
	return outErr
}
