package envelope

import (
	"context"
	"errors"
	"io"
	"sync"
	"time"

	"xbus.io/go-xbus/api"
)

// ErrTimeout ...
var ErrTimeout = errors.New("Timeout")

// Complete accumulate the complete envelope as a api.Envelope
func Complete(e Envelope, timeout time.Duration) (*api.Envelope, error) {
	ctx, cancel := context.WithTimeout(context.Background(), timeout)
	defer cancel()
	return CompleteWithContext(ctx, e)
}

// CompleteWithContext accumulate the complete envelope as a api.Envelope
func CompleteWithContext(ctx context.Context, e Envelope) (*api.Envelope, error) {
	done := make(chan struct{})
	defer close(done)
	go func() {
		select {
		case <-ctx.Done():

			e.CloseWithError(ctx.Err())
		case <-done:
		}
	}()
	var (
		result = api.Envelope{
			ID:   e.ID(),
			Last: true,
		}
		finalErr error
		mu       sync.Mutex
		wg       sync.WaitGroup
	)
	for msg, err := e.NextMessage(true); err != io.EOF; msg, err = e.NextMessage(true) {
		mu.Lock()
		result.EventIDs = append(result.EventIDs, msg.ID())
		mu.Unlock()
		wg.Add(1)
		go func(msg Message) {
			defer wg.Done()
			chunks, err := ReadAllChunks(msg.ChunkReader())
			if err != nil {
				mu.Lock()
				if finalErr == nil {
					finalErr = err
					e.CloseWithError(err)
				}
				mu.Unlock()
				return
			}
			var event = api.Event{
				ID:        msg.ID(),
				Type:      msg.MsgType(),
				Index:     1,
				ItemCount: uint64(len(chunks)),
				Checksum:  msg.Checksum(),
				Items:     api.ChunksAsItems(chunks),
			}

			mu.Lock()
			result.Events = append(result.Events, event)
			mu.Unlock()
		}(msg)
	}
	wg.Wait()
	if finalErr != nil {
		return nil, finalErr
	}
	return &result, nil
}
