/*Package envelope is a message oriented API to build envelopes */
package envelope

import (
	"io"

	"xbus.io/go-xbus/api"
)

// Envelope holds messages.
// The messages can be written and read on the fly, which makes the envelope
// more like a buffer holder than a message holder.
// It also means the envelope can be read only once, unless cloned first
type Envelope interface {
	ID() api.UUID

	// Returns the next message if any. It is a blocking call if 'wait' is true,
	// and all calls after the last message returns a 'EOF' error
	NextMessage(wait bool) (Message, error)

	// Clone the envelope and all the message in it.
	// The clone can be read without altering the original envelope read cursors.
	// Clone must be called before any call to "NextMessage"
	// Warning: using clone on big streamed envelopes can use a LOT of memory
	Clone() (Envelope, error)

	// Close the reading of the envelope and set a custom error.
	// The error cannot be ErrClosedEnvelope. Any subsequent
	// read or write in the envelope will fail with the error that was passed.
	CloseWithError(error) error
}

// NewEnvelope creates an envelope. The only possible error is when it fails
// to generate a UUID (which is unlikely), so you may consider MustNewEnvelope.
func NewEnvelope(messages ...Message) (Envelope, error) {
	id, err := api.UUIDV4()
	if err != nil {
		return nil, err
	}
	return NewEnvelopeWithID(id, messages...), nil
}

// MustNewEnvelope creates an envelope and panics if the UUID creation fails,
// which should be virtually never
func MustNewEnvelope(messages ...Message) Envelope {
	envelope, err := NewEnvelope(messages...)
	if err != nil {
		panic(err)
	}
	return envelope
}

// NewEnvelopeWithID creates an envelope with the given ID
func NewEnvelopeWithID(id api.UUID, messages ...Message) Envelope {
	builder := NewOpenedEnvelopeWithID(id)
	for _, m := range messages {
		if err := builder.Add(m); err != nil {
			// Error could only be a ErrClosedEnvelope, but we know we did not
			// close it yet.
			panic(err)
		}
	}
	if err := builder.Close(); err != nil {
		panic(err)
	}
	return builder
}

// NewOpenedEnvelope creates an envelope on which messages can be added
func NewOpenedEnvelope() (*Builder, error) {
	id, err := api.UUIDV4()
	if err != nil {
		return nil, err
	}
	return NewOpenedEnvelopeWithID(id), nil
}

// MustNewOpenedEnvelope creates an envelope on which messages can be added
func MustNewOpenedEnvelope() *Builder {
	builder, err := NewOpenedEnvelope()
	if err != nil {
		panic(err)
	}
	return builder
}

// NewOpenedEnvelopeWithID creates an envelope on which messages can be added
func NewOpenedEnvelopeWithID(id api.UUID) *Builder {
	return NewBuilder(id)
}

// ForEachMessage call a function for all messages of an envelope.
func ForEachMessage(env Envelope, fn func(Message) error) error {
	for msg, err := env.NextMessage(true); err != io.EOF; msg, err = env.NextMessage(true) {
		if err != nil {
			return err
		}
		if err := fn(msg); err != nil {
			env.CloseWithError(err)
			return err
		}
	}
	return nil
}
