package envelope

import (
	"xbus.io/go-xbus/api"
	"xbus.io/go-xbus/tools"
)

// BuilderFromFragment builds an Envelope from a serie of api.Envelope
// fragments.
type BuilderFromFragment struct {
	builder  *Builder
	sorter   *tools.EnvelopeFragmentSorter
	messages map[api.UUID]*StreamMessage

	Complete bool
	done     chan struct{}
}

// Envelope returns the envelope being built
func (eb *BuilderFromFragment) Envelope() Envelope {
	return eb.builder
}

// Done returns a channel that will be closed with the reception of fragments is complete
func (eb *BuilderFromFragment) Done() <-chan struct{} {
	return eb.done
}

// dispatchCurrent dispatches the current fragment of the sorter
// into the envelope
func (eb *BuilderFromFragment) dispatchCurrent() error {
	fragment, err := eb.sorter.Get()
	if err != nil {
		return err
	}
	for _, eventFragment := range fragment.Events {
		message, ok := eb.messages[eventFragment.ID]
		if !ok {
			message = NewStreamMessage(eventFragment.ID, eventFragment.Type)
			if err := eb.builder.Add(message); err != nil {
				return err
			}
			eb.messages[eventFragment.ID] = message
		}
		if message.Checksum() == 0 && eventFragment.Checksum != 0 {
			message.SetChecksum(eventFragment.Checksum)
		}
		if err := message.addChunks(api.ItemsAsChunks(eventFragment.Items)); err != nil {
			return err
		}
		if eb.sorter.Events[eventFragment.ID].Complete {
			if err := message.addChunks(nil); err != nil {
				return err
			}
		}
	}
	if eb.sorter.Complete {
		eb.Complete = true
		close(eb.done)
	}
	if eb.sorter.EventTypesKnown && eb.builder.Opened() {
		// we actually only care about events having at least a fragment,
		// but the current API only tell us when the event types are known
		return eb.builder.Close()
	}
	return nil

}

// Add adds a fragment
func (eb *BuilderFromFragment) Add(fragment *api.Envelope) error {
	eb.sorter.Add(fragment)
	for eb.sorter.Next() {
		if err := eb.dispatchCurrent(); err != nil {
			eb.Cancel(err)
			return err
		}
	}
	return nil
}

// Cancel stops builder and set an error on the envelope
func (eb *BuilderFromFragment) Cancel(err error) error {
	select {
	case <-eb.done:
	default:
		close(eb.done)
	}
	return eb.builder.CloseWithError(err)
}

// NewFromFragments returns a Envelope that can be fed with a stream
// of envelope fragments. After all fragments are received, the functions returns io.EOF
func NewFromFragments(envelopeID api.UUID) *BuilderFromFragment {
	builder := BuilderFromFragment{
		sorter:   tools.NewEnvelopeFragmentSorter(),
		messages: make(map[api.UUID]*StreamMessage),
		builder:  NewOpenedEnvelopeWithID(envelopeID),
		done:     make(chan struct{}),
	}

	return &builder
}
