package envelope

import (
	"bytes"
	"errors"
	"io"
	"io/ioutil"
	"sync"

	"xbus.io/go-xbus/api"
)

var (
	// MaxChunkSize is the maximum allowed size for a chunk. If bigger, any
	// api taking a chunk will refuse it
	MaxChunkSize = 1024 * 768

	// ChunkSplitThreadhold is the chunk size above which a paylod will be
	// splitted in chunks
	ChunkSplitThreadhold = 1024 * 250

	// DefaultChunkSize is the chunk size when a content is automatically
	// splitted
	DefaultChunkSize = 1024 * 50

	// ErrCheckum is returned by addChunk if a checksum error is detected
	ErrCheckum = errors.New("Checksum error")

	// ErrClosedEnvelope is returned by addMessage if the envelope is
	// already closed
	ErrClosedEnvelope = errors.New("Envelope is closed")

	// ErrClosedMessage is returned by Message readers/writers if the message is
	// already closed
	ErrClosedMessage = errors.New("Message is closed")
)

// ChunkWriter writes chunks of message. The chunk sizes are kept and the exact
// same chunks can be read using a ChunkReader when reading the message.
// A single chunk should be small, and if too big will make the resulting
// message impossible to send on the bus. The recommanded maximum size is
// approximatively 768K.
type ChunkWriter interface {
	io.Closer
	Write([]byte) error
	WriteString(string) error
}

// ChunkReader allow reading a messages by chunks as they were written by a
// ChunkWriter
type ChunkReader interface {
	ReadString() (string, error)
	ReadBytes() ([]byte, error)
	DataAvailable() bool
}

// Message is typed and holds some data
type Message interface {
	ID() api.UUID
	MsgType() string
	Checksum() api.Checksum
	// Clone the message. Cannot be called if any content access method was
	// already called
	Clone() Message

	Reader() io.Reader
	ChunkReader() ChunkReader
}

// MessageCloser is a Message that should be closed.
type MessageCloser interface {
	CloseWithError(err error) error
	Message
}

// ReadAllAsString returns a message full content as a string
func ReadAllAsString(m Message) (string, error) {
	b, err := ReadAll(m)
	return string(b), err
}

// ReadAll returns a message full content as a []byte
func ReadAll(m Message) ([]byte, error) {
	return ioutil.ReadAll(m.Reader())
}

// ReadAllChunks returns all the chunks of a ChunkReader
func ReadAllChunks(reader ChunkReader) ([][]byte, error) {
	var (
		chunks [][]byte
	)

	for {
		chunk, err := reader.ReadBytes()
		if err == io.EOF {
			return chunks, nil
		}
		if err != nil {
			return nil, err
		}
		chunks = append(chunks, chunk)
	}
}

// NewTextMessage creates a message with a simple string content
// The content, if too big, may be splitted in smaller chunks
func NewTextMessage(msgType string, content string) (Message, error) {
	return NewBytesMessage(msgType, []byte(content))
}

// MustNewTextMessage creates a message and panics if a message UUID cannot
// be generated, which should be virtually never
func MustNewTextMessage(msgType string, content string) Message {
	msg, err := NewTextMessage(msgType, content)
	if err != nil {
		panic(err)
	}
	return msg
}

// NewTextMessageWithID creates a message with a simple string content
// The content, if too big, may be splitted in smaller chunks
func NewTextMessageWithID(id api.UUID, msgType string, content string) (Message, error) {
	return NewBytesMessageWithID(id, msgType, []byte(content))
}

// NewBytesMessage creates a message with a simple []byte content
// The content, if too big, may be splitted in smaller chunks
func NewBytesMessage(msgType string, content []byte) (Message, error) {
	id, err := api.UUIDV4()
	if err != nil {
		return nil, err
	}
	return NewBytesMessageWithID(id, msgType, content)
}

// MustNewBytesMessage creates a message and panics if a message UUID cannot
// be generated, which should be virtually never
func MustNewBytesMessage(msgType string, content []byte) Message {
	msg, err := NewBytesMessage(msgType, content)
	if err != nil {
		panic(err)
	}
	return msg
}

// NewBytesMessageWithID creates a message with a simple []byte content
// The content, if too big, may be splitted in smaller chunks
func NewBytesMessageWithID(id api.UUID, msgType string, content []byte) (Message, error) {
	if len(content) > ChunkSplitThreadhold {
		m, w := NewMessageWriterWithID(id, msgType)
		if _, err := w.Write(content); err != nil {
			return nil, err
		}
		if err := w.Close(); err != nil {
			return nil, err
		}
		return m, nil
	}

	checksum := api.Checksum(0).
		Update(id[:16]).
		Update([]byte(msgType)).
		Update(content)

	return &SimpleMessage{id, checksum, msgType, content}, nil
}

// NewMessageChunkWriter returns a Message and a chunk writer to write its content.
// The check writer gives controls on the chunk sizes, and will refuse chunks
// that are too big
// The Close method must be called to end the stream
func NewMessageChunkWriter(msgType string) (Message, ChunkWriter, error) {
	id, err := api.UUIDV4()
	if err != nil {
		return nil, nil, err
	}
	m, w := NewMessageChunkWriterWithID(id, msgType, 0)
	return m, w, nil
}

// MustNewMessageChunkWriter returns a Message and a chunk writer to write its content.
// It panics if a UUID cannot be generated
// The check writer gives controls on the chunk sizes, and will refuse chunks
// that are too big
// The Close method must be called to end the stream
func MustNewMessageChunkWriter(msgType string) (Message, ChunkWriter) {
	m, w, err := NewMessageChunkWriter(msgType)
	if err != nil {
		panic(err)
	}
	return m, w
}

// NewMessageChunkWriterWithID returns a Message and a chunk writer to write its content.
// If checksum is non-zero, it will be checked when the chunk writer is closed.
// The check writer gives controls on the chunk sizes, and will refuse chunks
// that are too big
// The Close method must be called to end the stream
func NewMessageChunkWriterWithID(
	id api.UUID, msgType string, checksum api.Checksum,
) (Message, ChunkWriter) {
	m := NewStreamMessage(id, msgType)
	if checksum != 0 {
		m.SetChecksum(checksum)
	}
	return m, NewStreamMessageChunkWriter(m.addChunks)
}

// NewMessageWriter returns a Message and a writer to write its content
// The Close method must be called to end the stream
func NewMessageWriter(msgType string) (Message, io.WriteCloser, error) {
	id, err := api.UUIDV4()
	if err != nil {
		return nil, nil, err
	}
	m, w := NewMessageWriterWithID(id, msgType)
	return m, w, nil
}

// NewMessageWriterWithID returns a Message and a writer to write its content
// The Close method must be called to end the stream
func NewMessageWriterWithID(id api.UUID, msgType string) (Message, io.WriteCloser) {
	m := NewStreamMessage(id, msgType)
	return m, NewStreamMessageWriter(m.addChunks)
}

// SimpleMessage is a non-streamed implementation of Message
type SimpleMessage struct {
	id       api.UUID
	checksum api.Checksum
	msgType  string
	content  []byte
}

// ID returns the message ID
func (m *SimpleMessage) ID() api.UUID {
	return m.id
}

// Checksum returns the message checksum
func (m *SimpleMessage) Checksum() api.Checksum {
	return m.checksum
}

// MsgType returns the message type
func (m *SimpleMessage) MsgType() string {
	return m.msgType
}

// Clone the message
func (m *SimpleMessage) Clone() Message {
	return &SimpleMessage{id: m.id, msgType: m.msgType, content: m.content}
}

// Reader returns a io.Reader on the content
func (m *SimpleMessage) Reader() io.Reader {
	return bytes.NewBuffer(m.content)
}

type plainChunkReader struct {
	content []byte
	read    bool
}

// ReadString returns the whole content as string, only once
func (pcr *plainChunkReader) ReadString() (string, error) {
	if pcr.read {
		return "", io.EOF
	}
	pcr.read = true
	return string(pcr.content), nil
}

// ReadBytes load the content as bytes, only once. If the buffer is too small
// a io.ErrShortBuffer is returned and no data is copied to the buffer
func (pcr *plainChunkReader) ReadBytes() ([]byte, error) {
	if pcr.read {
		return nil, io.EOF
	}
	pcr.read = true
	return pcr.content, nil
}

func (pcr *plainChunkReader) DataAvailable() bool {
	return !pcr.read
}

// ChunkReader returns a ChunkReader on the content
func (m *SimpleMessage) ChunkReader() ChunkReader {
	return &plainChunkReader{content: m.content}
}

// NewStreamMessage creates a StreamMessage
func NewStreamMessage(id api.UUID, msgType string) *StreamMessage {
	m := StreamMessage{
		id:          id,
		msgType:     msgType,
		curChecksum: api.Checksum(0).Update(id[:16]).Update([]byte(msgType)),
	}
	m.cond = sync.NewCond(&m.mut)
	return &m
}

// StreamMessage is a streamed message
type StreamMessage struct {
	id          api.UUID
	msgType     string
	checksum    api.Checksum
	curChecksum api.Checksum

	mut    sync.Mutex
	cond   *sync.Cond
	chunks [][]byte
	err    error
	done   bool

	clones []*StreamMessage
}

type streamMessageChunkReader struct {
	m *StreamMessage
}

func (r *streamMessageChunkReader) DataAvailable() bool {
	r.m.mut.Lock()
	defer r.m.mut.Unlock()

	return r.m.done || r.m.err == nil && len(r.m.chunks) != 0
}

func (r *streamMessageChunkReader) ReadBytes() ([]byte, error) {
	r.m.mut.Lock()

	defer r.m.mut.Unlock()

	for r.m.err == nil && !r.m.done && len(r.m.chunks) == 0 {
		r.m.cond.Wait()
	}
	if r.m.err != nil {
		return nil, r.m.err
	} else if len(r.m.chunks) == 0 {
		if !r.m.done {
			panic("invalid state")
		}
		return nil, io.EOF
	} else {
		chunk := r.m.chunks[0]
		r.m.chunks = r.m.chunks[1:]
		return chunk, nil
	}
}

func (r *streamMessageChunkReader) ReadString() (string, error) {
	b, err := r.ReadBytes()
	if err != nil {
		return "", err
	}
	return string(b), err
}

// NewChunkReaderReader returns a io.Reader that consumes a ChunkReader
func NewChunkReaderReader(chunkReader ChunkReader) io.Reader {
	return &ChunkReaderReader{
		chunkreader: chunkReader,
		buf:         make([]byte, MaxChunkSize),
	}
}

// ChunkReaderReader implements io.Reader on top of ChunkReader
type ChunkReaderReader struct {
	chunkreader ChunkReader

	buf      []byte
	curChunk []byte
}

// Read  implements the io.Reader interface
func (r *ChunkReaderReader) Read(buf []byte) (int, error) {
	// TODO if buf is bigger than curChunk and we have data available,
	// read as much data as we can
	if len(r.curChunk) == 0 {
		chunk, err := r.chunkreader.ReadBytes()
		if err != nil {
			return 0, err
		}
		r.curChunk = chunk
	}
	n := copy(buf, r.curChunk)
	r.curChunk = r.curChunk[n:]
	return n, nil
}

// CloseWithError set a the given error on the message, which will block any
// subsequent read/write
func (m *StreamMessage) CloseWithError(err error) error {
	m.mut.Lock()
	if m.err != nil {
		err := m.err
		m.mut.Unlock()
		return err
	}
	m.err = err
	m.mut.Unlock()
	m.cond.Broadcast()
	return nil
}

func (m *StreamMessage) addChunks(chunks [][]byte) error {
	m.mut.Lock()
	// if append panics we don't want a deadlock in addition
	defer m.mut.Unlock()

	if m.err != nil {
		return m.err
	}

	if chunks == nil {
		if m.checksum == 0 {
			m.checksum = m.curChecksum
		} else {
			if m.curChecksum != m.checksum {
				m.err = ErrCheckum
				m.cond.Broadcast()
				return ErrCheckum
			}
		}
		m.done = true
	} else {
		m.chunks = append(m.chunks, chunks...)
		for _, chunk := range chunks {
			m.curChecksum = m.curChecksum.Update(chunk)
		}
	}

	m.cond.Broadcast()

	for _, clone := range m.clones {
		// XXX is it alright to ignore clone errors ?
		clone.addChunks(chunks)
	}
	return nil
}

// ID returns the message ID
func (m *StreamMessage) ID() api.UUID {
	return m.id
}

// MsgType returns the message type
func (m *StreamMessage) MsgType() string {
	return m.msgType
}

// Checksum returns the message checksum if known, 0 otherwise
func (m *StreamMessage) Checksum() (c api.Checksum) {
	m.mut.Lock()
	c = m.checksum
	m.mut.Unlock()
	return
}

// SetChecksum sets the checksum. If called, the calculated checksum will be
// compared to it on closing
func (m *StreamMessage) SetChecksum(c api.Checksum) {
	m.mut.Lock()
	m.checksum = c
	m.mut.Unlock()
}

// Clone the message. Cannot be called if any content access method was
// already called
func (m *StreamMessage) Clone() Message {
	m.mut.Lock()
	defer m.mut.Unlock()

	clone := NewStreamMessage(m.id, m.msgType)
	if m.chunks != nil {
		clone.addChunks(m.chunks)
	}
	m.clones = append(m.clones, clone)
	return clone
}

// Reader returns a io.Reader for the message
func (m *StreamMessage) Reader() io.Reader {
	return NewChunkReaderReader(m.ChunkReader())
}

// ChunkReader returns a ChunkReader for the message
func (m *StreamMessage) ChunkReader() ChunkReader {
	return &streamMessageChunkReader{m}
}

// NewStreamMessageChunkWriter implements ChunkWriter and output a chunk for reach call to Write
func NewStreamMessageChunkWriter(addChunks func([][]byte) error) ChunkWriter {
	return &streamMessageChunkWriter{addChunks}
}

type streamMessageChunkWriter struct {
	addChunks func([][]byte) error
}

func (w *streamMessageChunkWriter) Write(data []byte) error {
	var chunk = make([]byte, len(data))
	copy(chunk, data)
	return w.addChunks([][]byte{chunk})
}

func (w *streamMessageChunkWriter) WriteString(data string) error {
	return w.addChunks([][]byte{[]byte(data)})
}

func (w *streamMessageChunkWriter) Close() error {
	return w.addChunks(nil)
}

// NewStreamMessageWriter implements io.WriteCloser and auto-split in chunks
func NewStreamMessageWriter(addChunks func([][]byte) error) io.WriteCloser {
	return &streamMessageWriter{
		addChunks: addChunks,
	}
}

type streamMessageWriter struct {
	addChunks func([][]byte) error
	buf       []byte
	started   bool
}

func (w *streamMessageWriter) Write(data []byte) (int, error) {
	var datasize = len(data)
	if !w.started && len(w.buf)+len(data) < ChunkSplitThreadhold {
		w.buf = append(w.buf, data...)
		return datasize, nil
	}

	w.started = true

	var chunks [][]byte

	for len(w.buf)+len(data) >= DefaultChunkSize {
		var (
			chunk []byte
			n     int
		)
		if len(w.buf) > 0 {
			if len(w.buf) > DefaultChunkSize {
				chunk = w.buf[:DefaultChunkSize]
				n = DefaultChunkSize
			} else {
				chunk = w.buf
				n = len(w.buf)
			}
			w.buf = w.buf[n:]
		}
		if n < DefaultChunkSize {
			chunk = append(chunk, data[:(DefaultChunkSize-n)]...)
			data = data[DefaultChunkSize:]
		}
		chunks = append(chunks, chunk)
	}
	w.buf = append([]byte{}, data[:]...)

	if err := w.addChunks(chunks); err != nil {
		return 0, err
	}
	return datasize, nil
}

func (w *streamMessageWriter) Close() error {
	if err := w.addChunks([][]byte{w.buf}); err != nil {
		return err
	}
	return w.addChunks(nil)
}

// ForEachChunk calls a function for each chunk of a message
func ForEachChunk(msg Message, fn func([]byte) error) error {
	reader := msg.ChunkReader()
	for chunk, err := reader.ReadBytes(); err != io.EOF; chunk, err = reader.ReadBytes() {
		if err != nil {
			return err
		}
		if err := fn(chunk); err != nil {
			return err
		}
	}
	return nil
}

// ForEachChunkString calls a function for each chunk of a message
func ForEachChunkString(msg Message, fn func(string) error) error {
	reader := msg.ChunkReader()
	for chunk, err := reader.ReadString(); err != io.EOF; chunk, err = reader.ReadString() {
		if err != nil {
			return err
		}
		if err := fn(chunk); err != nil {
			return err
		}
	}
	return nil
}
