package xbus

import (
	"errors"
	"fmt"
	"io"
	"sync"
	"time"

	api "xbus.io/go-xbus/api"
	"xbus.io/go-xbus/envelope"
	"xbus.io/go-xbus/tools"
)

var (
	// ErrEventTypesNotKnown is returned by functions that work with the
	// event types if they were not yet received.
	ErrEventTypesNotKnown = errors.New("Event types are not known yet")

	// ErrWrongStatus is when addEnvelope is called with a current
	// status != api.EnvelopeAck_RECEIVING
	ErrWrongStatus = errors.New("Receiver is done receiving")

	// ErrTimeout is when a wait operation timed out
	ErrTimeout = errors.New("Timeout")

	// ErrReceptionError is returned by Complete() if current status is error
	ErrReceptionError = errors.New("Reception error")

	// ErrIncompleteEnvelope is when the source is closed and not all fragments
	// where received
	ErrIncompleteEnvelope = errors.New("Incomplete envelope")
)

// EventValidatorMap is a map of EventValidator by EventID
type EventValidatorMap map[api.UUID]tools.EventValidator

// NewEnvelopeReceiver initialize a EnvelopeReceiver
// Deprecated: use envelope.NewFromFragments()
func NewEnvelopeReceiver(input string, e api.Envelope) (*EnvelopeReceiver, error) {
	er := NewEnvelopeReceiverSimple(e.ID)
	er.Input = input

	err := er.AddFragment(e)
	if err != nil {
		return nil, err
	}
	return er, nil
}

// NewEnvelopeReceiverSimple initialize a EnvelopeReceiver without an initial
// fragment
// Deprecated: use envelope.NewFromFragments()
func NewEnvelopeReceiverSimple(id api.UUID) *EnvelopeReceiver {
	builder, err := NewEnvelopeReceiverFromBuilder("", envelope.NewFromFragments(id))
	if err != nil {
		// Since no fragment is added yet, there is no reason to get any error
		panic(err)
	}
	return builder
}

// NewEnvelopeReceiverFromBuilder creates a EnvelopeReceiver from a envelope.BuilderFromFragment
func NewEnvelopeReceiverFromBuilder(input string, builder *envelope.BuilderFromFragment) (*EnvelopeReceiver, error) {
	er := EnvelopeReceiver{
		ID:       builder.Envelope().ID(),
		builder:  builder,
		envelope: builder.Envelope(),

		eventsByType: make(map[string]api.UUID),
		readers:      make(map[api.UUID]envelope.ChunkReader),
		itemCounters: make(map[api.UUID]uint64),
		itemCount:    make(map[api.UUID]uint64),

		status: api.EnvelopeAck_NOSTATUS,
	}
	er.cond = sync.NewCond(&er.mu)
	er.mu.Lock()
	defer er.mu.Unlock()
	if err := er.readMessages(); err != nil {
		er.builder.Cancel(err)
		return nil, err
	}
	return &er, nil
}

// NewEnvelopeReceiverEmpty initialize a EnvelopeReceiver without an initial
// fragment, nor an envelopeID
// Deprecated: use envelope.NewFromFragments()
func NewEnvelopeReceiverEmpty() *EnvelopeReceiver {
	return NewEnvelopeReceiverSimple(api.UUID{})
}

// EnvelopeReceiver handles an incoming envelope
// Deprecated: use envelope.NewFromFragments()
type EnvelopeReceiver struct {
	ID    api.UUID
	Input string

	builder  *envelope.BuilderFromFragment
	envelope envelope.Envelope

	messages     []envelope.Message
	eventTypes   []string
	eventsByType map[string]api.UUID
	readers      map[api.UUID]envelope.ChunkReader
	itemCounters map[api.UUID]uint64
	itemCount    map[api.UUID]uint64

	status    api.EnvelopeAck_ReceptionStatus
	lastError error

	mu   sync.Mutex
	cond *sync.Cond
}

// Cancel should be called by the envelope consumer if for some reasons it
// decides not to consume the envelope anymore
func (r *EnvelopeReceiver) Cancel() {
	r.builder.Cancel(fmt.Errorf("EnvelopeReceiver Cancel"))
}

// Status returns the reception status
func (r *EnvelopeReceiver) Status() api.EnvelopeAck_ReceptionStatus {
	r.mu.Lock()
	defer r.mu.Unlock()
	return r.status
}

// LastError returns the reception status
func (r *EnvelopeReceiver) LastError() error {
	r.mu.Lock()
	defer r.mu.Unlock()
	return r.lastError
}

func (r *EnvelopeReceiver) readMessages() error {
	if r.status == api.EnvelopeAck_NOSTATUS {
		r.status = api.EnvelopeAck_RECEIVING
	}
	for message, err := r.builder.Envelope().NextMessage(false); message != nil || err != nil; message, err = r.builder.Envelope().NextMessage(false) {
		if err == io.EOF {
			if len(r.eventTypes) == 0 {
				for _, message := range r.messages {
					r.eventTypes = append(r.eventTypes, message.MsgType())
					r.eventsByType[message.MsgType()] = message.ID()
				}
			}
			r.cond.Broadcast()
			break
		}
		if err != nil {
			r.SetError(err)
			return err
		}
		r.messages = append(r.messages, message)
		r.readers[message.ID()] = message.ChunkReader()
	}
	if r.builder.Complete {
		r.status = api.EnvelopeAck_ACCEPTED
		// The event types must be know before we returned (by contract). Wait
		// until the message consumption routine is over
		r.waitEventTypes(time.Second)
		r.cond.Broadcast()
	}
	return nil
}

// AddFragment add an envelope fragment
func (r *EnvelopeReceiver) AddFragment(e api.Envelope) error {
	r.mu.Lock()
	defer r.mu.Unlock()

	err := r.builder.Add(&e)
	if err != nil {
		r.setError(err)
		return err
	}

	if err := r.readMessages(); err != nil {
		return err
	}

	return nil
}

// Finalize signals the end of the fragments stream. If fragment(s) are missing,
// the receiver state will change to ERROR.
func (r *EnvelopeReceiver) Finalize() error {
	r.mu.Lock()
	defer r.mu.Unlock()

	if r.status == api.EnvelopeAck_ERROR {
		return ErrWrongStatus
	}

	if r.status == api.EnvelopeAck_RECEIVING {
		r.setError(ErrIncompleteEnvelope)
		return r.lastError
	}

	return nil
}

// SetError change the receiver state to ERROR with a message
func (r *EnvelopeReceiver) SetError(err error) {
	r.mu.Lock()
	defer r.mu.Unlock()

	r.setError(err)
}

// setError change the receiver state to ERROR with a message
func (r *EnvelopeReceiver) setError(err error) {
	r.status = api.EnvelopeAck_ERROR
	r.lastError = err
	if r.builder != nil {
		r.builder.Cancel(err)
	}
	r.cond.Broadcast()
}

// EventTypes returns the types of event in the envelope, or nil if
// the types are not all known yet.
func (r *EnvelopeReceiver) EventTypes() []string {
	r.mu.Lock()
	defer r.mu.Unlock()

	return r.eventTypes
}

// Events returns all information available about the events except their items.
// Returns nil if the event types are not known yet
func (r *EnvelopeReceiver) Events() []api.Event {
	r.mu.Lock()
	defer r.mu.Unlock()

	if len(r.eventTypes) == 0 {
		return nil
	}

	return r.PartialEvents()
}

// PartialEvents returns all information available about the already known
// events except their items.
func (r *EnvelopeReceiver) PartialEvents() []api.Event {
	var events []api.Event

	for _, message := range r.messages {
		events = append(events, api.Event{
			ID:        message.ID(),
			Type:      message.MsgType(),
			ItemCount: r.itemCount[message.ID()],
			Checksum:  message.Checksum(),
		})
	}
	return events
}

// WaitEventTypes returns the types of the event, or nil if the timeout
// expires and the types are still unknown
func (r *EnvelopeReceiver) WaitEventTypes(timeout time.Duration) []string {
	r.mu.Lock()
	defer r.mu.Unlock()
	return r.waitEventTypes(timeout)
}

func (r *EnvelopeReceiver) waitEventTypes(timeout time.Duration) []string {
	var eventtypes = r.eventTypes

	if eventtypes != nil || timeout == 0 {
		return eventtypes
	}

	deadline := time.Now().Add(timeout)

	// XXX I suspect a race-condition is possible if the timer is not precise
	// enough and hits the condition just before Now().Before(deadline) becomes
	// false
	// We need to investigate this a little more, maybe unittests will help
	timer := time.AfterFunc(timeout, func() {
		r.mu.Lock()
		r.cond.Broadcast()
		r.mu.Unlock()
	})
	defer timer.Stop()

	for eventtypes == nil && time.Now().Before(deadline) {
		r.cond.Wait()
		eventtypes = r.eventTypes
	}

	return eventtypes
}

// waitNextItems returns at most 'max' items from the given eventType
// as soon a they are available. If not and the status is receiving, it waits
func (r *EnvelopeReceiver) waitNextItems(eventID api.UUID, max int) (items []api.Item) {
	r.mu.Lock()

	for r.status == api.EnvelopeAck_RECEIVING &&
		!r.readers[eventID].DataAvailable() {
		r.cond.Wait()
	}
	r.mu.Unlock()

	return r.NextEventItems(eventID, max)
}

// NextItems retrieve at most 'max' items from the given eventType
func (r *EnvelopeReceiver) NextItems(eventType string, max int) (items []api.Item) {
	r.mu.Lock()
	eventID := r.eventsByType[eventType]
	r.mu.Unlock()

	return r.NextEventItems(eventID, max)
}

// NextEventItems retrieve at most 'max' items from the given event ID
func (r *EnvelopeReceiver) NextEventItems(eventID api.UUID, max int) (items []api.Item) {
	r.mu.Lock()
	defer r.mu.Unlock()

	count := 0
	reader, ok := r.readers[eventID]
	if !ok {
		panic("No reader for event: " + eventID.String())
	}

	for reader.DataAvailable() && count < max {
		count++
		chunk, err := reader.ReadBytes()
		if err == io.EOF {
			r.itemCount[eventID] = r.itemCounters[eventID]
			break
		}
		if err != nil {
			return nil
		}
		items = append(items, api.Item(chunk))
		r.itemCounters[eventID] = r.itemCounters[eventID] + 1
	}

	return
}

// WaitCompletion waits until the full envelope was received.
// Returns false if the reception is not over after timeout
func (r *EnvelopeReceiver) WaitCompletion(timeout time.Duration) bool {
	select {
	case <-time.After(timeout):
		return false
	case <-r.Done():
		return true
	}
}

// GetItemChan returns a channel on which items are sent by a goroutine
// as soon as they are received.
// The channel is closed as soon as the reception is stopped
// The function should be called once at most for each event type
func (r *EnvelopeReceiver) GetItemChan(eventType string) <-chan api.Item {
	c := make(chan api.Item)
	eventID, ok := r.eventsByType[eventType]
	if !ok {
		panic("No such event type in the envelope: " + eventType)
	}
	reader, ok := r.readers[eventID]
	if !ok {
		panic("No reader for event: " + eventID.String())
	}

	go func(reader envelope.ChunkReader, c chan<- api.Item) {
		for {
			chunk, err := reader.ReadBytes()
			if err != nil {
				close(c)
				return
			}
			c <- api.Item(chunk)
		}
	}(reader, c)
	return c
}

// Done is a chan that is closed when recetion is over
func (r *EnvelopeReceiver) Done() <-chan struct{} {
	return r.builder.Done()
}

// Complete returns the complete envelope.
func (r *EnvelopeReceiver) Complete(timeout time.Duration) (*api.Envelope, error) {
	if !r.WaitCompletion(timeout) {
		return nil, ErrTimeout
	}
	r.mu.Lock()
	defer r.mu.Unlock()
	if r.status == api.EnvelopeAck_ERROR {
		if r.lastError != nil {
			return nil, r.lastError
		}
		return nil, ErrReceptionError
	}
	env := api.Envelope{
		ID:   r.ID,
		Last: true,
	}
	for _, message := range r.messages {
		chunks, err := envelope.ReadAllChunks(r.readers[message.ID()])
		if err != nil {
			return nil, err
		}
		env.EventIDs = append(env.EventIDs, message.ID())
		env.Events = append(env.Events, api.Event{
			ID:        message.ID(),
			Type:      message.MsgType(),
			Index:     1,
			Items:     api.ChunksAsItems(chunks),
			ItemCount: uint64(len(chunks)),
			Checksum:  message.Checksum(),
		})
	}
	return &env, nil
}
