package xbus

import (
	"context"
	"sync"
	"time"

	api "xbus.io/go-xbus/api"
	"xbus.io/go-xbus/envelope"
)

var (
	// DefaultMaxChunksByFragment is the default max number of chunks by fragment
	DefaultMaxChunksByFragment = 1000
)

// SendEnvelopeFunc ...
type SendEnvelopeFunc func(*api.Envelope) (api.EnvelopeAck_ReceptionStatus, error)

// NewEnvelopeSender creates a new envelope sender
func NewEnvelopeSender(
	eventTypes []string,
	maxSize int,
	sendEnvelope SendEnvelopeFunc,
) *EnvelopeSender {

	sender := EnvelopeSender{
		sendEnvelope:   sendEnvelope,
		maxSize:        maxSize,
		maxItems:       DefaultMaxChunksByFragment,
		builder:        envelope.MustNewOpenedEnvelope(),
		eventIndexes:   make(map[string]int),
		eventIDIndexes: make(map[api.UUID]int),
		noItems:        true,
		sendResult:     make(chan error),
	}

	for _, eventType := range eventTypes {
		if _, err := sender.AddEvent(eventType); err != nil {
			panic(err)
		}
	}

	return &sender
}

// EnvelopeSender streams items as one or several envelopes
// to a given Actor output.
type EnvelopeSender struct {
	sendEnvelope SendEnvelopeFunc

	builder *envelope.Builder

	eventIndexes   map[string]int
	eventIDIndexes map[api.UUID]int
	writers        []envelope.ChunkWriter
	noItems        bool

	status api.EnvelopeAck_ReceptionStatus

	maxSize  int
	maxDelay time.Duration
	maxItems int

	mu sync.Mutex

	sendResult chan error
	ctx        context.Context
	ctxCancel  func()
}

func (s *EnvelopeSender) doSend() {
	s.sendResult <- SendEnvelope(
		s.ctx, s.builder, s.maxSize, s.maxDelay, s.maxItems, s.sendEnvelope)
}

// SetMaxDelay set the maximum delay between two 'send' calls. If no data is
// available and empty fragment is sent.
func (s *EnvelopeSender) SetMaxDelay(delay time.Duration) {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.maxDelay = delay
}

// GetID returns the ID of the envelope
func (s *EnvelopeSender) GetID() api.UUID {
	return s.builder.ID()
}

// FreezeEventList signal that no more event will be added
func (s *EnvelopeSender) FreezeEventList() {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.builder.Close()
}

func (s *EnvelopeSender) addItems(index int, items []api.Item) error {
	if s.noItems {
		s.noItems = false
		s.ctx, s.ctxCancel = context.WithCancel(context.Background())
		go s.doSend()
	}
	writer := s.writers[index]
	for _, item := range items {
		if err := writer.Write(item); err != nil {
			return err
		}
	}
	return nil
}

// AddItems pushes items to the event of the given type in the current envelope
// This function is deprecated, AddEventItems is a safer alternative
func (s *EnvelopeSender) AddItems(eventType string, items ...api.Item) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	index, ok := s.eventIndexes[eventType]
	if !ok {
		panic("Invalid eventType")
	}
	return s.addItems(index, items)
}

// AddEventWithID adds an event to the envelope (no items)
func (s *EnvelopeSender) AddEventWithID(id api.UUID, eventType string) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	msg, writer := envelope.NewMessageChunkWriterWithID(id, eventType, 0)

	if err := s.builder.Add(msg); err != nil {
		return err
	}

	if _, ok := s.eventIndexes[eventType]; !ok {
		s.eventIndexes[eventType] = len(s.writers)
	}
	s.eventIDIndexes[id] = len(s.writers)

	s.writers = append(s.writers, writer)

	return nil
}

// AddEvent adds an event to the envelope (no items)
func (s *EnvelopeSender) AddEvent(eventType string) (api.UUID, error) {
	id, err := api.UUIDV4()
	if err != nil {
		return api.UUID{}, nil
	}
	return id, s.AddEventWithID(id, eventType)
}

// AddEventItems pushes chunks to the given event
func (s *EnvelopeSender) AddEventItems(id api.UUID, items ...api.Item) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	index, ok := s.eventIDIndexes[id]
	if !ok {
		panic("Invalid event id")
	}
	return s.addItems(index, items)
}

// CloseEvent marks the end of items for an event
func (s *EnvelopeSender) CloseEvent(id api.UUID, checksum api.Checksum) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	index := s.eventIDIndexes[id]
	if err := s.writers[index].Close(); err != nil {
		return err
	}
	s.writers[index] = nil
	return nil
}

// Flush forces the buffered items to be sent right away
// It is now a noop
func (s *EnvelopeSender) Flush() {
}

// Cancel closes the steams without finalizing the sending
func (s *EnvelopeSender) Cancel() {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.builder.CloseWithError(ErrEmissionCanceled)

	if !s.noItems {
		s.ctxCancel()
		<-s.sendResult
	}
}

// Close closes the streams. Must be called to finalize the emission, except
// if Cancel() was called before
func (s *EnvelopeSender) Close() error {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.builder.Close()

	for _, w := range s.writers {
		if w != nil {
			w.Close()
		}
	}

	if s.noItems {
		status, err := s.sendEnvelope(nil)
		if err != nil {
			s.status = api.EnvelopeAck_ERROR
			return err
		}
		s.status = status
	} else {
		// TODO cancel it if too long?
		defer s.ctxCancel()
		return <-s.sendResult
	}

	return nil
}

// SendEnvelope sends a Envelope in chunks
func SendEnvelope(
	ctx context.Context,
	env envelope.Envelope,
	maxSize int,
	maxDelay time.Duration,
	maxChunks int,
	sendEnvelope SendEnvelopeFunc,
) error {
	var (
		errC = make(chan error)
	)

	sender := NewLegacyEnvelopeSenderWithIDs(
		env.ID(),
		nil,
		nil,
		maxSize,
		sendEnvelope,
	)
	sender.SetMaxDelay(maxDelay)
	sender.maxItems = maxChunks

	go func() {
		var wg sync.WaitGroup

		// Start reading all messages
		if err := envelope.ForEachMessage(env, func(m envelope.Message) error {
			sender.AddEventWithID(m.ID(), m.MsgType())
			wg.Add(1)
			go func() {
				defer wg.Done()
				if err := envelope.ForEachChunk(m, func(chunk []byte) error {
					return sender.AddEventItems(m.ID(), chunk)
				}); err != nil {
					errC <- err
				}
				if err := sender.CloseEvent(m.ID(), m.Checksum()); err != nil {
					errC <- err
				}
				return
			}()
			return nil
		}); err != nil {
			errC <- err
		}
		sender.FreezeEventList()

		wg.Wait()
		// Signal that we finished reading the envelope
		close(errC)
	}()

	// flush the errC chan on exit
	defer func() {
		for range errC {
		}
	}()

	select {
	case err := <-errC:
		if err == nil {
			return sender.Close()
		}
		sender.Cancel()
		return err
	case <-ctx.Done():
		env.CloseWithError(ctx.Err())
		sender.Cancel()
		return ctx.Err()
	}
}
