package xbus

import api "xbus.io/go-xbus/api"

// ItemQueue is a queue for items
// It is not thread safe, and would most probably benefit from a ring-buffer
// based implementation to avoid excessive reallocation when the items are
// consumed quickly
type ItemQueue []api.Item

// Add adds item(s) to the queue
func (q *ItemQueue) Add(items ...api.Item) {
	*q = append(*q, items...)
}

// Remove removes the first element and returns it
func (q *ItemQueue) Remove() api.Item {
	if len(*q) == 0 {
		panic("Remove called on empty queue")
	}
	item := (*q)[0]
	*q = (*q)[1:]
	return item
}

// RemoveN at most the N first elements and returns them
func (q *ItemQueue) RemoveN(n int) []api.Item {
	if len(*q) == 0 {
		panic("Remove called on empty queue")
	}
	if len(*q) < n {
		n = len(*q)
	}
	items := (*q)[:n]
	*q = (*q)[n:]
	return items
}

// RemoveS returns items with a total size not exceeding maxBytes
func (q *ItemQueue) RemoveS(maxBytes int) ([]api.Item, int) {
	var size, index int
	for i, item := range *q {
		nSize := size + len(item)
		if nSize <= maxBytes {
			size = nSize
			index = i
		} else {
			break
		}
	}
	return q.RemoveN(index + 1), size
}

// RemoveNS returns items with a total size not exceeding maxBytes,
// and a count not exceeding n
func (q *ItemQueue) RemoveNS(n, maxBytes int) ([]api.Item, int) {
	var (
		size  int
		index = -1
	)
	for i, item := range *q {
		nSize := size + len(item)
		if nSize <= maxBytes {
			size = nSize
			index = i
			if index+1 == n {
				break
			}
		} else {
			break
		}
	}
	if index == -1 {
		return nil, 0
	}
	return q.RemoveN(index + 1), size
}
