package xbus

import (
	"errors"
	"fmt"
	"sync"
	"time"

	api "xbus.io/go-xbus/api"
	"xbus.io/go-xbus/envelope"
)

var (
	// ErrEventListFrozen is returns when the event list is complete and one
	// attempt to add a new event
	ErrEventListFrozen = errors.New("Event list is frozen")
)

// NewLegacyEnvelopeSender creates a new envelope sender
func NewLegacyEnvelopeSender(
	eventTypes []string,
	maxSize int,
	sendEnvelope SendEnvelopeFunc,
) *LegacyEnvelopeSender {
	return NewLegacyEnvelopeSenderWithIDs(
		api.UUID{},
		eventTypes,
		make([]api.UUID, len(eventTypes)),
		maxSize,
		sendEnvelope,
	)
}

// NewLegacyEnvelopeSenderWithIDs returns a LegacyEnvelopeSender with forced envelope and
// event ids. If any ID is zero, an ID is forged
func NewLegacyEnvelopeSenderWithIDs(
	envelopeID api.UUID,
	eventTypes []string,
	eventIDs []api.UUID,
	maxSize int,
	sendEnvelope SendEnvelopeFunc,
) *LegacyEnvelopeSender {
	if envelopeID.IsZero() {
		var err error
		envelopeID, err = api.UUIDV4()
		if err != nil {
			panic(err)
		}
	}
	es := LegacyEnvelopeSender{
		sendEnvelope:   sendEnvelope,
		itemQueues:     make([]ItemQueue, len(eventTypes)),
		eventIndexes:   make(map[string]int),
		eventIDIndexes: make(map[api.UUID]int),
		maxSize:        maxSize,
		maxItems:       1000,
		envelope: api.Envelope{
			ID: envelopeID,
		},
	}
	for i, eType := range eventTypes {
		eventID := eventIDs[i]
		if eventID.IsZero() {
			var err error
			eventID, err = api.UUIDV4()
			if err != nil {
				panic(err)
			}
		}
		es.AddEventWithID(eventID, eType)
	}
	if len(eventTypes) != 0 {
		es.FreezeEventList()
	}
	return &es
}

// LegacyEnvelopeSender streams items as one or several envelopes
// to a given Actor output.
type LegacyEnvelopeSender struct {
	sendEnvelope SendEnvelopeFunc
	queuesSize   int
	itemQueues   []ItemQueue
	envelope     api.Envelope
	status       api.EnvelopeAck_ReceptionStatus
	maxSize      int
	maxItems     int

	minDelay       time.Duration
	keepAliveTimer *time.Timer

	eventListComplete bool
	eventListSent     bool
	eventIndexes      map[string]int
	eventIDIndexes    map[api.UUID]int
	closedEvents      []bool
	finishedEvents    []bool

	mu sync.Mutex
}

// send emits a fragment
func (s *LegacyEnvelopeSender) send(fragment *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
	if s.keepAliveTimer != nil {
		s.keepAliveTimer.Reset(s.minDelay)
	}
	return s.sendEnvelope(fragment)
}

// SetMaxDelay set the maximum delay between two 'send' calls. If no data is
// available and empty fragment is sent.
func (s *LegacyEnvelopeSender) SetMaxDelay(delay time.Duration) {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.minDelay = delay

	if delay == 0 {
		if s.keepAliveTimer != nil {
			s.keepAliveTimer.Stop()
			s.keepAliveTimer = nil
		}
		return
	}
	if s.keepAliveTimer == nil {
		s.keepAliveTimer = time.NewTimer(delay)
		go func(c <-chan time.Time) {
			for range c {
				s.Flush()
			}
		}(s.keepAliveTimer.C)
	} else {
		s.keepAliveTimer.Reset(delay)
	}
}

// GetID returns the ID of the envelope
func (s *LegacyEnvelopeSender) GetID() api.UUID {
	return s.envelope.ID
}

// FreezeEventList signal that no more event will be added
func (s *LegacyEnvelopeSender) FreezeEventList() {
	s.mu.Lock()
	defer s.mu.Unlock()

	s.eventListComplete = true
}

func (s *LegacyEnvelopeSender) addItems(index int, items []api.Item) error {
	s.itemQueues[index].Add(items...)
	event := &s.envelope.Events[index]
	event.ItemCount += uint64(len(items))
	for _, item := range items {
		event.Checksum = event.Checksum.Update(item)
		s.queuesSize += len(item)
	}
	return s.autoFlush()
}

// AddItems pushes items to the event of the given type in the current envelope
// This function is deprecated, AddEventItems is a safer alternative
func (s *LegacyEnvelopeSender) AddItems(eventType string, items ...api.Item) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	index, ok := s.eventIndexes[eventType]
	if !ok {
		panic("Invalid eventType")
	}
	return s.addItems(index, items)
}

// AddEventWithID adds an event to the envelope (no items)
func (s *LegacyEnvelopeSender) AddEventWithID(id api.UUID, eventType string) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.eventListComplete {
		return ErrEventListFrozen
	}

	if _, ok := s.eventIndexes[eventType]; !ok {
		s.eventIndexes[eventType] = len(s.envelope.Events)
	}
	s.eventIDIndexes[id] = len(s.envelope.Events)

	s.envelope.EventIDs = append(s.envelope.EventIDs, id)
	s.envelope.Events = append(s.envelope.Events, api.Event{
		ID:       id,
		Type:     eventType,
		Index:    0,
		Checksum: api.Checksum(0).Update(id[:]).Update([]byte(eventType)),
	})
	s.itemQueues = append(s.itemQueues, ItemQueue{})
	s.closedEvents = append(s.closedEvents, false)
	s.finishedEvents = append(s.finishedEvents, false)
	return nil
}

// AddEvent adds an event to the envelope (no items)
func (s *LegacyEnvelopeSender) AddEvent(eventType string) (api.UUID, error) {
	id, err := api.UUIDV4()
	if err != nil {
		return api.UUID{}, nil
	}
	return id, s.AddEventWithID(id, eventType)
}

// AddEventItems pushes chunks to the given event
func (s *LegacyEnvelopeSender) AddEventItems(id api.UUID, items ...api.Item) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	index, ok := s.eventIDIndexes[id]
	if !ok {
		panic("Invalid event id")
	}
	return s.addItems(index, items)
}

// CloseEvent marks the end of items for an event
func (s *LegacyEnvelopeSender) CloseEvent(id api.UUID, checksum api.Checksum) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	index, ok := s.eventIDIndexes[id]
	if !ok {
		panic("Invalid event id")
	}
	if checksum != 0 {
		if s.envelope.Events[index].Checksum != checksum {
			return envelope.ErrCheckum
		}
	}
	s.closedEvents[index] = true
	return nil
}

func (s *LegacyEnvelopeSender) nextFragment() (api.Envelope, bool) {
	env := api.Envelope{
		ID: s.envelope.ID,
	}
	if s.eventListComplete && !s.eventListSent {
		env.EventIDs = s.envelope.EventIDs
		s.eventListSent = true
	}
	envSize := 0
	for i := range s.envelope.Events {
		if s.finishedEvents[i] {
			continue
		}
		event := &s.envelope.Events[i]

		var items []api.Item
		if len(s.itemQueues[i]) > 0 {
			var size int
			items, size = s.itemQueues[i].RemoveNS(s.maxItems, s.maxSize-envSize)
			s.queuesSize -= size
			envSize += size
		}

		// Skip empty event if it does not need to be finalized and is already
		// started
		if !s.closedEvents[i] && len(items) == 0 && event.Index != 0 {
			continue
		}

		event.Index++

		env.Events = append(env.Events, api.Event{
			ID:    event.ID,
			Type:  event.Type,
			Index: event.Index,
			Items: items,
		})
		if s.closedEvents[i] && len(s.itemQueues[i]) == 0 && !s.finishedEvents[i] {
			env.Events[len(env.Events)-1].ItemCount = event.ItemCount
			env.Events[len(env.Events)-1].Checksum = event.Checksum
			s.finishedEvents[i] = true
		}
	}
	if s.queuesSize != 0 && envSize == 0 {
		panic(fmt.Sprintf(
			"nextFragment produced an empty envelope but queuesize (%d) is not zero",
			s.queuesSize))
	}
	return env, s.queuesSize == 0
}

// autoFlush emit envelopes until queue sizes is lower that the cache limit
func (s *LegacyEnvelopeSender) autoFlush() error {
	biggestQueueLen := func() int {
		n := 0
		for _, q := range s.itemQueues {
			if len(q) > n {
				n = len(q)
			}
		}
		return n
	}
	for biggestQueueLen() > s.maxItems || s.queuesSize > s.maxSize {
		env, _ := s.nextFragment()
		status, err := s.send(&env)
		if err != nil {
			return err
		}
		s.status = status
	}
	return nil
}

// flush sends out all the fragments expect the last and returns it
func (s *LegacyEnvelopeSender) flush() (*api.Envelope, error) {
	if s.status == api.EnvelopeAck_ACCEPTED ||
		s.status == api.EnvelopeAck_ERROR {
		return nil, errors.New("invalid state, cannot flush")
	}
	for {
		env, last := s.nextFragment()
		if last {
			return &env, nil
		}
		status, err := s.send(&env)
		if err != nil {
			panic(err)
		}
		s.status = status
		if s.status != api.EnvelopeAck_RECEIVING {
			return nil, errors.New("invalid state, cannot continue to flush")
		}
	}
}

// Flush forces the buffered items to be sent right away
func (s *LegacyEnvelopeSender) Flush() {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.status == api.EnvelopeAck_ACCEPTED ||
		s.status == api.EnvelopeAck_ERROR {
		return
	}
	for {
		env, last := s.nextFragment()
		status, err := s.send(&env)
		if err != nil {
			panic(err)
		}
		s.status = status
		if last {
			break
		}
	}
}

// Cancel closes the steams without finalizing the sending
func (s *LegacyEnvelopeSender) Cancel() {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.keepAliveTimer != nil {
		s.keepAliveTimer.Stop()
		s.keepAliveTimer = nil
	}

	s.status = api.EnvelopeAck_ERROR
}

// Close closes the streams. Must be called to finalize the emission, except
// if Cancel() was called before
func (s *LegacyEnvelopeSender) Close() error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.keepAliveTimer != nil {
		s.keepAliveTimer.Stop()
		s.keepAliveTimer = nil
	}

	// Force closing remaining events and make sure EventIDs is sent
	s.eventListComplete = true
	for i := range s.closedEvents {
		s.closedEvents[i] = true
	}

	env, err := s.flush()
	if err != nil {
		return err
	}

	var noItems bool
	for _, event := range s.envelope.Events {
		if event.ItemCount != 0 {
			noItems = false
			break
		}
	}
	env.Last = true
	var (
		status api.EnvelopeAck_ReceptionStatus
	)
	if noItems {
		status, err = s.send(nil)
	} else {
		status, err = s.send(env)
	}
	if err != nil {
		return err
	}
	s.status = status
	if status != api.EnvelopeAck_ACCEPTED {
		return fmt.Errorf("legacy sender expects final status to be 'ACCEPTED', got '%s'", status)
	}
	return nil
}
